/**
 * Title :	Property view
 * Description: Property editor
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package bsys.propertyview;

import java.awt.Component;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.io.File;
import java.net.URISyntaxException;
import java.util.EventObject;
import java.util.Set;

import javax.swing.CellEditor;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.table.TableCellEditor;

import bsys.propertyview.PropertyView.Attribute;
import bsys.propertyview.PropertyView.Handler;

/**
 * Versatile super editor
 * 
 * @author Bernard Bou
 */
public class Editor implements TableCellEditor, CellEditorListener
{
	/**
	 * View
	 */
	private final PropertyView thePropertyView;

	/**
	 * Handler used to retrieve id map
	 */
	private Handler theHandler;

	/**
	 * Attribute being edited
	 */
	private Attribute theAttribute;

	/**
	 * Color editor
	 */
	private final ColorEditor theColorEditor;

	/**
	 * File editor
	 */
	private final FileEditor theFileEditor;

	/**
	 * List editor
	 */
	private final ListEditor theListEditor;

	/**
	 * Long text editor
	 */
	private final LongTextEditor theLongTextEditor;

	/**
	 * Default editor
	 */
	private final DefaultCellEditor theDefaultEditor;

	/**
	 * The delegate subeditor
	 */
	private TableCellEditor theSubEditor;

	/**
	 * The subeditors
	 */
	private final TableCellEditor[] theSubEditors;

	/**
	 * Constructor
	 * 
	 * @param thisPropertyView
	 *        view
	 */
	public Editor(final PropertyView thisPropertyView)
	{
		this.thePropertyView = thisPropertyView;

		// subeditors
		this.theColorEditor = new ColorEditor();
		this.theFileEditor = new FileEditor();
		this.theListEditor = new ListEditor();
		this.theLongTextEditor = new LongTextEditor();
		this.theDefaultEditor = new DefaultCellEditor(new JTextField());
		this.theSubEditors = new TableCellEditor[] { this.theColorEditor, this.theFileEditor, this.theListEditor, this.theLongTextEditor, this.theDefaultEditor };

		// edited attribute
		this.theAttribute = null;
	}

	/**
	 * Set get/set handlers
	 * 
	 * @param thisHandler
	 *        handler
	 */
	public void setHandler(final Handler thisHandler)
	{
		this.theHandler = thisHandler;
	}

	// I N T E R F A C E

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableCellEditor#getTableCellEditorComponent(javax.swing.JTable, java.lang.Object, boolean, int, int)
	 */
	@Override
	public Component getTableCellEditorComponent(final JTable table, final Object thisValue, final boolean isSelected, final int row, final int column)
	{
		if (column == PropertyView.VALUE)
		{
			this.theAttribute = (Attribute) thisValue;

			// pick subeditor
			switch (this.theAttribute.theDescriptor.theType)
			{
			// color
			case COLOR:
				this.theSubEditor = this.theColorEditor;
				break;

			// lists
			case STROKE:
				this.theListEditor.setEditable(false);
				this.theListEditor.setImageMap(PropertyView.theStrokeIcons);
				populateWithStrokes(this.theListEditor.getComboBox());
				this.theSubEditor = this.theListEditor;
				break;

			case TERMINATOR:
				this.theListEditor.setImageMap(PropertyView.theTerminatorIcons);
				this.theListEditor.setEditable(false);
				populateWithTerminators(this.theListEditor.getComboBox());
				this.theSubEditor = this.theListEditor;
				break;

			case FONTFACE:
				this.theListEditor.setImageMap(null);
				this.theListEditor.setEditable(false);
				populateWithFonts(this.theListEditor.getComboBox());
				this.theSubEditor = this.theListEditor;
				break;

			case BOOLEAN:
				this.theListEditor.setImageMap(null);
				this.theListEditor.setEditable(false);
				populateWithTrueFalse(this.theListEditor.getComboBox());
				this.theSubEditor = this.theListEditor;
				break;

			case REFID:
				this.theListEditor.setImageMap(null);
				this.theListEditor.setEditable(false);
				populateWithIds(this.theListEditor.getComboBox());
				this.theSubEditor = this.theListEditor;
				break;

			case LINK:
				this.theListEditor.setImageMap(null);
				this.theListEditor.setEditable(true);
				populateWithUrls(this.theListEditor.getComboBox());
				addPopulateWithSharpIds(this.theListEditor.getComboBox());
				this.theSubEditor = this.theListEditor;
				break;

			// files
			case IMAGE:
				this.theFileEditor.setDirectory(getImageRepository());
				this.theSubEditor = this.theFileEditor;
				break;

			// text
			case LONGTEXT:
				this.theSubEditor = this.theLongTextEditor;
				break;

			case TEXT:
				if (this.theAttribute.theDescriptor.thePossibleValues != null)
				{
					this.theListEditor.setImageMap(null);
					this.theListEditor.setEditable(false);
					populateWithStrings(this.theListEditor.getComboBox(), this.theAttribute.theDescriptor.thePossibleValues);
					this.theSubEditor = this.theListEditor;
					break;
				}
				// fall through

				// text
			case INTEGER:
			case DOUBLE:
			case ID:
			case FONTSIZE:
			default:
				this.theSubEditor = this.theDefaultEditor;
				break;
			}

			// invoke subeditor
			return this.theSubEditor.getTableCellEditorComponent(table, this.theAttribute.theValue, isSelected, row, column);
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.CellEditor#getCellEditorValue()
	 */
	@Override
	public Object getCellEditorValue()
	{
		switch (this.theAttribute.theDescriptor.theType)
		{
		case BOOLEAN:
		{
			final String thisValue = (String) this.theSubEditor.getCellEditorValue();
			if (thisValue == null || thisValue.equals(PropertyView.theDefaultString))
				return null;
			return Boolean.valueOf(thisValue);
		}

		case INTEGER:
		case FONTSIZE:
		{
			final String thisValue = (String) this.theSubEditor.getCellEditorValue();
			if (thisValue == null || thisValue.equals(PropertyView.theDefaultString))
				return null;
			try
			{
				return Integer.valueOf(thisValue);
			}
			catch (final NumberFormatException e)
			{
				// do nothing
			}
			return null;
		}

		case DOUBLE:
		{
			final String thisValue = (String) this.theSubEditor.getCellEditorValue();
			if (thisValue == null || thisValue.equals(PropertyView.theDefaultString))
				return null;
			try
			{
				return Double.valueOf(thisValue);
			}
			catch (final NumberFormatException e)
			{
				// do nothing
			}
			return null;
		}

		case COLOR:
		{
			return this.theSubEditor.getCellEditorValue();
		}

		default:
		{
			final String thisValue = (String) this.theSubEditor.getCellEditorValue();
			if (thisValue == null || thisValue.isEmpty() || thisValue.equals(PropertyView.theDefaultString))
				return null;
			return thisValue;
		}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.CellEditor#shouldSelectCell(java.util.EventObject)
	 */
	@Override
	public boolean shouldSelectCell(final EventObject thisEvent)
	{
		return this.theSubEditor.shouldSelectCell(thisEvent);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.CellEditor#stopCellEditing()
	 */
	@Override
	public boolean stopCellEditing()
	{
		return this.theSubEditor.stopCellEditing();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.DefaultCellEditor#cancelCellEditing()
	 */
	@Override
	public void cancelCellEditing()
	{
		this.theSubEditor.cancelCellEditing();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.CellEditor#isCellEditable(java.util.EventObject)
	 */
	@Override
	public boolean isCellEditable(final EventObject thisAnEvent)
	{
		return true;
	}

	// H E L P E R S

	/**
	 * Get image repository
	 * 
	 * @return image repository as file
	 */
	private File getImageRepository()
	{
		try
		{
			return new File(this.thePropertyView.theImageRepository.toURI());
		}
		catch (final URISyntaxException e)
		{
			// do nothing
		}
		return null;
	}

	// P O P U L A T E . C O M B O B O X

	/**
	 * Populate combobox with ids
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void populateWithIds(final JComboBox thisCombo)
	{
		final Set<String> theseIds = this.theHandler.theIdGetter.ids();
		thisCombo.removeAllItems();
		thisCombo.addItem(PropertyView.theDefaultString);
		for (final String thisString : theseIds)
		{
			thisCombo.addItem(thisString);
		}
	}

	/**
	 * Populate combobox with stroke types
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void populateWithStrokes(final JComboBox thisCombo)
	{
		populateWithStrings(thisCombo, PropertyView.theStrokeStrings);
	}

	/**
	 * Populate combobox with true/false
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void populateWithTrueFalse(final JComboBox thisCombo)
	{
		populateWithStrings(thisCombo, new String[] { "false", "true" }); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Populate combobox with urls
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void populateWithUrls(final JComboBox thisCombo)
	{
		populateWithStrings(thisCombo, new String[] { "http://", "ftp://" }); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Populate combobox with ids referencing nodes
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void addPopulateWithSharpIds(final JComboBox thisCombo)
	{
		final Set<String> theseIds = this.theHandler.theIdGetter.ids();
		for (final String thisString : theseIds)
		{
			thisCombo.addItem("#" + thisString); //$NON-NLS-1$
		}
	}

	/**
	 * Populate combobox with fonts
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void populateWithFonts(final JComboBox thisCombo)
	{
		populateWithStrings(thisCombo, new String[] { "SansSerif", "Serif", "MonoSpaced", "Dialog", "DialogInput" }); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
		final Font[] theseFonts = GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts();
		for (final Font thisTheseFont : theseFonts)
		{
			thisCombo.addItem(thisTheseFont.getFontName());
		}
	}

	/**
	 * Populate combobox with terminators
	 * 
	 * @param thisCombo
	 *        combobox
	 */
	private void populateWithTerminators(final JComboBox thisCombo)
	{
		populateWithStrings(thisCombo, PropertyView.theTerminatorStrings);
	}

	/**
	 * Populate combobox with strings list
	 * 
	 * @param thisCombo
	 *        combobox
	 * @param theseStrings
	 *        list of strings
	 */
	private void populateWithStrings(final JComboBox thisCombo, final String[] theseStrings)
	{
		thisCombo.removeAllItems();
		thisCombo.setMaximumRowCount(theseStrings.length + 1);
		thisCombo.removeAllItems();
		thisCombo.addItem(PropertyView.theDefaultString);
		for (final String thisString : theseStrings)
		{
			thisCombo.addItem(thisString);
		}
	}

	// N O T I F I C A T I O N

	/*
	 * (non-Javadoc)
	 * @see javax.swing.event.CellEditorListener#editingCanceled(javax.swing.event.ChangeEvent)
	 */
	@Override
	public void editingCanceled(final ChangeEvent e)
	{
		System.out.println("Editor notification: Canceled"); //$NON-NLS-1$
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.event.CellEditorListener#editingStopped(javax.swing.event.ChangeEvent)
	 */
	@Override
	public void editingStopped(final ChangeEvent e)
	{
		System.out.println("Editor notification: Stopped"); //$NON-NLS-1$
	}

	// L I S T E N E R S

	/*
	 * (non-Javadoc)
	 * @see javax.swing.CellEditor#addCellEditorListener(javax.swing.event.CellEditorListener)
	 */
	@Override
	public void addCellEditorListener(final CellEditorListener thisListener)
	{
		for (final CellEditor thisSubEditor : this.theSubEditors)
		{
			thisSubEditor.addCellEditorListener(thisListener);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.CellEditor#removeCellEditorListener(javax.swing.event.CellEditorListener)
	 */
	@Override
	public void removeCellEditorListener(final CellEditorListener thisListener)
	{
		for (final CellEditor thisSubEditor : this.theSubEditors)
		{
			thisSubEditor.removeCellEditorListener(thisListener);
		}
	}
}