/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: 2.0.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */

package treebolic;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.MediaTracker;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.SwingWorker;

import treebolic.component.Progress;
import treebolic.component.Statusbar;
import treebolic.component.Toolbar;
import treebolic.control.Controller;
import treebolic.control.Controller.MatchMode;
import treebolic.control.Controller.MatchScope;
import treebolic.control.Mounter;
import treebolic.core.AbstractLayerOut;
import treebolic.core.Complex;
import treebolic.core.LayerOut;
import treebolic.core.Weigher;
import treebolic.model.IEdge;
import treebolic.model.INode;
import treebolic.model.IProvider;
import treebolic.model.IProviderContext;
import treebolic.model.Location;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.view.View;

/**
 * Treebolic's main panel. Call either init() to feed the data
 * 
 * @author Bernard Bou
 */
public class Widget extends JPanel implements IWidget, IProviderContext
{
	private static final long serialVersionUID = 020000L;

	// B E H A V I O U R

	/**
	 * Loading data on separate thread
	 */
	static private boolean debug = false;

	/**
	 * Loading data on separate thread
	 */
	static private final boolean threaded = true;

	/**
	 * XML is parsed by DOM parser
	 */
	static private boolean fromDom = true;

	/**
	 * Animate on start flag
	 */
	static private final boolean animateOnStart = true;

	// V E R S I O N

	/**
	 * Version
	 */
	private static final String theVersion = "provider";

	// C O N T E X T

	/**
	 * Context for operations (provides some glue to the context, application)
	 * 
	 * @see IContext
	 */
	private final IContext theContext;

	// M O D E L - V I E W - C O N T R O L L E R

	/**
	 * Model
	 */
	private Model theModel;

	/**
	 * View
	 */
	private View theView;

	/**
	 * Controller
	 */
	private final Controller theController;

	// D A T A . P R O V I D E R

	/**
	 * Provider
	 */
	private IProvider theProvider;

	// A G E N T S

	/**
	 * Weigher
	 */
	private final Weigher theWeigher;

	/**
	 * Provider
	 */
	private final AbstractLayerOut theLayerOut;

	// C O M P O N E N T S

	/**
	 * Toolbar
	 */
	private Toolbar theToolbar;

	/**
	 * Status bar
	 */
	private Statusbar theStatusbar;

	/**
	 * Progress
	 */
	private final Progress theProgress;

	// I M A G E . M A N A G E M E N T

	/**
	 * Media tracker
	 */
	private final MediaTracker theMediaTracker;

	/**
	 * Name-to-image map, used to void loading twice the same images
	 */
	private Hashtable<String, Image> theImages;

	/**
	 * Images base Url
	 */
	private URL theImagesBase;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 * 
	 * @param thisContext
	 *        context
	 */
	public Widget(final IContext thisContext)
	{
		this.theContext = thisContext;

		// media tracker
		this.theMediaTracker = new MediaTracker(this);

		// layout
		setLayout(new BorderLayout());

		// first components
		this.theProgress = new Progress();
		add(this.theProgress, BorderLayout.CENTER);

		// size
		setPreferredSize(new Dimension(600, 600));

		// controller
		this.theController = new Controller();
		this.theController.connect(this);

		// weigher
		this.theWeigher = new Weigher();

		// layer out
		this.theLayerOut = new LayerOut();
		this.theController.connect(this.theLayerOut);
	}

	// I N I T

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#init()
	 */
	@Override
	public void init()
	{
		// source
		String thisSource = getParameter("doc");
		if (thisSource == null || thisSource.isEmpty())
		{
			thisSource = getParameter("xml");
		}

		// provider
		final String thisProviderName = getParameter("provider");

		// init
		init(thisProviderName, thisSource);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#init(java.lang.String, java.lang.String)
	 */
	@Override
	public void init(final String thatProviderName, final String thisSource)
	{
		String thisProviderName = thatProviderName;
		if (thisProviderName == null)
		{
			thisProviderName = Widget.fromDom ? "treebolic.provider.xml.dom.Provider" : "treebolic.provider.xml.fast.Provider";
		}

		if ("true".equals(getParameter("debug")))
		{
			Widget.debug = true;
		}
		if (Widget.debug)
		{
			System.out.println("PROVIDER=" + thisProviderName);
		}

		// make provider
		this.theProvider = makeProvider(thisProviderName);
		if (this.theProvider == null)
		{
			putProgress("Cannot load provider <" + thisProviderName + ">", true);
			return;
		}

		// pass context reference to provider, so that it can access some services
		this.theProvider.connect(this);

		// init
		initProcess(thisSource);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#reinit(java.lang.String)
	 */
	@Override
	public void reinit(final String thisSource)
	{
		if (this.theProvider == null)
		{
			putProgress("Null provider " + this.theProvider, true);
			return;
		}

		// init
		initProcess(thisSource);
	}

	/**
	 * Init (typically called by embedding applet's init()). Data source and data provider have not yet been determined.
	 * 
	 * @param thisSource
	 *        source (anything the provider will make sense of)
	 */
	protected void initProcess(final String thisSource)
	{
		// further (possibly lengthy) loading is done on separate thread
		if (!Widget.threaded)
		{
			init(this.theProvider, thisSource);
		}
		else
		{
			final SwingWorker<Void, Void> thisWorker = new SwingWorker<Void, Void>()
			{
				@SuppressWarnings("synthetic-access")
				@Override
				protected Void doInBackground() throws Exception
				{
					try
					{
						init(Widget.this.theProvider, thisSource);
					}
					catch (final Throwable e)
					{
						System.err.println("Worker thread: " + e.toString());
						e.printStackTrace();
					}
					return null;
				}

				@Override
				protected void done()
				{
					// do nothing
				}
			};
			thisWorker.execute();
		}
	}

	/**
	 * Init from provider and source
	 * 
	 * @param thisProvider
	 *        data provider
	 * @param thisSource
	 *        data source
	 */
	protected void init(final IProvider thisProvider, final String thisSource)
	{
		if (Widget.debug)
		{
			System.out.println("SOURCE=" + thisSource);
		}
		putProgress("Loading " + thisSource, false);
		final Model thisModel = thisProvider.makeModel(thisSource);
		if (thisModel == null)
		{
			putProgress("Provider " + thisProvider.getClass().getCanonicalName() + " returned null model from source <" + thisSource + ">", true);
			return;
		}
		putProgress("Loaded " + thisSource, false);

		// load model
		init(thisModel);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#init(treebolic.model.Model)
	 */
	@Override
	public void init(final Model thisModel)
	{
		if (thisModel == null)
		{
			putProgress("Null model", true);
			putMessage("Aborting");
			return;
		}

		// model/view/controller
		this.theModel = thisModel;

		// view
		this.theView = new View();

		// connect view and controller to each other
		this.theController.connect(this.theView);
		this.theView.connect(this.theController);

		// connect both controller and view to model
		this.theView.connect(thisModel);
		this.theController.connect(thisModel);

		// connect view to layerout
		this.theView.connect(this.theLayerOut);

		// initiate image loading
		putProgress("Images loading...", false);
		loadImages();
		putProgress("Images loaded.", false);

		// enforce settings
		this.theLayerOut.apply(this.theModel.theSettings);
		this.theView.apply(this.theModel.theSettings);

		// weigh model
		putProgress("Weighing nodes ...", false);
		this.theWeigher.weigh(thisModel.theTree.getRoot());

		// lay out model
		putProgress("Laying out ...", false);
		this.theLayerOut.layout(thisModel.theTree.getRoot());

		// initial transform
		this.theView.applyInitialTransform();

		// remove all previous components (possibly progress or previous view and tools)
		removeAll();

		// status bar
		if (this.theModel.theSettings.theHasStatusbarFlag != null && this.theModel.theSettings.theHasStatusbarFlag)
		{
			this.theStatusbar = new Statusbar();
			add(this.theStatusbar, BorderLayout.SOUTH);
		}

		// toolbar
		if (this.theModel.theSettings.theHasToolbarFlag != null && this.theModel.theSettings.theHasToolbarFlag)
		{
			final boolean hasTooltip = this.theModel.theSettings.theHasToolTipFlag != null ? this.theModel.theSettings.theHasToolTipFlag : true;
			final boolean toolTipDisplaysContent = this.theModel.theSettings.theToolTipDisplaysContentFlag != null ? this.theModel.theSettings.theToolTipDisplaysContentFlag : true;
			final boolean edgesHyperbolic = this.theModel.theSettings.theHyperbolicEdgesFlag != null ? this.theModel.theSettings.theHyperbolicEdgesFlag : true;
			final boolean focusOnHover = this.theModel.theSettings.theFocusOnHoverFlag != null ? this.theModel.theSettings.theFocusOnHoverFlag : true;
			this.theToolbar = new Toolbar(this.theController, hasTooltip, toolTipDisplaysContent, edgesHyperbolic, focusOnHover);
			add(this.theToolbar, Toolbar.isVertical ? BorderLayout.EAST : BorderLayout.NORTH);
		}

		// show view
		add(this.theView, BorderLayout.CENTER);
		validate();

		// animate
		final INode thisFocus = getFocusNode();
		if (!Widget.animateOnStart)
		{
			// no animation
			this.theView.moveToCenter(thisFocus);
		}
		else
		{
			// animation to focus
			if (this.theModel.theSettings.theXMoveTo != null && this.theModel.theSettings.theXMoveTo < 1. || this.theModel.theSettings.theYMoveTo != null && this.theModel.theSettings.theYMoveTo < 1.)
			{
				// move required
				final Complex thisTo = new Complex(this.theModel.theSettings.theXMoveTo == null ? 0. : this.theModel.theSettings.theXMoveTo, this.theModel.theSettings.theYMoveTo == null ? 0. : this.theModel.theSettings.theYMoveTo);
				if (thisTo.abs2() > 1.)
				{
					thisTo.normalize().multiply(.9);
				}
				this.theView.animateTo(thisFocus, thisTo, false);
			}
			else
			{
				this.theView.animateToCenter(thisFocus, false);
			}
		}
	}

	// N A V I G A T I O N

	/**
	 * Mount a source at node
	 * 
	 * @param thisMountingNode
	 *        mounting node
	 * @param thisSource
	 *        mounted source
	 */
	public synchronized void mount(final INode thisMountingNode, final String thisSource)
	{
		putInformation("Mount", thisSource, false);

		if (this.theProvider == null)
		{
			putInformation("Mount", "Null provider", false);

			// make provider
			this.theProvider = makeProvider("treebolic.provider.xml.dom.Provider");
			if (this.theProvider == null)
			{
				putInformation("Mount", "Cannot load provider treebolic.provider.xml.dom.Provider", true);
				return;
			}

			// pass context reference to provider, so that it can access some services
			this.theProvider.connect(this);
		}

		// make model
		final Tree thisTree = this.theProvider.makeTree(thisSource, false);
		if (thisTree == null)
		{
			putInformation("Mount", "failed to get " + thisSource, false);
			return;
		}

		// extract subroot
		final INode thisMountedRoot = thisTree.getRoot();

		// images
		loadImages(thisMountedRoot);

		// ensure edge list is non null
		if (this.theModel.theTree.getEdges() == null)
		{
			this.theModel.theTree.setEdges(new ArrayList<IEdge>());
		}

		// graft nodes
		if (!Mounter.graft(thisMountingNode, thisMountedRoot, this.theModel.theTree.getEdges(), thisTree.getEdges()))
		{
			putInformation("Mount", "failed mounting", false);
			return;
		}

		// weigh
		this.theWeigher.weigh(thisMountedRoot);
		thisMountingNode.setChildrenWeight(thisMountedRoot.getChildrenWeight());
		thisMountingNode.setMinWeight(thisMountedRoot.getMinWeight());

		// compute locations : layout
		final MountPoint.Mounting thisMountingPoint = (MountPoint.Mounting) thisMountingNode.getMountPoint();
		this.theLayerOut.layout(thisMountedRoot, thisMountingNode.getLocation().hyper.center0, thisMountingPoint.theHalfWedge, thisMountingPoint.theOrientation);

		// notify view
		this.theView.mount(thisMountedRoot);
	}

	/**
	 * Unmount at mountpoint node
	 * 
	 * @param thisMountedNode
	 *        mounted node
	 */
	public synchronized void umount(final INode thisMountedNode)
	{
		putInformation("UnMount", "", false);

		// model
		final INode thisMountingNode = Mounter.prune(thisMountedNode, this.theModel.theTree.getEdges());
		if (thisMountingNode == null)
		{
			putInformation("Unmount", "failed unmounting", false);
			return;
		}

		// compute locations : copy
		final Location thisMountedNodeLocation = thisMountedNode.getLocation();
		final Location thisMountingNodeLocation = thisMountingNode.getLocation();
		thisMountingNodeLocation.hyper.clone(thisMountedNodeLocation.hyper);

		// notify view
		this.theView.umount(thisMountingNode);
	}

	// A C C E S S

	/**
	 * Get version
	 * 
	 * @return version
	 */
	@Override
	public String getVersion()
	{
		return Widget.theVersion;
	}

	/**
	 * Get context
	 * 
	 * @return context
	 */
	public IContext getContext()
	{
		return this.theContext;
	}

	// P R O V I D E R - F A C T O R Y

	/**
	 * Make provider
	 * 
	 * @param thisProviderName
	 *        provider name
	 * @return provider
	 */
	private IProvider makeProvider(final String thisProviderName)
	{
		try
		{
			final Class<?> thisClass = Class.forName(thisProviderName);
			final Class<?>[] theseArgsClass = new Class[] {};
			final Object[] theseArgs = new Object[] {};

			final Constructor<?> thisConstructor = thisClass.getConstructor(theseArgsClass);
			final Object thisInstance = thisConstructor.newInstance(theseArgs);
			return (IProvider) thisInstance;
		}
		catch (final ClassNotFoundException e)
		{
			System.err.println("Provider factory: " + e.toString());
		}
		catch (final NoSuchMethodException e)
		{
			System.err.println("Provider factory: " + e.toString());
		}
		catch (final IllegalAccessException e)
		{
			System.err.println("Provider factory: " + e.toString());
		}
		catch (final InstantiationException e)
		{
			System.err.println("Provider factory: " + e.toString());
		}
		catch (final IllegalArgumentException e)
		{
			System.err.println("Provider factory: " + e.toString());
		}
		catch (final InvocationTargetException e)
		{
			System.err.println("Provider factory: " + e.toString());
		}
		return null;
	}

	// P R O V I D E R . C O N T E X T

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProviderContext#getParameter(java.lang.String)
	 */
	@Override
	public String getParameter(final String thisName)
	{
		// delegate to context
		return this.theContext.getParameter(thisName);
	}

	// U R L . F A C T O R Y

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProviderContext#makeURL(java.lang.String)
	 */
	@Override
	public URL makeURL(final String thisSource)
	{
		if (thisSource == null)
		{
			if (Widget.debug)
			{
				System.out.println("MAKEURL= null (null source)");
			}
			return null;
		}

		// try to consider it well-formed full-fledged url
		try
		{
			final URL thisUrl = new URL(thisSource);
			if (Widget.debug)
			{
				System.out.println("URL=" + new URL(thisSource).toString());
			}
			return thisUrl;
		}
		catch (final MalformedURLException e)
		{
			// do nothing
		}

		// default to source relative to a base
		try
		{
			final URL thisBaseUrl = getBase();
			if (Widget.debug)
			{
				System.out.println("BASE URL=" + thisBaseUrl.toString());
			}
			final URL thisUrl = new URL(thisBaseUrl, thisSource);
			if (Widget.debug)
			{
				System.out.println("URL=" + thisUrl.toString());
			}
			return thisUrl;
		}
		catch (final MalformedURLException e)
		{
			// do nothing
		}
		if (Widget.debug)
		{
			System.out.println("URL= null (fail)");
		}
		return null;
	}

	/**
	 * Get base url
	 * 
	 * @return base url
	 */
	private URL getBase()
	{
		return this.theContext.getBase();
	}

	/**
	 * Get image Url
	 * 
	 * @param thisImageSource
	 *        image source
	 * @return image url
	 */
	private URL makeImageURL(final String thisImageSource)
	{
		try
		{
			return new URL(this.theImagesBase, thisImageSource);
		}
		catch (final MalformedURLException thisException)
		{
			// do nothing
		}
		return null;
	}

	// I M A G E . L O A D E R

	private void loadImages()
	{
		this.theImages = new Hashtable<String, Image>();
		this.theImagesBase = this.theContext.getImagesBase();
		if (Widget.debug)
		{
			System.out.println("IMAGEBASE URL=" + this.theImagesBase.toString());
		}
		loadImages(this.theModel.theTree.getRoot());
		loadImages(this.theModel.theTree.getEdges());
		loadTopImages(this.theModel.theSettings);
		try
		{
			this.theMediaTracker.waitForAll();
		}
		catch (final InterruptedException thisException)
		{
			System.err.println("Loading images interrupted: " + thisException.toString());
		}
	}

	/**
	 * Load node and node's children's images
	 * 
	 * @param thisNode
	 *        starting node
	 */
	private void loadImages(final INode thisNode)
	{
		if (thisNode == null)
			return;

		// node image
		String thisSource = thisNode.getImageFile();
		if (thisSource != null)
		{
			// cache lookup
			Image thisImage = this.theImages.get(thisSource);
			if (thisImage == null)
			{
				// cache miss
				thisImage = loadImage(thisSource);
			}

			// store image in the node
			thisNode.setImage(thisImage);
		}

		// edge image
		thisSource = thisNode.getEdgeImageFile();
		if (thisSource != null)
		{
			// cache lookup
			Image thisImage = this.theImages.get(thisSource);
			if (thisImage == null)
			{
				// cache miss
				thisImage = loadImage(thisSource);
			}
			thisNode.setEdgeImage(thisImage);
		}

		// recurse on mounting node obfuscated by mounted node)
		final MountPoint thisMountPoint = thisNode.getMountPoint();
		if (thisMountPoint != null && thisMountPoint instanceof MountPoint.Mounted)
		{
			final MountPoint.Mounted thisMountedPoint = (MountPoint.Mounted) thisMountPoint;
			final INode theMountingNode = thisMountedPoint.theMountingNode;
			loadImages(theMountingNode);
		}

		// recurse children
		final List<INode> theseChildren = thisNode.getChildren();
		if (theseChildren != null)
		{
			for (final INode thisChild : thisNode.getChildren())
			{
				loadImages(thisChild);
			}
		}
	}

	/**
	 * Load edges' images
	 * 
	 * @param thisEdgeList
	 *        edge list
	 */
	private void loadImages(final List<IEdge> thisEdgeList)
	{
		if (thisEdgeList != null)
		{
			for (final IEdge thisEdge : thisEdgeList)
			{
				loadImage(thisEdge);
			}
		}
	}

	/**
	 * Load edge image
	 * 
	 * @param thisEdge
	 *        edge
	 */
	private void loadImage(final IEdge thisEdge)
	{
		if (thisEdge == null)
			return;
		final String thisSource = thisEdge.getImageFile();
		if (thisSource != null)
		{
			// cache lookup
			Image thisImage = this.theImages.get(thisSource);
			if (thisImage == null)
			{
				// cache miss
				thisImage = loadImage(thisSource);
			}
			thisEdge.setImage(thisImage);
		}
	}

	/**
	 * Load top images
	 * 
	 * @param theseSettings
	 *        settings
	 */
	private void loadTopImages(final Settings theseSettings)
	{
		final Image thisBackgroundImage = loadImage(theseSettings.theBackgroundImage);
		final Image thisDefaultNodeImage = loadImage(theseSettings.theDefaultNodeImage);
		final Image thisDefaultTreeEdgeImage = loadImage(theseSettings.theDefaultTreeEdgeImage);
		final Image thisDefaultEdgeImage = loadImage(theseSettings.theDefaultEdgeImage);
		this.theView.setImages(thisBackgroundImage, thisDefaultNodeImage, thisDefaultTreeEdgeImage, thisDefaultEdgeImage);
	}

	/**
	 * Load image given its source
	 * 
	 * @param thisSource
	 *        source
	 * @return image
	 */
	private Image loadImage(final String thisSource)
	{
		if (thisSource == null)
			return null;

		final URL thisUrl = makeImageURL(thisSource);
		if (thisUrl != null)
		{
			if (Widget.debug)
			{
				System.out.println("IMAGE=" + thisUrl);
			}

			// delegate image loading to context
			final Image thisImage = this.theContext.getImage(thisUrl);
			if (thisImage != null)
			{
				// cache image
				this.theImages.put(thisSource, thisImage);
				// track it
				this.theMediaTracker.addImage(thisImage, thisSource.hashCode());
				return thisImage;
			}
		}
		return null;
	}

	// F O C U S . N O D E

	/**
	 * Get node to initially receive focus
	 * 
	 * @return node
	 */
	private INode getFocusNode()
	{
		// focus node
		String thisFocusNodeId = null;
		if (this.theModel.theSettings.theFocus != null)
		{
			thisFocusNodeId = this.theModel.theSettings.theFocus;
		}

		// anchor of document base
		final URL thisBase = this.theContext.getHtmlDocumentBase();
		if (thisBase != null)
		{
			thisFocusNodeId = thisBase.getRef();
		}

		// animate
		INode thisFocus = null;
		if (thisFocusNodeId == null)
		{
			thisFocus = this.theModel.theTree.getRoot();
		}
		else
		{
			thisFocus = this.theModel.theTree.findNodeById(this.theModel.theTree.getRoot(), thisFocusNodeId);

			// default to root if not found
			if (thisFocus == null)
			{
				thisFocus = this.theModel.theTree.getRoot();
			}
		}
		return thisFocus;
	}

	// P R O G R E S S

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProviderContext#putProgress(java.lang.String)
	 */
	@Override
	public void putProgress(final String thisString, final boolean fail)
	{
		this.theProgress.put(thisString, fail);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProviderContext#putMessage(java.lang.String)
	 */
	@Override
	public void putMessage(final String thisString)
	{
		this.theContext.showStatus(thisString);
	}

	// S T A T U S

	/**
	 * Get target
	 * 
	 * @return target
	 */
	public String getTarget()
	{
		if (this.theStatusbar != null)
			return this.theStatusbar.get();
		return null;
	}

	/**
	 * Put information
	 * 
	 * @param thisHeader
	 *        header
	 * @param thisContent
	 *        content
	 */
	public void putInformation(final String thisHeader, final String thisContent, final boolean isLink)
	{
		if (this.theStatusbar == null)
			return;
		this.theStatusbar.put(thisHeader, thisContent, isLink ? Statusbar.PutType.LINK : Statusbar.PutType.INFO);
	}

	// JavaScript

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#focus(java.lang.String)
	 */
	@Override
	public void focus(final String thisNodeId)
	{
		this.theController.focus(thisNodeId);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#match(java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
	public String match(final String thisTargetString, final String thisScopeString, final String thisModeString)
	{
		MatchScope thisScope = MatchScope.LABEL;
		if (thisScopeString != null)
		{
			if (thisScopeString.equals("label"))
			{
				thisScope = MatchScope.LABEL;
			}
			else if (thisScopeString.equals("content"))
			{
				thisScope = MatchScope.CONTENT;
			}
			else if (thisScopeString.equals("link"))
			{
				thisScope = MatchScope.LINK;
			}
			else if (thisScopeString.equals("id"))
			{
				thisScope = MatchScope.ID;
			}
		}

		MatchMode thisMode = MatchMode.EQUALS;
		if (thisModeString != null)
		{
			if (thisModeString.equals("equals"))
			{
				thisMode = MatchMode.EQUALS;
			}
			else if (thisModeString.equals("startswith"))
			{
				thisMode = MatchMode.STARTSWITH;
			}
			else if (thisModeString.equals("includes"))
			{
				thisMode = MatchMode.INCLUDES;
			}
		}
		return this.theController.match(thisTargetString, thisScope, thisMode);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.IWidget#link(java.lang.String, java.lang.String)
	 */
	@Override
	public void link(final String thisUrlString, final String thisUrlTarget)
	{
		this.theController.linkTo(thisUrlString, thisUrlTarget);
	}
}
