/**
 * Title : Treebolic
 * Description: Treebolic
 * Version: 2.x
 * Copyright:	(c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.commons;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * Class searcher
 * 
 * @author Walter Angerer 'greenhorn'
 * @author Bernard Bou
 */
public class Searcher
{
	/**
	 * Find classes matching pattern in classpath
	 * 
	 * @param thatPattern
	 *        pattern
	 * @return result set
	 */
	static public Set<String> findClasses(final String thatPattern)
	{
		String thisPattern = thatPattern;
		final Set<String> thisList = new TreeSet<String>();
		final char thisSeparator = File.separatorChar;
		thisPattern = thisPattern.replaceAll("\\\\.", thisSeparator == '/' ? "/" : "\\\\\\\\"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		for (final String thisFileName : Searcher.findFiles(thisPattern + "\\.class$").keySet())
		{
			thisList.add(thisFileName.replace(thisSeparator, '.').substring(0, thisFileName.length() - 6));
		}
		return thisList;
	}

	/**
	 * Find files matching pattern in classpath
	 * 
	 * @param thisPattern
	 *        pattern
	 * @return result set
	 */
	static public Map<String, String> findFiles(final String thisPattern)
	{
		final Map<String, String> thisList = new TreeMap<String, String>();
		final String thisClassPath = System.getProperty("java.class.path"); //$NON-NLS-1$
		final String[] thesePathElements = thisClassPath.split(File.pathSeparator);
		for (final String thisPathElement : thesePathElements)
		{
			try
			{
				final File thisFile = new File(thisPathElement);
				if (thisFile.isDirectory())
				{
					thisList.putAll(Searcher.findInDirectory(thisFile, thisPattern, thisPathElement));
				}
				else
				{
					thisList.putAll(Searcher.findInFile(thisFile, thisPattern));
				}
			}
			catch (final IOException e)
			{
				System.err.println("Find files: " + e.toString()); //$NON-NLS-1$
			}
		}
		return thisList;
	}

	/**
	 * Find in JAR file
	 * 
	 * @param thisFile
	 *        JAR file
	 * @param thisPattern
	 *        pattern
	 * @return file with matching file path
	 * @throws IOException
	 */
	static private Map<String, String> findInFile(final File thisFile, final String thisPattern) throws IOException
	{
		final Map<String, String> thisMap = new TreeMap<String, String>();
		if (thisFile.canRead() && thisFile.getAbsolutePath().endsWith(".jar")) //$NON-NLS-1$
		{
			final JarFile thisJar = new JarFile(thisFile);
			for (final Enumeration<JarEntry> theseEntries = thisJar.entries(); theseEntries.hasMoreElements();)
			{
				final JarEntry thisEntry = theseEntries.nextElement();
				if (thisEntry.getName().matches(thisPattern))
				{
					thisMap.put(thisEntry.getName(), thisJar.getName());
				}
			}
		}
		return thisMap;
	}

	/**
	 * Find in directory
	 * 
	 * @param thisDirectory
	 *        directory
	 * @param thisPattern
	 *        pattern
	 * @return file with matching file path
	 * @throws IOException
	 */
	static private Map<String, String> findInDirectory(final File thisDirectory, final String thisPattern, final String thisPathElement) throws IOException
	{
		final Map<String, String> thisMap = new TreeMap<String, String>();
		for (final File thisDirectoryEntry : thisDirectory.listFiles())
		{
			if (thisDirectoryEntry.getAbsolutePath().matches(thisPattern))
			{
				final String thisFileName = thisDirectoryEntry.getAbsolutePath().substring(thisPathElement.length() + 1);
				thisMap.put(thisFileName, thisPathElement);
			}
			else if (thisDirectoryEntry.isDirectory())
			{
				thisMap.putAll(Searcher.findInDirectory(thisDirectoryEntry, thisPattern, thisPathElement));
			}
			else
			{
				thisMap.putAll(Searcher.findInFile(thisDirectoryEntry, thisPattern));
			}
		}
		return thisMap;
	}

	/**
	 * Find files matching pattern in classpath
	 * 
	 * @param thisPattern
	 *        pattern
	 * @return list of urls
	 */
	static public List<String> findFileUrls(final String thisPattern)
	{
		return Searcher.toUrls(Searcher.findFiles(thisPattern));
	}

	/**
	 * Convert file map to list of urls
	 * 
	 * @param thisMap
	 *        path to container map
	 * @return list of urls
	 */
	static public List<String> toUrls(final Map<String, String> thisMap)
	{
		final List<String> theseUrls = new ArrayList<String>();
		for (final Map.Entry<String, String> thisEntry : thisMap.entrySet())
		{
			final String thisContainer = thisEntry.getValue();
			final String thisPath = thisEntry.getKey();
			String thisUrlString = null;
			if (thisContainer.endsWith(".jar"))
			{
				thisUrlString = "jar:file:" + thisContainer + "!/" + thisPath; //$NON-NLS-1$ //$NON-NLS-2$
			}
			else
			{
				thisUrlString = "file:" + thisContainer + '/' + thisPath; //$NON-NLS-1$
			}
			theseUrls.add(thisUrlString);
		}
		return theseUrls;
	}

	/**
	 * Main
	 * 
	 * @param args
	 *        args
	 */
	public static void main(final String[] args)
	{
		for (final String thisClassName : Searcher.findClasses(".*Provider"))
		{
			System.out.println("PROVIDER CLASS: " + thisClassName); //$NON-NLS-1$
		}

		for (final Map.Entry<String, String> thisEntry : Searcher.findFiles(".*xsl$").entrySet())
		{
			System.out.println("XSL FILE: " + thisEntry.getKey() + " @ " + thisEntry.getValue()); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}
}
