/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.control;

import java.awt.Point;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;

import javax.swing.JPopupMenu;

import treebolic.Widget;
import treebolic.component.PopupMenu;
import treebolic.core.AbstractLayerOut;
import treebolic.core.Complex;
import treebolic.model.INode;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.view.View;

/**
 * Controller
 * 
 * @author Bernard Bou
 */
public class Controller extends Commander
{
	// connected

	/**
	 * Connected widget
	 */
	private Widget theWidget;

	/**
	 * Connected model
	 */
	private Model theModel;

	/**
	 * Connected view
	 */
	private View theView;

	/**
	 * Connected layout agent
	 */
	private AbstractLayerOut theLayerOut;

	// behaviour

	/**
	 * Verbose status flag (use for debug purposes)
	 */
	static private final boolean verbose = false;

	// action

	/**
	 * Event types
	 */
	static public enum Event
	{
		SELECT, HOVER, DRAG, LEAVEDRAG, MOVE, ROTATE, FOCUS, MOUNT, LINK, POPUP
	}

	/**
	 * Action types
	 */
	public static enum Action
	{
		GOTO, SEARCH, FOCUS, LINK, MOUNT
	};

	/**
	 * Match scope
	 */
	public static enum MatchScope
	{
		LABEL, CONTENT, LINK, ID
	};

	/**
	 * Match mode
	 */
	public static enum MatchMode
	{
		EQUALS, STARTSWITH, INCLUDES
	};

	/**
	 * Match scopes
	 */
	static private final String[] theMatchScopeString = { "label", "content", "link", "id" };

	/**
	 * Match modes
	 */
	static private final String[] theMatchModeString = { "equals", "starts with", "includes" };

	// C O N T R O L L E R

	/**
	 * Constructor
	 */
	public Controller()
	{
		super();
		this.theWidget = null;
		this.theModel = null;
		this.theView = null;
		this.theLayerOut = null;
	}

	// C O N N E C T

	/**
	 * Connect with widget
	 * 
	 * @param thisWidget
	 *        widget
	 */
	public void connect(final Widget thisWidget)
	{
		this.theWidget = thisWidget;
	}

	/**
	 * Connect
	 * 
	 * @param thisModel
	 *        model
	 */
	public void connect(final Model thisModel)
	{
		this.theModel = thisModel;
	}

	/**
	 * Connect with view
	 * 
	 * @param thisView
	 *        view
	 */
	public void connect(final View thisView)
	{
		this.theView = thisView;
	}

	/**
	 * Connect with layout agent
	 * 
	 * @param thisLayerOut
	 *        layerout
	 */
	public void connect(final AbstractLayerOut thisLayerOut)
	{
		this.theLayerOut = thisLayerOut;
	}

	// R E F E R E N C E . F O R . C O M M A N D E R

	/*
	 * (non-Javadoc)
	 * @see treebolic.control.Commander#getModel()
	 */
	@Override
	protected Model getModel()
	{
		return this.theModel;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.control.Commander#getView()
	 */
	@Override
	protected View getView()
	{
		return this.theView;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.control.Commander#getLayerOut()
	 */
	@Override
	protected AbstractLayerOut getLayerOut()
	{
		return this.theLayerOut;
	}

	// E V E N T S

	/**
	 * Handle events
	 * 
	 * @param thisEventType
	 *        event type
	 * @param thisParameter
	 *        event-specific object
	 */
	@SuppressWarnings("unchecked")
	public void handle(final Event thisEventType, final Object thisParameter)
	{
		switch (thisEventType)
		{
		case MOVE:
		{
			// translate as per vector(start,end)
			final Pair<Complex> thisVector = (Pair<Complex>) thisParameter;
			this.theView.composeTranslate(thisVector.theFirst, thisVector.theSecond);
			break;
		}

		case ROTATE:
		{
			// rotate as per vector(start,end)
			final Pair<Complex> thisVector = (Pair<Complex>) thisParameter;
			this.theView.composeRotate(thisVector.theFirst, thisVector.theSecond);
			break;
		}

		case HOVER:
		{
			final INode thisNode = (INode) thisParameter;
			final String thisLink = thisNode.getLink();

			// tooltip
			this.theView.setToolTip(thisNode);

			// cursor
			this.theView.setHoverCursor(thisLink != null && !thisLink.isEmpty());

			// status
			putNode(thisNode);
			break;
		}

		case DRAG:
		{
			this.theView.enterDrag();
			break;
		}

		case LEAVEDRAG:
		{
			this.theView.leaveDrag();
			break;
		}

		case SELECT:
		{
			break;
		}

		case FOCUS:
		{
			final INode thisNode = (INode) thisParameter;
			this.theView.animateToCenter(thisNode, false);
			break;
		}

		case MOUNT:
		{
			final INode thisNode = (INode) thisParameter;
			final MountPoint thisMountPoint = thisNode.getMountPoint();
			if (thisMountPoint != null)
			{
				// mount/umount
				if (thisMountPoint instanceof MountPoint.Mounted)
				{
					this.theWidget.umount(thisNode);
				}
				else
				{
					final MountPoint.Mounting thisMointingPoint = (MountPoint.Mounting) thisMountPoint;
					this.theWidget.mount(thisNode, thisMointingPoint.theURL);
				}
			}
			break;
		}

		case LINK:
		{
			final INode thisNode = (INode) thisParameter;
			final String thisLink = thisNode.getLink();
			final String thisTarget = thisNode.getTarget();
			if (thisLink != null)
			{
				linkTo(thisLink, thisTarget);
			}
			break;
		}

		case POPUP:
		{
			final Pair<Object> thisPair = (Pair<Object>) thisParameter;
			final Point thisPoint = (Point) thisPair.theFirst;
			final INode thisNode = (INode) thisPair.theSecond;
			popup(thisPoint.x, thisPoint.y, thisNode);
			break;
		}

		default:
			System.err.println("Unhandled event: " + thisEventType.toString());
		}
	}

	// D I S P A T C H

	/**
	 * Dispatch action
	 * 
	 * @param thisAction
	 *        action
	 * @param thatLink
	 *        url
	 * @param thisLinkTarget
	 *        url link target
	 * @param thisMatchTarget
	 *        match target
	 * @param thisMatchScope
	 *        match scope
	 * @param thisMatchMode
	 *        match mode
	 * @param thisNode
	 *        node
	 */
	public void dispatch(final Action thisAction, final String thatLink, final String thisLinkTarget, final String thisMatchTarget, final MatchScope thisMatchScope, final MatchMode thisMatchMode, final INode thisNode)
	{
		String thisLink = thatLink;
		switch (thisAction)
		{
		case GOTO:
			final String thisString = this.theWidget.getTarget();
			String thisExpandedUrl = null;
			if (thisLink != null)
			{
				thisLink = decode(thisLink);
				thisExpandedUrl = PopupMenu.expandPercent(thisLink, thisString, thisNode);
			}
			linkTo(thisExpandedUrl, thisLinkTarget);
			break;

		case LINK:
			handle(Controller.Event.LINK, thisNode);
			break;

		case MOUNT:
			handle(Controller.Event.MOUNT, thisNode);
			break;

		case SEARCH:

			// target
			final String thisEdit = this.theWidget.getTarget();
			String thisTarget;
			if (thisMatchTarget == null || thisMatchTarget.isEmpty())
			{
				thisTarget = thisEdit;
			}
			else
			{
				thisTarget = PopupMenu.expandPercent(thisMatchTarget, thisEdit, thisNode);
			}

			// search
			if (thisTarget != null)
			{
				final String thisMessage = "search for node whose " + Controller.theMatchScopeString[thisMatchScope.ordinal()] + " " + Controller.theMatchModeString[thisMatchMode.ordinal()] + " <" + thisTarget + "> from " + thisNode.getLabel() + "\n";
				this.theWidget.putInformation(null, "searching " + thisMessage, false);
				final INode thisResult = match(thisTarget, thisMatchScope, thisMatchMode, thisNode);
				if (thisResult != null)
				{
					this.theWidget.putInformation(thisTarget, thisMessage + thisResult, false);
					this.theView.animateToCenter(thisResult, false);
				}
				else
				{
					this.theWidget.putInformation(thisTarget, thisMessage + "null", false);
				}
			}
			break;

		case FOCUS:
			this.theView.animateToCenter(thisNode, false);
			break;

		default:
			System.err.println("Unsupported dispatch action=" + thisAction + " link=" + thisLink + " context=" + thisNode);
		}
	}

	// D I S P L A Y

	/**
	 * Display node
	 * 
	 * @param thisNode
	 *        node
	 */
	private void putNode(final INode thisNode)
	{
		// data
		final String thisLabel = thisNode.getLabel();
		String thisContent = thisNode.getContent();
		if (thisContent == null)
		{
			thisContent = "";
		}
		final String thisLink = thisNode.getLink();
		if (thisLink != null && !thisLink.isEmpty())
		{
			thisContent += " &lt;" + decode(thisLink) + "&gt;";
		}
		final MountPoint thisMountPoint = thisNode.getMountPoint();
		if (thisMountPoint != null && thisMountPoint instanceof MountPoint.Mounting)
		{
			final MountPoint.Mounting thisMountingPoint = (MountPoint.Mounting) thisMountPoint;
			thisContent += " [mountpoint=" + thisMountingPoint.theURL + "]";
		}
		if (Controller.verbose)
		{
			thisContent += " weight=" + thisNode.getWeight();
		}
		this.theWidget.putInformation(thisLabel, thisContent, false);
	}

	// P O P U P

	public void popup(final int x, final int y, final INode thisNode)
	{
		final JPopupMenu thisMenu = PopupMenu.makePopup(this, this.theWidget.getTarget(), thisNode, this.theModel.theSettings);
		thisMenu.show(this.theView, x, y);
	}

	// M A T C H . N O D E

	/**
	 * Match node against for string
	 * 
	 * @param thisString
	 *        string to match
	 * @param thisScope
	 *        scope
	 * @param thisMode
	 *        mode
	 * @param thisNode
	 *        node
	 * @return node
	 */
	private INode match(final String thisString, final MatchScope thisScope, final MatchMode thisMode, final INode thisNode)
	{
		// match
		String thisTarget = null;
		switch (thisScope)
		{
		case CONTENT:
			thisTarget = thisNode.getContent();
			break;

		case LINK:
			thisTarget = thisNode.getLink();
			break;

		case ID:
			thisTarget = thisNode.getId();
			break;

		case LABEL:
		default:
			thisTarget = thisNode.getLabel();
			break;
		}

		// try to match this node
		if (thisTarget != null)
		{
			switch (thisMode)
			{
			case EQUALS:
				if (thisTarget.equals(thisString))
					return thisNode;
				break;

			case INCLUDES:
				if (thisTarget.indexOf(thisString) != -1)
					return thisNode;
				break;

			case STARTSWITH:
			default:
				if (thisTarget.startsWith(thisString))
					return thisNode;
				break;
			}
		}

		// try to match match this node's children
		for (final INode thisChild : thisNode.getChildren())
		{
			final INode thisResult = match(thisString, thisScope, thisMode, thisChild);
			if (thisResult != null)
				return thisResult;
		}
		return null;
	}

	/**
	 * Match node against string (wrapper)
	 * 
	 * @param thisString
	 *        string to match
	 * @param thisScope
	 *        scope (LABEL, CONTENT, LINK, ID)
	 * @param thisMode
	 *        mode (EQUALS, STARTSWITH, INCLUDES)
	 * @return node id
	 */
	public String match(final String thisString, final MatchScope thisScope, final MatchMode thisMode)
	{
		final INode thisNode = match(thisString, thisScope, thisMode, this.theModel.theTree.getRoot());
		if (thisNode != null)
			return thisNode.getId();
		return null;
	}

	// S E A R C H . F O R . N O D E . F R O M . L O C A T I O N

	/**
	 * Find node
	 * 
	 * @param vx
	 *        at view x position
	 * @param vy
	 *        at view y position
	 * @return found node or null
	 */
	public INode findNode(final int vx, final int vy)
	{
		final Complex thisEuclidianLocation = this.theView.viewToUnitCircle(vx, vy);
		return this.theModel.theTree.findNodeAt(this.theModel.theTree.getRoot(), thisEuclidianLocation);
	}

	// F O C U S

	/**
	 * Focus node
	 * 
	 * @param thisNodeId
	 *        node id to get focus
	 */
	public void focus(final String thisNodeId)
	{
		final INode thisNode = this.theModel.theTree.findNodeById(this.theModel.theTree.getRoot(), thisNodeId);
		if (thisNode != null)
		{
			this.theView.animateToCenter(thisNode, false);
		}
	}

	// N A V I G A T I O N

	/**
	 * Decode encoded URL (for display)
	 * 
	 * @param thisString
	 *        encode URL string
	 * @return decoded URL string
	 */
	private String decode(final String thisString)
	{
		try
		{
			return URLDecoder.decode(thisString, "UTF8");
		}
		catch (final UnsupportedEncodingException e)
		{
			System.err.println("Can't decode " + thisString + " - " + e);
		}
		return thisString;
	}

	/**
	 * Follow hypertext link
	 * 
	 * @param thisHref
	 *        url string
	 * @param thisTarget
	 *        target frame
	 */
	public void linkTo(final String thisHref, final String thisTarget)
	{
		if (thisHref == null)
			return;

		// reference hook : find node with identifier
		if (thisHref.startsWith("#"))
		{
			final String thisBookmark = thisHref.substring(1);
			final INode thisFocus = this.theModel.theTree.findNodeById(this.theModel.theTree.getRoot(), thisBookmark);
			if (thisFocus != null)
			{
				this.theView.animateToCenter(thisFocus, false);
				return;
			}
		}

		final String thisDecodedLink = decode(thisHref);

		// status
		this.theWidget.putInformation("LINK", "Link to " + thisDecodedLink + " ," + thisTarget, true);
		this.theWidget.getContext().showStatus("Link to " + thisDecodedLink);

		// jump link
		if (!this.theWidget.getContext().linkTo(thisDecodedLink, thisTarget))
		{
			this.theWidget.reinit(thisDecodedLink);
		}
	}

	// S P A C E . C O N V E R S I O N

	/**
	 * Convert view space to unit circle
	 * 
	 * @param thisPoint
	 *        view space coordinate
	 * @return unit circle coordinate
	 */
	public Complex viewToUnitCircle(final Point thisPoint)
	{
		return this.theView.viewToUnitCircle(thisPoint.x, thisPoint.y);
	}
}
