/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */

package treebolic.control;

import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import treebolic.core.Complex;
import treebolic.core.EuclidianLocation;
import treebolic.model.INode;

/**
 * Mouse adapter
 * 
 * @author Bernard Bou
 */
public class MouseAdapter implements MouseListener, MouseMotionListener
{
	// D A T A

	/**
	 * Controller
	 */
	private final Controller theController;

	// D R A G A N D D R O P

	/**
	 * Drag mode enum
	 */
	static private enum DragMode
	{
		TRANSLATE, ROTATE
	}

	/**
	 * Drag mode
	 */
	private DragMode theDragMode;

	/**
	 * Whether mouse was dragged
	 */
	private boolean wasDragged = false;

	/**
	 * Whether mouse was moved
	 */
	private boolean wasMoved = false;

	// popup behaviour

	/**
	 * Whether it has popup
	 */
	public boolean hasPopUp = true;

	// translation
	/**
	 * Drag starting point
	 */
	private final Complex theDragStart = new Complex();

	/**
	 * Drag end point
	 */
	private Complex theDragEnd = new Complex();

	// selection
	/**
	 * Hot node
	 */
	private INode theHotNode = null;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 * 
	 * @param thisController
	 *        controller
	 */
	public MouseAdapter(final Controller thisController)
	{
		this.theController = thisController;
	}

	// M O U S E L I S T E N E R

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseClicked(final MouseEvent thisEvent)
	{
		final int vx = thisEvent.getX();
		final int vy = thisEvent.getY();
		final INode thisNode = this.theController.findNode(vx, vy);
		if (thisNode != null)
		{
			// right click
			if (this.hasPopUp && (thisEvent.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK)
			{
				this.theController.handle(Controller.Event.POPUP, new Pair<Object>(new Point(thisEvent.getX(), thisEvent.getY()), thisNode));
			}
			else if (thisEvent.isControlDown())
			{
				this.theController.handle(Controller.Event.MOUNT, thisNode);
			}
			else if (thisEvent.isShiftDown())
			{
				this.theController.handle(Controller.Event.LINK, thisNode);
			}
			else if (!thisEvent.isControlDown() && !thisEvent.isShiftDown())
			{
				this.theController.handle(Controller.Event.FOCUS, thisNode);
			}
		}
		thisEvent.consume();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	@Override
	public void mousePressed(final MouseEvent thisEvent)
	{
		final Complex thisDragStart = this.theController.viewToUnitCircle(new Point(thisEvent.getX(), thisEvent.getY()));
		this.theDragMode = thisEvent.isShiftDown() ? DragMode.ROTATE : DragMode.TRANSLATE;
		this.theDragStart.set(thisDragStart);
		this.theDragEnd.set(this.theDragStart);
		thisEvent.consume();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseReleased(final MouseEvent thisEvent)
	{
		// drag
		if (this.wasDragged)
		{
			this.wasDragged = false;
			this.theController.handle(Controller.Event.LEAVEDRAG, null);
		}
		else
		{
			// selection
			final int vx = thisEvent.getX();
			final int vy = thisEvent.getY();
			final INode thisNode = this.theController.findNode(vx, vy);
			if (thisNode != null)
			{
				this.theController.handle(Controller.Event.SELECT, thisNode);
			}
		}
		thisEvent.consume();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseEntered(final MouseEvent thisEvent)
	{
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseExited(final MouseEvent thisEvent)
	{
		// do nothing
	}

	// M O U S E M O T I O N L I S T E N E R

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseMoved(final MouseEvent thisEvent)
	{
		final int vx = thisEvent.getX();
		final int vy = thisEvent.getY();
		final INode thisNode = this.theController.findNode(vx, vy);
		final boolean again = this.theHotNode == thisNode;
		this.theHotNode = thisNode;
		if (thisNode != null && !again)
		{
			this.theController.handle(Controller.Event.HOVER, thisNode);
		}
		thisEvent.consume();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseDragged(final MouseEvent thisEvent)
	{
		final double theMaxShiftSpan = .5;

		// avoid wide mouse shift which will lead to cross-circle xlations
		Complex thisDragEnd = this.theController.viewToUnitCircle(new Point(thisEvent.getX(), thisEvent.getY()));
		if (EuclidianLocation.getDistance(this.theDragStart, thisDragEnd) > theMaxShiftSpan)
		{
			// keep shift direction but limit span
			thisDragEnd = Complex.makeFromArgAbs(thisDragEnd.sub(this.theDragStart).arg(), theMaxShiftSpan).add(this.theDragStart);
		}
		this.theDragEnd = thisDragEnd;
		this.wasMoved = true;
		this.wasDragged = true;
		this.theController.handle(Controller.Event.DRAG, null);
		thisEvent.consume();
	}

	// D R A G M O V E

	/**
	 * Rotate
	 */
	private synchronized void rotate()
	{
		this.theController.handle(Controller.Event.ROTATE, new Pair<Complex>(new Complex(this.theDragStart), new Complex(this.theDragEnd)));
		this.theDragStart.set(this.theDragEnd); // eat
	}

	/**
	 * Move
	 */
	private synchronized void move()
	{
		this.theController.handle(Controller.Event.MOVE, new Pair<Complex>(this.theDragStart, this.theDragEnd));
		this.theDragStart.set(this.theDragEnd); // eat
	}

	/**
	 * Drag
	 * 
	 * @return true if successful
	 */
	public boolean drag()
	{
		if (this.wasMoved)
		{
			// move it now
			switch (this.theDragMode)
			{
			case TRANSLATE:
				move();
				break;

			case ROTATE:
				rotate();
			}
			this.wasMoved = false;
			return true;
		}
		return false;
	}

	// H O T N O D E

	/**
	 * Reset hot node
	 */
	public void resetHotNode()
	{
		this.theHotNode = null;
	}

	/**
	 * Get hot node
	 * 
	 * @return hot node
	 */
	public INode getHotNode()
	{
		return this.theHotNode;
	}
}
