/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */

package treebolic.core;

import java.text.DecimalFormat;

/**
 * Complex
 * 
 * @author Bernard Bou
 */
public class Complex
{
	// C O N S T A N T S

	/**
	 * Zero
	 */
	static public final Complex ZERO = new Complex(0., 0.);

	/**
	 * One
	 */
	static public final Complex ONE = new Complex(1., 0.);

	/**
	 * North orientation
	 */
	static public final Complex NORTH = new Complex(0., 1.);

	/**
	 * South orientation
	 */
	static public final Complex SOUTH = new Complex(0., -1.);

	/**
	 * East orientation
	 */
	static public final Complex EAST = Complex.ONE;

	/**
	 * West orientation
	 */
	static public final Complex WEST = new Complex(-1., 0.);

	// D A T A

	/**
	 * Real part
	 */
	public double re;

	/**
	 * Imaginary part
	 */
	public double im;

	// C O N S T R U C T O R S

	/**
	 * Default constructor which defaults to (0,0)
	 */
	public Complex()
	{
		// this(0.,0.);
	}

	/**
	 * Construct complex from real and imaginary parts
	 * 
	 * @param x
	 *        real part
	 * @param y
	 *        imaginary part
	 */
	public Complex(final double x, final double y)
	{
		this.re = x;
		this.im = y;
	}

	/**
	 * Copy constructor
	 * 
	 * @param z
	 *        copied complex
	 */
	public Complex(final Complex z)
	{
		this.re = z.re;
		this.im = z.im;
	}

	/**
	 * Construct complex from real part, imaginary part defaulting to 0
	 * 
	 * @param x
	 *        real part
	 */
	public Complex(final double x)
	{
		this.re = x;
		this.im = 0.;
	}

	/**
	 * Construct complex from argument and mag re = cos(arg) im = sin(arg)
	 * 
	 * @param thisArg
	 *        argument a
	 * @param thisMag
	 *        magnitude m
	 * @return new complex of this argument a and magnitude m
	 */
	static public Complex makeFromArgAbs(final double thisArg, final double thisMag)
	{
		return new Complex(thisMag * Math.cos(thisArg), thisMag * Math.sin(thisArg));
	}

	/**
	 * Construct complex of magnitude 1 from argument re = cos(arg) im = sin(arg)
	 * 
	 * @param thisArg
	 *        argument a
	 * @return new complex of magnitude 1 and argument a
	 */
	static public Complex makeFromArg(final double thisArg)
	{
		return new Complex(Math.cos(thisArg), Math.sin(thisArg));
	}

	// S E T

	/**
	 * Set complex to value of another complex, no new complex object is created
	 * 
	 * @param z
	 *        complex
	 * @return original complex object, whose value has been set to z
	 */
	public Complex set(final Complex z)
	{
		this.re = z.re;
		this.im = z.im;
		return this;
	}

	/**
	 * Set complex to values of real and imaginary parts no new complex object is created
	 * 
	 * @param x
	 *        real part
	 * @param y
	 *        imaginary part
	 * @return original complex object, whose value has been set to x,z
	 */
	public Complex set(final double x, final double y)
	{
		this.re = x;
		this.im = y;
		return this;
	}

	// E Q U A L S

	/**
	 * Complex equality
	 * 
	 * @param z
	 *        complex
	 * @return true if this complex z0 is equal to complex z
	 */
	public boolean equals(final Complex z)
	{
		return this.re == z.re && this.im == z.im;
	}

	// A D D

	/**
	 * Add complex z to this complex
	 * 
	 * @param z
	 *        complex to be added to this complex
	 * @return this complex, whose value has been set to z0+z
	 */
	public Complex add(final Complex z)
	{
		this.re += z.re;
		this.im += z.im;
		return this;
	}

	/**
	 * Set this complex to z1+z2
	 * 
	 * @param z1
	 *        complex
	 * @param z2
	 *        complex
	 * @return this complex, whose value has been set to z1+z2 and whose original value is not used
	 */
	public Complex add(final Complex z1, final Complex z2)
	{
		this.re = z1.re + z2.re;
		this.im = z1.im + z2.im;
		return this;
	}

	// S U B S T R A C T

	/**
	 * Substract complex z from this complex
	 * 
	 * @param z
	 *        complex
	 * @return this complex, whose value has been set to z0-z
	 */
	public Complex sub(final Complex z)
	{
		this.re -= z.re;
		this.im -= z.im;
		return this;
	}

	/**
	 * Set this complex to z1-z2
	 * 
	 * @param z1
	 *        complex
	 * @param z2
	 *        complex
	 * @return this complex whose value has been set to z1-z2 and whose original value is not used
	 */
	public Complex sub(final Complex z1, final Complex z2)
	{
		this.re = z1.re - z2.re;
		this.im = z1.im - z2.im;
		return this;
	}

	// C O N J U G A T E

	/**
	 * Conjugate of this complex
	 * 
	 * @return this complex, whose value has been to the conjugate of this complex
	 */
	public Complex conj()
	{
		this.im = -this.im;
		return this;
	}

	/**
	 * Conjugate of z
	 * 
	 * @param z
	 *        complex
	 * @return this complex, whose value has been set to conjugate of z and whose original value is not used
	 */
	public Complex conj(final Complex z)
	{
		this.re = z.re;
		this.im = -z.im;
		return this;
	}

	// N E G A T I O N

	/**
	 * Negation of this complex
	 * 
	 * @return this complex, whose value has been set to the negation of this complex
	 */
	public Complex neg()
	{
		this.re = -this.re;
		this.im = -this.im;
		return this;
	}

	/**
	 * Negation of complex z
	 * 
	 * @param z
	 *        complex
	 * @return this complex, whose value has been set to the negation of z and whose original value is not used
	 */
	public Complex neg(final Complex z)
	{
		this.re = -z.re;
		this.im = -z.im;
		return this;
	}

	// D I V

	/**
	 * Division of this complex by z
	 * 
	 * @param z
	 *        complex
	 * @return this complex, whose value has been set to the division of this complex by complex z (=z0/z)
	 */
	public Complex div(final Complex z)
	{
		final double d = z.re * z.re + z.im * z.im; // (x2.x2 + y2.y2)
		final double t = (this.re * z.re + this.im * z.im) / d; // (x1.x2 + y1.y2) / (x2.x2 + y2.y2)
		this.im = (this.im * z.re - this.re * z.im) / d; // (y1.x2 - x1*y2) / (x2.x2 + y2.y2)
		this.re = t;
		return this;
	}

	/**
	 * Division
	 * 
	 * @param z1
	 *        complex
	 * @param z2
	 *        complex
	 * @return this complex, whose value has been set to the division of z1 by z2 (=z1/z2) and whose original value is not used
	 */
	public Complex div(final Complex z1, final Complex z2)
	{
		final double d = z2.re * z2.re + z2.im * z2.im;
		this.re = (z1.re * z2.re + z1.im * z2.im) / d;
		this.im = (z1.im * z2.re - z1.re * z2.im) / d;
		return this;
	}

	/**
	 * Inverse of this complex
	 * 
	 * @return this complex, as the result of inverting its original value (=1/z0)
	 */
	public Complex onediv()
	{
		// 1/z
		final double d = this.re * this.re + this.im * this.im; // (x.x + y.y)
		this.re = this.re / d;
		this.im = -this.im / d; // -y / (x.x + y.y)
		return this;
	}

	// M U L T I P L Y

	/**
	 * Multiplication of this complex by z
	 * 
	 * @param z
	 * @return this complex, whose value has been multiplied by z (=z0.z)
	 */
	public Complex mul(final Complex z)
	{
		final double t = this.re * z.im + this.im * z.re; // x1.y2 + x2.y1
		this.re = this.re * z.re - this.im * z.im; // x1.x2 - y1.y2
		this.im = t;
		return this;
	}

	/**
	 * Multiplication
	 * 
	 * @param a
	 *        complex
	 * @param b
	 *        complex
	 * @return this complex as the result of a.b and whose original value not used
	 */
	public Complex mul(final Complex a, final Complex b)
	{
		this.re = a.re * b.re - a.im * b.im;
		this.im = a.re * b.im + a.im * b.re;
		return this;
	}

	// M U L T I P L Y / D I V I D E . B Y . D O U B L E

	/**
	 * Multiply complex by double
	 * 
	 * @param z
	 *        complex
	 * @param m
	 *        multiplicator
	 * @return this complex, as the result of multiplying z by m
	 */
	public Complex multiply(final Complex z, final double m)
	{
		this.im = z.im * m;
		this.re = z.re * m;
		return this;
	}

	/**
	 * Multiply this complex by double
	 * 
	 * @param m
	 *        multiplicator
	 * @return this complex, multiplied by m
	 */
	public Complex multiply(final double m)
	{
		this.im *= m;
		this.re *= m;
		return this;
	}

	/**
	 * Divide complex by double
	 * 
	 * @param z
	 *        complex
	 * @param d
	 *        divisor
	 * @return this complex, as the result of dividying z by d
	 */
	public Complex divide(final Complex z, final double d)
	{
		this.im = z.im / d;
		this.re = z.re / d;
		return this;
	}

	/**
	 * Divide this complex by double
	 * 
	 * @param d
	 *        divisor
	 * @return this complex, divided by d
	 */
	public Complex divide(final double d)
	{
		this.im /= d;
		this.re /= d;
		return this;
	}

	// A R G U M E N T / A N G L E
	// re = abs().cos(arg)
	// im = abs().sin(arg)

	/**
	 * Argument
	 * 
	 * @return argument of complex
	 */
	public double arg()
	{
		return Math.atan2(this.im, this.re);
	}

	// M A G N I T U D E / N O R M / R O
	// re = mag().cos(arg)
	// im = mag().sin(arg)

	/**
	 * Magnitude
	 * 
	 * @return magnitude
	 */
	public double mag()
	{
		return Math.sqrt(this.re * this.re + this.im * this.im);
	}

	/**
	 * Squared magnitude
	 * 
	 * @return squared magnitude
	 */
	public double abs2()
	{
		return this.re * this.re + this.im * this.im;
	}

	// N O R M A L I Z E . T O . M A G N I T U D E . O N E . C O M P L E X

	/**
	 * Map to magnitude one complex
	 * 
	 * @return magnitude 1 complex with same argument
	 */
	public Complex normalize()
	{
		final double r = mag();
		this.re /= r;
		this.im /= r;
		return this;
	}

	// S T R I N G

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		final DecimalFormat thisFormat = new DecimalFormat("0.###");
		return "(" + thisFormat.format(this.re) + " ; " + thisFormat.format(this.im) + ")";
	}
}
