/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.core;

/**
 * Hypercircle node location in hyperspace
 * 
 * @author Bernard Bou
 */
public class HyperLocation
{
	// C O N S T A N T S

	/**
	 * Where border starts (distance to origin)
	 */
	static final private double BORDER = .95;

	// D A T A

	/**
	 * Center as computed by initial layout
	 */
	public final Complex center0;

	/**
	 * Center as mapped by current transform
	 */
	public final Complex center;

	/**
	 * Radius
	 */
	public double radius;

	/**
	 * Distance to (0,0)
	 */
	public double dist;

	// S T A T E

	/**
	 * Computation state (set whenever either center or radius or both change)
	 */
	public boolean isDirty;

	// B O R D E R

	/**
	 * Whether this hypercircle nears border
	 */
	public boolean isBorder;

	// C O N S T R U C T O R

	/**
	 * Construct hypercircle
	 */
	public HyperLocation()
	{
		this.center0 = new Complex();
		this.center = new Complex();
		this.radius = 0.;
		this.isBorder = false;
		this.isDirty = false;
	}

	// S E T

	/**
	 * Set hypercircle data
	 * 
	 * @param o
	 *        center
	 * @param r
	 *        radius
	 */
	public void set(final Complex o, final double r)
	{
		this.center0.set(o);
		this.center.set(o);
		this.dist = this.center.mag();
		this.isBorder = this.dist > HyperLocation.BORDER;
		this.radius = r;
		this.isDirty = true;
	}

	/**
	 * Reset to untransformed initial position
	 */
	public void reset()
	{
		this.center.set(this.center0);
		this.dist = this.center.mag();
		this.isBorder = this.dist > HyperLocation.BORDER;
		this.isDirty = true;
	}

	/**
	 * Clone
	 * 
	 * @param thisHyperLocation
	 *        hyper location
	 */
	public void clone(final HyperLocation thisHyperLocation)
	{
		this.center0.set(thisHyperLocation.center0);
		this.center.set(thisHyperLocation.center);
		this.dist = thisHyperLocation.dist;
		this.isBorder = thisHyperLocation.isBorder;
		this.radius = thisHyperLocation.radius;
		this.isDirty = true;
	}

	// T R A N S F O R M

	/**
	 * Apply transform to this hypercircle
	 * 
	 * @param t
	 *        transform
	 */
	public void transform(final IHyperTransform t)
	{
		// map
		t.map(this.center.set(this.center0));

		// distance to (0,0)
		this.dist = this.center.mag();

		// normalization (should not occur with proper transform)
		if (this.dist > 1.)
		{
			this.center.normalize();
			this.dist = this.center.mag();
		}

		// compute if off-limit
		this.isBorder = this.dist > HyperLocation.BORDER;
		this.isDirty = true;
	}

	// H Y P E R B O L I C . D I S T A N C E

	/**
	 * Distance between 2 points
	 * 
	 * @param z1
	 *        point1
	 * @param z2
	 *        point2
	 * @return distance between two points
	 */
	static public double getHyperDistance(final Complex z1, final Complex z2)
	{
		// 2*atanh((z1-z2)/(1-z1*~z2));
		// 2*atanh(z) if distance from 0,0 origin
		final Complex denom = new Complex().conj(z2).mul(z1).neg().add(Complex.ONE);
		final Complex z = new Complex().sub(z1, z2).div(denom);
		return 2. * XMath.atanh(z.mag());
	}

	// S T R I N G

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		return "o0=" + this.center0 + ", o=" + this.center + ", r=" + this.radius + ", d=" + this.dist;
	}
}
