/**
 * Title :	Treebolic generator
 * Description: Treebolic generator
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.generator.domtree;

import java.awt.Color;
import java.awt.Component;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextPane;
import javax.swing.JTree;
import javax.swing.text.BadLocationException;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledDocument;
import javax.swing.tree.TreeCellRenderer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Renderer
 * 
 * @author Bernard Bou
 */
public class Renderer implements TreeCellRenderer
{
	// icons

	/**
	 * Element icon
	 */
	static private ImageIcon theElementIcon;

	/**
	 * Text icon
	 */
	static private ImageIcon theTextIcon;

	/**
	 * Default icon
	 */
	static private ImageIcon theDefaultIcon;

	// styles

	/**
	 * Style for element name
	 */
	static private SimpleAttributeSet theElementNameStyle = new SimpleAttributeSet();

	/**
	 * Style for element data
	 */
	static private SimpleAttributeSet theElementValueStyle = new SimpleAttributeSet();

	/**
	 * Style for node name
	 */
	static private SimpleAttributeSet theDefaultNameStyle = new SimpleAttributeSet();

	/**
	 * Style for node data
	 */
	static private SimpleAttributeSet theDefaultValueStyle = new SimpleAttributeSet();

	/**
	 * Style for id data
	 */
	static private SimpleAttributeSet theIdStyle = new SimpleAttributeSet();

	/**
	 * Patterns
	 */
	private final Pattern theIdPattern = Pattern.compile("id=\"(([^\"]*))\""); //$NON-NLS-1$

	/**
	 * Pattern list
	 */
	protected List<Pattern> thePatterns;

	/**
	 * Styles for patterns
	 */
	protected Map<Pattern, SimpleAttributeSet> thePatternToStyleMap;

	static
	{
		// styles
		StyleConstants.setFontFamily(Renderer.theElementNameStyle, "Dialog"); //$NON-NLS-1$
		StyleConstants.setFontSize(Renderer.theElementNameStyle, 10);
		StyleConstants.setBold(Renderer.theElementNameStyle, true);
		StyleConstants.setBackground(Renderer.theElementNameStyle, new Color(255, 255, 200));

		StyleConstants.setFontFamily(Renderer.theElementValueStyle, "Dialog"); //$NON-NLS-1$
		StyleConstants.setFontSize(Renderer.theElementValueStyle, 10);

		StyleConstants.setFontFamily(Renderer.theDefaultNameStyle, "Dialog"); //$NON-NLS-1$
		StyleConstants.setFontSize(Renderer.theDefaultNameStyle, 10);
		StyleConstants.setItalic(Renderer.theDefaultNameStyle, true);
		StyleConstants.setForeground(Renderer.theDefaultNameStyle, Color.GRAY);

		StyleConstants.setFontFamily(Renderer.theDefaultValueStyle, "Dialog"); //$NON-NLS-1$
		StyleConstants.setFontSize(Renderer.theDefaultValueStyle, 10);
		StyleConstants.setForeground(Renderer.theDefaultValueStyle, Color.GRAY);

		StyleConstants.setForeground(Renderer.theIdStyle, Color.BLUE);

		// icons
		Renderer.theElementIcon = new ImageIcon(Renderer.class.getResource("images/treeelement.gif")); //$NON-NLS-1$
		Renderer.theTextIcon = new ImageIcon(Renderer.class.getResource("images/treetext.gif")); //$NON-NLS-1$
		Renderer.theDefaultIcon = new ImageIcon(Renderer.class.getResource("images/treedefault.gif")); //$NON-NLS-1$
	}

	// components

	/**
	 * Renderer component
	 */
	protected JPanel thePanel;

	/**
	 * Icon subcomponent
	 */
	protected JLabel theIconComponent;

	/**
	 * Text sub component
	 */
	protected JTextPane theTextComponent;

	/**
	 * Constructor
	 */
	public Renderer()
	{
		// components
		this.theIconComponent = new JLabel();
		this.theTextComponent = new JTextPane();

		// assemble
		this.theIconComponent.setAlignmentY(Component.BOTTOM_ALIGNMENT);
		this.theTextComponent.setAlignmentY(Component.BOTTOM_ALIGNMENT);
		this.thePanel = new JPanel();
		this.thePanel.setOpaque(false);
		this.thePanel.setLayout(new BoxLayout(this.thePanel, BoxLayout.X_AXIS));
		this.thePanel.add(this.theIconComponent);
		this.thePanel.add(this.theTextComponent);

		// styles for patterns
		this.thePatterns = new ArrayList<Pattern>();
		this.thePatterns.add(this.theIdPattern);
		this.thePatternToStyleMap = new HashMap<Pattern, SimpleAttributeSet>();
		this.thePatternToStyleMap.put(this.theIdPattern, Renderer.theIdStyle);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeCellRenderer#getTreeCellRendererComponent(javax.swing.JTree, java.lang.Object, boolean, boolean, boolean, int, boolean)
	 */
	@Override
	public Component getTreeCellRendererComponent(final JTree tree, final Object value, final boolean sel, final boolean expanded, final boolean leaf, final int row, final boolean hasFocus)
	{
		final Node thisNode = (Node) value;

		// decorator
		final DefaultDecorator thisDecorator = makeDecorator(thisNode);

		// strings
		final String thisName = thisDecorator.getName();
		final String thisValue = thisDecorator.getValue();
		final int thisNameLength = thisName.length();

		// document
		final StyledDocument thisStyledDocument = this.theTextComponent.getStyledDocument();

		// set text
		try
		{
			thisStyledDocument.remove(0, thisStyledDocument.getLength());
			thisStyledDocument.insertString(0, thisName, getNameStyle(thisNode));
			if (thisValue != null)
			{
				thisStyledDocument.insertString(thisNameLength, " ", null); //$NON-NLS-1$
				thisStyledDocument.insertString(thisNameLength + 1, thisValue, getValueStyle(thisNode));
			}
		}
		catch (final BadLocationException thisException)
		{
			// do nothing
		}

		// patterns
		applyStyleToPatterns(thisValue, thisNameLength);

		// icon
		final Icon thisIcon = getIconStyle(thisNode);
		this.theIconComponent.setIcon(thisIcon);

		return this.thePanel;
	}

	/**
	 * Apply styles to patterns
	 * 
	 * @param thisString
	 *        value string
	 * @param thisOffset
	 *        offset of value in text
	 */
	protected void applyStyleToPatterns(final String thisString, final int thisOffset)
	{
		if (thisString != null)
		{
			final StyledDocument thisStyledDocument = this.theTextComponent.getStyledDocument();
			for (final Pattern thisPattern : this.thePatterns)
			{
				final Matcher thisMatcher = thisPattern.matcher(thisString);
				while (thisMatcher.find())
				{
					for (int g = 1; g < thisMatcher.groupCount(); g++)
					{
						final int from = thisMatcher.start(g) + thisOffset + 1;
						final int to = thisMatcher.end(g) + thisOffset + 1;
						thisStyledDocument.setCharacterAttributes(from, to - from, this.thePatternToStyleMap.get(thisPattern), false);
					}
				}
			}
		}
	}

	/**
	 * Make node decorator
	 * 
	 * @param thisNode
	 *        node
	 * @return node decorator
	 */
	protected DefaultDecorator makeDecorator(final Node thisNode)
	{
		return thisNode instanceof Element ? new ElementDecorator(thisNode) : new DefaultDecorator(thisNode);
	}

	/**
	 * Get name style for node
	 * 
	 * @param thisNode
	 *        node
	 * @return style
	 */
	protected SimpleAttributeSet getNameStyle(final Node thisNode)
	{
		return thisNode instanceof Element ? Renderer.theElementNameStyle : Renderer.theDefaultNameStyle;
	}

	/**
	 * Get value style for node
	 * 
	 * @param thisNode
	 *        node
	 * @return style
	 */
	protected SimpleAttributeSet getValueStyle(final Node thisNode)
	{
		return thisNode instanceof Element ? Renderer.theElementValueStyle : Renderer.theDefaultValueStyle;
	}

	/**
	 * Get icon for node
	 * 
	 * @param thisNode
	 *        node
	 * @return style
	 */
	protected Icon getIconStyle(final Node thisNode)
	{
		switch (thisNode.getNodeType())
		{
		case Node.ELEMENT_NODE:
			return Renderer.theElementIcon;
		case Node.TEXT_NODE:
			return Renderer.theTextIcon;
		default:
			return Renderer.theDefaultIcon;
		}
	}
}