/**
 * Title :	Treebolic generator
 * Description: Treebolic generator
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.generator.domtree;

import java.util.Vector;

import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class TreeModelAdapter implements TreeModel
{
	// D A T A

	/**
	 * DOM document
	 */
	private final Document theDocument;

	/**
	 * From node
	 */
	private final String theFromNode;

	/**
	 * Whether to consider DOM elements
	 */
	private boolean elementsOnly = false;

	/**
	 * Listener support
	 */
	private final Vector<TreeModelListener> theListenerList = new Vector<TreeModelListener>();

	// C O N S T R U C T

	/**
	 * Constructor
	 * 
	 * @param thisDocument
	 *        document
	 * @param thisFromNode
	 *        from node
	 * @param thisElementsOnlyFlag
	 *        whether to consider elements only
	 */
	public TreeModelAdapter(final Document thisDocument, final String thisFromNode, final boolean thisElementsOnlyFlag)
	{
		this.theDocument = thisDocument;
		this.theFromNode = thisFromNode;
		this.elementsOnly = thisElementsOnlyFlag;
	}

	// T R E E M O D E L

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#getRoot()
	 */
	@Override
	public Object getRoot()
	{
		if (this.theFromNode == null)
			return this.theDocument.getDocumentElement();
		else
		{
			final NodeList theseElements = this.theDocument.getElementsByTagName(this.theFromNode);
			return theseElements == null || theseElements.getLength() < 1 ? null : theseElements.item(0);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#isLeaf(java.lang.Object)
	 */
	@Override
	public boolean isLeaf(final Object thisNode)
	{
		return childCountOf((Node) thisNode) == 0;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#getChildCount(java.lang.Object)
	 */
	@Override
	public int getChildCount(final Object thisNode)
	{
		return childCountOf((Node) thisNode);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#getChild(java.lang.Object, int)
	 */
	@Override
	public Object getChild(final Object thisNode, final int thisIndex)
	{
		return nthChildOf((Node) thisNode, thisIndex);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#getIndexOfChild(java.lang.Object, java.lang.Object)
	 */
	@Override
	public int getIndexOfChild(final Object thisNode, final Object thisChild)
	{
		return indexOf((Node) thisNode, (Node) thisChild);
	}

	// M O D I F I C A T I O N

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#valueForPathChanged(javax.swing.tree.TreePath, java.lang.Object)
	 */
	@Override
	public void valueForPathChanged(final TreePath path, final Object newValue)
	{
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#addTreeModelListener(javax.swing.event.TreeModelListener)
	 */
	@Override
	public void addTreeModelListener(final TreeModelListener thisListener)
	{
		if (thisListener != null && !this.theListenerList.contains(thisListener))
		{
			this.theListenerList.addElement(thisListener);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.tree.TreeModel#removeTreeModelListener(javax.swing.event.TreeModelListener)
	 */
	@Override
	public void removeTreeModelListener(final TreeModelListener thisListener)
	{
		if (thisListener != null)
		{
			this.theListenerList.removeElement(thisListener);
		}
	}

	// T R E E M O D E L . H E L P E R S

	/**
	 * Get nth child node
	 * 
	 * @param thisNode
	 *        node
	 * @param thatIndex
	 *        index
	 * @return nth child
	 */
	private Node nthChildOf(final Node thisNode, final int thatIndex)
	{
		if (!this.elementsOnly)
			return thisNode.getChildNodes().item(thatIndex);

		// nth element node
		int thisIndex = 0;
		for (int i = 0; i < thisNode.getChildNodes().getLength(); i++)
		{
			final Node thisChild = thisNode.getChildNodes().item(i);
			if (thisChild.getNodeType() == Node.ELEMENT_NODE && thisIndex++ == thatIndex)
				return thisChild;
		}
		return null;
	}

	/**
	 * Get child count
	 * 
	 * @param thisNode
	 *        node
	 * @return number of children
	 */
	private int childCountOf(final Node thisNode)
	{
		if (!this.elementsOnly)
			return thisNode.getChildNodes().getLength();

		// number of element nodes
		int thisCount = 0;
		for (int i = 0; i < thisNode.getChildNodes().getLength(); i++)
		{
			final Node thisChild = thisNode.getChildNodes().item(i);
			if (thisChild.getNodeType() == Node.ELEMENT_NODE)
			{
				++thisCount;
			}
		}
		return thisCount;
	}

	/**
	 * Get index of child node
	 * 
	 * @param thisNode
	 *        node
	 * @param thatChild
	 *        child
	 * @return index or -1 if not found
	 */
	private int indexOf(final Node thisNode, final Node thatChild)
	{
		final int thisCount = childCountOf(thisNode);
		for (int i = 0; i < thisCount; i++)
		{
			final Node thisChild = nthChildOf(thisNode, i);
			if (thatChild == thisChild)
				return i;
		}
		return -1;
	}
}
