/**
 * Title :	Treebolic generator
 * Description: Treebolic generator
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.generator.tree;

import java.awt.Component;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JTree;
import javax.swing.TransferHandler;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreePath;

import treebolic.generator.Pair;
import treebolic.model.MenuItem;
import treebolic.model.TreeMutableNode;

/**
 * Transfer handler for drag and drop operations
 * 
 * @author Bernard Bou
 */
public class TreeTransferHandler extends TransferHandler
{
	private static final long serialVersionUID = 1L;

	// F L A V O R

	/**
	 * Data flavour
	 */
	private static DataFlavor theFlavor;

	static
	{
		try
		{
			TreeTransferHandler.theFlavor = new DataFlavor(DataFlavor.javaJVMLocalObjectMimeType + ";class=\"" + DefaultMutableTreeNode.class.getName() + "\""); //$NON-NLS-1$ //$NON-NLS-2$
		}
		catch (final ClassNotFoundException e)
		{
			e.printStackTrace();
		}
	}

	// T R A N S F E R A B L E . O B J E C T

	/**
	 * Transferable data
	 */
	class TransferableNode implements Transferable
	{
		/**
		 * Node
		 */
		private final DefaultMutableTreeNode theNode;

		/**
		 * Constructor
		 * 
		 * @param thisNode
		 *        node object to transfer
		 */
		public TransferableNode(final Object thisNode)
		{
			this.theNode = (DefaultMutableTreeNode) thisNode;
		}

		// flavor
		/*
		 * (non-Javadoc)
		 * @see java.awt.datatransfer.Transferable#getTransferDataFlavors()
		 */
		@SuppressWarnings("synthetic-access")
		@Override
		public DataFlavor[] getTransferDataFlavors()
		{
			final DataFlavor[] theseFlavors = { TreeTransferHandler.theFlavor };
			return theseFlavors;
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.datatransfer.Transferable#isDataFlavorSupported(java.awt.datatransfer.DataFlavor)
		 */
		@SuppressWarnings("synthetic-access")
		@Override
		public boolean isDataFlavorSupported(final DataFlavor thisFlavor)
		{
			return thisFlavor.equals(TreeTransferHandler.theFlavor);
		}

		// data
		/*
		 * (non-Javadoc)
		 * @see java.awt.datatransfer.Transferable#getTransferData(java.awt.datatransfer.DataFlavor)
		 */
		@Override
		public Object getTransferData(final DataFlavor flavor) throws UnsupportedFlavorException, IOException
		{
			return this.theNode;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.TransferHandler#getSourceActions(javax.swing.JComponent)
	 */
	@Override
	public int getSourceActions(final JComponent c)
	{
		return TransferHandler.MOVE;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.TransferHandler#createTransferable(javax.swing.JComponent)
	 */
	@Override
	protected Transferable createTransferable(final JComponent thisSourceComponent)
	{
		final JTree thisSourceTree = (JTree) thisSourceComponent;
		return new TransferableNode(thisSourceTree.getSelectionModel().getSelectionPath().getLastPathComponent());
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.TransferHandler#canImport(javax.swing.TransferHandler.TransferSupport)
	 */
	@Override
	public boolean canImport(final TransferSupport thisSupport)
	{
		final Pair<TreePath, TreePath> theseNodes = getNodes(thisSupport);
		if (theseNodes != null)
			return testMove(theseNodes.theFirst, theseNodes.theSecond);
		return false;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.TransferHandler#importData(javax.swing.TransferHandler.TransferSupport)
	 */
	@Override
	public boolean importData(final TransferSupport thisSupport)
	{
		final Pair<TreePath, TreePath> theseNodes = getNodes(thisSupport);
		if (theseNodes != null)
		{
			if (thisSupport.isDrop())
				return move((Tree) thisSupport.getComponent(), theseNodes.theFirst, theseNodes.theSecond);
		}
		return false;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.TransferHandler#exportDone(javax.swing.JComponent, java.awt.datatransfer.Transferable, int)
	 */
	@Override
	protected void exportDone(final JComponent thisComponent, final Transferable data, final int action)
	{
		// do nothing
	}

	// I M P L E M E N T

	/**
	 * Get pair of source tree and destination treepaths
	 * 
	 * @param thisSupport
	 *        transfer support
	 * @return source tree and destination treepaths
	 */
	private Pair<TreePath, TreePath> getNodes(final TransferSupport thisSupport)
	{
		// source flavour
		if (!thisSupport.isDataFlavorSupported(TreeTransferHandler.theFlavor))
			return null;

		// destination tree
		final Component thisComponent = thisSupport.getComponent();
		final JTree thisDestinationTree = (JTree) thisComponent;

		// destination data
		final DropLocation thisDropLocation = thisSupport.getDropLocation();
		final TreePath thisDestinationPath = thisDestinationTree.getClosestPathForLocation(thisDropLocation.getDropPoint().x, thisDropLocation.getDropPoint().y);
		final DefaultMutableTreeNode thisDestinationNode = (DefaultMutableTreeNode) thisDestinationPath.getLastPathComponent();
		final Object thisDestinationObject = thisDestinationNode.getUserObject();

		// source data
		TreePath thisSourcePath = null;
		Object thisData = null;
		final Transferable thisTransferable = thisSupport.getTransferable();
		try
		{
			thisData = thisTransferable.getTransferData(TreeTransferHandler.theFlavor);
		}
		catch (final UnsupportedFlavorException thisException)
		{
			return null;
		}
		catch (final IOException thisException)
		{
			return null;
		}
		final DefaultMutableTreeNode thisNode = (DefaultMutableTreeNode) thisData;
		final Object thisSourceObject = thisNode.getUserObject();
		thisSourcePath = new TreePath(thisNode.getPath());

		// type check
		if (thisSourceObject instanceof TreeMutableNode && thisDestinationObject instanceof TreeMutableNode)
			return new Pair<TreePath, TreePath>(thisSourcePath, thisDestinationPath);
		if (thisSourceObject instanceof MenuItemWrapper && thisDestinationObject instanceof MenuWrapper)
			return new Pair<TreePath, TreePath>(thisSourcePath, thisDestinationPath);

		return null;
	}

	/**
	 * Whether source node can drag and drop to destination node
	 * 
	 * @param thisSource
	 *        source node
	 * @param thisDestination
	 *        destination node
	 * @return true if source node can drag and drop to destination node
	 */
	private boolean testMove(final TreePath thisSource, final TreePath thisDestination)
	{
		if (thisDestination != null && thisSource != null)
		{
			final DefaultMutableTreeNode thisDestinationNode = (DefaultMutableTreeNode) thisDestination.getLastPathComponent();
			final DefaultMutableTreeNode thisSourceNode = (DefaultMutableTreeNode) thisSource.getLastPathComponent();
			final Object thisDestinationObject = thisDestinationNode.getUserObject();
			final Object thisSourceObject = thisSourceNode.getUserObject();
			if (thisDestinationObject instanceof TreeMutableNode && thisSourceObject instanceof TreeMutableNode)
				return !thisSource.isDescendant(thisDestination);
			else if (thisDestinationObject instanceof MenuWrapper && thisSourceObject instanceof MenuItemWrapper)
				return !thisSource.isDescendant(thisDestination);
		}
		return false;
	}

	/**
	 * Move source node as child of destination node
	 * 
	 * @param thisSource
	 *        source node
	 * @param thisDestination
	 *        destination node
	 * @return true if successful
	 */
	private boolean move(final Tree thisTree, final TreePath thisSource, final TreePath thisDestination)
	{
		// tree
		final DefaultMutableTreeNode thisParentTreeNode = (DefaultMutableTreeNode) thisDestination.getLastPathComponent();
		final DefaultMutableTreeNode thisTreeNode = (DefaultMutableTreeNode) thisSource.getLastPathComponent();

		// tree remove
		thisTree.removeFromParent(thisTreeNode);

		// tree add
		thisTree.addToParent(thisParentTreeNode, thisTreeNode);

		// objects
		final Object thisParentObject = thisParentTreeNode.getUserObject();
		final Object thisObject = thisTreeNode.getUserObject();

		if (thisObject instanceof TreeMutableNode && thisParentObject instanceof TreeMutableNode)
		{
			final TreeMutableNode thisParentNode = (TreeMutableNode) thisParentObject;
			final TreeMutableNode thisNode = (TreeMutableNode) thisObject;

			// object remove
			thisNode.removeFromParent();

			// object add
			thisNode.addToParent(thisParentNode);
		}
		else
		{
			final MenuWrapper thisParentNode = (MenuWrapper) thisParentObject;
			final MenuItemWrapper thisNode = (MenuItemWrapper) thisObject;
			final List<MenuItem> thisMenu = thisParentNode.theMenu;
			final MenuItem thisMenuItem = thisNode.theMenuItem;
			thisMenu.remove(thisMenuItem);
			thisMenu.add(thisMenu.size(), thisMenuItem);
		}
		return true;
	}
}
