/**
 * Title :	Treebolic generator
 * Description: Treebolic generator
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.generator.tree;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.swing.JScrollPane;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

import treebolic.generator.Controller;
import treebolic.model.IEdge;
import treebolic.model.MenuItem;
import treebolic.model.TreeMutableEdge;
import treebolic.model.TreeMutableNode;

/**
 * Tree view
 * 
 * @author Bernard Bou
 */
public class TreeView extends JScrollPane
{
	private static final long serialVersionUID = 1L;

	// D A T A

	/**
	 * Tree component
	 */
	private final Tree theTree;

	/**
	 * Dirty flag
	 */
	public boolean dirty;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 */
	public TreeView()
	{
		this.theTree = new Tree();
		setViewportView(this.theTree);
	}

	// C O N N E C T

	/**
	 * Connect to controller
	 * 
	 * @param thisController
	 *        controller
	 */
	public void connect(final Controller thisController)
	{
		this.theTree.connect(thisController);
	}

	// M O D E L

	/**
	 * Set tree model
	 * 
	 * @param thisTreeModel
	 */
	public void set(final TreeModel thisTreeModel)
	{
		// add listener to tree model
		if (thisTreeModel != null)
		{
			thisTreeModel.addTreeModelListener(new TreeModelListener()
			{
				@Override
				public void treeNodesChanged(final TreeModelEvent thisEvent)
				{
					TreeView.this.dirty = true;
				}

				@Override
				public void treeNodesInserted(final TreeModelEvent thisE)
				{
					TreeView.this.dirty = true;
				}

				@Override
				public void treeNodesRemoved(final TreeModelEvent thisE)
				{
					TreeView.this.dirty = true;
				}

				@Override
				public void treeStructureChanged(final TreeModelEvent thisE)
				{
					TreeView.this.dirty = true;
				}
			});
		}
		this.theTree.setModel(thisTreeModel);

		// visibility
		if (thisTreeModel != null)
		{
			this.theTree.expandAll();
		}
	}

	/**
	 * Add object
	 */
	public void editAdd()
	{
		// selected tree node
		final TreePath thisParentPath = this.theTree.getSelectionPath();
		if (thisParentPath == null)
			return;

		// parent tree node
		final DefaultMutableTreeNode thisParentTreeNode = (DefaultMutableTreeNode) thisParentPath.getLastPathComponent();
		if (thisParentTreeNode == null)
			return;

		// parent's number of children
		final int thisPos = thisParentTreeNode.getChildCount();

		// object
		final Object thisObject = thisParentTreeNode.getUserObject();
		if (thisObject instanceof treebolic.model.MutableNode)
		{
			// parent
			final TreeMutableNode thisParentNode = (TreeMutableNode) thisObject;

			// id
			final String thisId = TreeMutableNode.makeNodeId();

			// label
			String thisLabel = thisParentNode.getLabel();
			if (thisLabel == null)
			{
				thisLabel = ""; //$NON-NLS-1$
			}
			else
			{
				thisLabel += "-"; //$NON-NLS-1$
			}
			thisLabel += thisPos;

			// new node (handle parent with addToParent)
			final TreeMutableNode thisNode = new TreeMutableNode(null, thisId);
			thisNode.setLabel(thisLabel);

			// object add to parent
			thisNode.addToParent(thisParentNode);

			// new tree node
			final DefaultMutableTreeNode thisTreeNode = new DefaultMutableTreeNode();
			thisTreeNode.setUserObject(thisNode);

			// tree add to parent
			this.theTree.addToParent(thisParentTreeNode, thisTreeNode);
		}
		else if (thisObject instanceof EdgesWrapper)
		{
			final EdgesWrapper thisWrapper = (EdgesWrapper) thisObject;
			List<IEdge> thisEdgeList = thisWrapper.theEdgeList;
			if (thisEdgeList == null)
			{
				thisEdgeList = new ArrayList<IEdge>();
				thisWrapper.theEdgeList = thisEdgeList;
				thisWrapper.theModel.theTree.setEdges(thisEdgeList);
			}

			// new edge
			final TreeMutableEdge thisEdge = new TreeMutableEdge(null, null);

			// object add
			thisEdgeList.add(thisEdge);

			// new tree node
			final DefaultMutableTreeNode thisTreeNode = new DefaultMutableTreeNode();
			thisTreeNode.setUserObject(thisEdge);

			// tree add to parent
			this.theTree.addToParent(thisParentTreeNode, thisTreeNode);
		}
		else if (thisObject instanceof MenuWrapper)
		{
			final MenuWrapper thisWrapper = (MenuWrapper) thisObject;
			List<MenuItem> thisMenu = thisWrapper.theMenu;
			if (thisMenu == null)
			{
				thisMenu = new ArrayList<MenuItem>();
				thisWrapper.theMenu = thisMenu;
				thisWrapper.theSettings.theMenu = thisMenu;
			}

			// new menu item
			final MenuItem thisMenuItem = new MenuItem();

			// object add
			thisMenu.add(thisMenuItem);

			// new tree node
			final DefaultMutableTreeNode thisTreeNode = new DefaultMutableTreeNode();
			thisTreeNode.setUserObject(new MenuItemWrapper(thisMenuItem, thisWrapper.theSettings));

			// tree add to parent
			this.theTree.addToParent(thisParentTreeNode, thisTreeNode);
		}
	}

	/**
	 * Remove object
	 */
	public void editRemove()
	{
		// selected tree node
		final TreePath thisPath = this.theTree.getSelectionPath();
		if (thisPath == null)
			return;
		final DefaultMutableTreeNode thisTreeNode = (DefaultMutableTreeNode) thisPath.getLastPathComponent();
		if (thisTreeNode == null)
			return;

		// parent tree node
		final DefaultMutableTreeNode thisParentTreeNode = (DefaultMutableTreeNode) thisTreeNode.getParent();
		if (thisParentTreeNode == null)
			return;

		// object remove
		final Object thisObject = thisTreeNode.getUserObject();
		final Object thisParentObject = thisParentTreeNode.getUserObject();
		if (thisParentObject instanceof TreeMutableNode)
		{
			final TreeMutableNode thisNode = (TreeMutableNode) thisObject;
			thisNode.removeFromParent();
		}
		else if (thisParentObject instanceof EdgesWrapper)
		{
			final EdgesWrapper thisWrapper = (EdgesWrapper) thisParentObject;
			thisWrapper.theEdgeList.remove(thisObject);
		}
		else if (thisParentObject instanceof MenuWrapper)
		{
			final MenuWrapper thisMenuWrapper = (MenuWrapper) thisParentObject;
			final MenuItemWrapper thisMenuItemWrapper = (MenuItemWrapper) thisObject;
			thisMenuWrapper.theMenu.remove(thisMenuItemWrapper.theMenuItem);
		}
		else
			return;

		// tree remove
		this.theTree.removeFromParent(thisTreeNode);
	}

	/**
	 * Search for tree node whose user object has given class name
	 * 
	 * @param thisClassName
	 *        class name
	 * @return tree node
	 */
	public DefaultMutableTreeNode search(final String thisClassName)
	{
		final TreeModel thisModel = this.theTree.getModel();
		final DefaultMutableTreeNode thisNode = (DefaultMutableTreeNode) thisModel.getRoot();
		return search(thisNode, thisClassName);
	}

	// S E A R C H

	/**
	 * Search for tree node whose user object has given class name
	 * 
	 * @param thisNode
	 *        start node
	 * @param thisClassName
	 *        class name
	 * @return tree node
	 */
	@SuppressWarnings("unchecked")
	private DefaultMutableTreeNode search(final DefaultMutableTreeNode thisNode, final String thisClassName)
	{
		if (thisNode.getUserObject().getClass().getName().endsWith(thisClassName))
			return thisNode;

		for (final Enumeration<DefaultMutableTreeNode> theseChildren = thisNode.children(); theseChildren.hasMoreElements();)
		{
			final DefaultMutableTreeNode thisChild = theseChildren.nextElement();
			final DefaultMutableTreeNode thisResult = search(thisChild, thisClassName);
			if (thisResult != null)
				return thisResult;
		}
		return null;
	}

	// S E L E C T

	/**
	 * Select tree node
	 * 
	 * @param thisNode
	 *        node
	 */
	public void select(final DefaultMutableTreeNode thisNode)
	{
		final TreePath thisPath = new TreePath(thisNode.getPath());
		this.theTree.setSelectionPath(thisPath);
	}

	// E X P A N D / C O L L A P S E

	/**
	 * Expand
	 */
	public void expand()
	{
		this.theTree.expandAll();
	}

	/**
	 * Collapse
	 */
	public void collapse()
	{
		this.theTree.collapseAll();
	}
}
