/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.model;

import java.util.List;

import treebolic.core.Complex;
import treebolic.core.EuclidianLocation;

/**
 * Model
 * 
 * @author Bernard Bou
 */
public class Tree
{
	// D A T A

	/**
	 * Root node
	 */
	private INode theRoot;

	/**
	 * Edge list
	 */
	private List<IEdge> theEdges;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 * 
	 * @param thisRoot
	 *        root node
	 * @param theseEdges
	 *        edge list
	 */
	public Tree(final INode thisRoot, final List<IEdge> theseEdges)
	{
		this.theRoot = thisRoot;
		this.theEdges = theseEdges;
	}

	// A C C E S S

	/**
	 * Get root node
	 * 
	 * @return root node
	 */
	public INode getRoot()
	{
		return this.theRoot;
	}

	/**
	 * Get edges
	 * 
	 * @return edges
	 */
	public List<IEdge> getEdges()
	{
		return this.theEdges;
	}

	/**
	 * Set root node
	 * 
	 * @param thisRoot
	 *        node
	 */
	public void setRoot(final INode thisRoot)
	{
		this.theRoot = thisRoot;
	}

	/**
	 * Set edges
	 * 
	 * @param theseEdges
	 *        edge list
	 */
	public void setEdges(final List<IEdge> theseEdges)
	{
		this.theEdges = theseEdges;
	}

	// S E A R C H I N G

	/**
	 * Find node nearest to location
	 * 
	 * @param thisStart
	 *        start node (only descendants are considered)
	 * @param thisLocation
	 *        unit circle location
	 * @return node if found, null otherwise
	 */
	public INode findNodeAt(final INode thisStart, final Complex thisLocation)
	{
		if (thisStart == null)
			return null;

		INode thisResult = thisStart;
		Location thisResultLocation = thisStart.getLocation();

		// find nearest (using squares as a measure)
		double thisDistance = EuclidianLocation.getDistanceSquare(thisResultLocation.hyper.center, thisLocation);
		final List<INode> theseChildren = thisStart.getChildren();
		if (theseChildren != null)
		{
			for (final INode thisChild : theseChildren)
			{
				final INode thisTargetNode = findNodeAt(thisChild, thisLocation);
				if (thisTargetNode != null)
				{
					final Location thisTargetLocation = thisTargetNode.getLocation();
					if (!thisTargetLocation.hyper.isBorder)
					{
						final double thatDistance = EuclidianLocation.getDistanceSquare(thisTargetLocation.hyper.center, thisLocation);
						if (thatDistance < thisDistance)
						{
							thisResult = thisTargetNode;
							thisResultLocation = thisTargetLocation;
							thisDistance = thatDistance;
						}
					}
				}
			}
		}

		if (thisResult != null && thisResultLocation.euclidian.radius * thisResultLocation.euclidian.radius <= thisDistance)
			return null;
		return thisResult;
	}

	/**
	 * Find node by id
	 * 
	 * @param thisStart
	 *        start node
	 * @param thisId
	 *        target id
	 * @return node if found, null otherwise
	 */
	public INode findNodeById(final INode thisStart, final String thisId)
	{
		if (thisStart == null)
			return null;

		// node test
		if (thisId.equals(thisStart.getId()))
			return thisStart;

		// children
		for (final INode thisChildNode : thisStart.getChildren())
		{
			final INode thisNode = findNodeById(thisChildNode, thisId);
			if (thisNode != null)
				return thisNode;
		}
		return null;
	}
}
