/**
 * Title: Treebolic DOT provider
 * Description: Treebolic DOT provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.dot;

import java.io.InputStream;
import java.net.URL;
import java.util.Hashtable;

import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.MutableEdge;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.Utils;
import treebolic.model.graph.Converter;
import treebolic.model.graph.Graph;
import treebolic.model.graph.GraphEdge;
import treebolic.model.graph.MutableGraph;
import treebolic.model.graph.MutableGraphNode;
import att.grappa.Attribute;
import att.grappa.GraphEnumeration;
import att.grappa.GrappaConstants;
import att.grappa.Parser;

/**
 * Dot parser
 * 
 * @author Bernard Bou
 */
public class DotParser
{
	/**
	 * Parse model from dot document url
	 * 
	 * @return model if successful
	 */
	static public Model parseModel(final URL thisUrl)
	{
		// parse dot
		final att.grappa.Graph thisDotGraph = DotParser.parseDotGraph(thisUrl);
		if (thisDotGraph == null)
			return null;

		final treebolic.model.Tree thisTree = DotParser.parseTree(thisDotGraph);
		final Settings theseSettings = DotParser.parseSettings(thisDotGraph);
		return thisTree != null ? new Model(thisTree, theseSettings) : null;
	}

	/**
	 * Parse tree from dot document url
	 * 
	 * @return tree if successful
	 */
	static public Tree parseTree(final URL thisUrl)
	{
		// parse dot
		final att.grappa.Graph thisDotGraph = DotParser.parseDotGraph(thisUrl);
		if (thisDotGraph == null)
			return null;
		return DotParser.parseTree(thisDotGraph);
	}

	// D O T P A R S I N G

	/**
	 * Parse dot graph
	 * 
	 * @param thisURL
	 *        dot file url
	 * @return dot graph
	 */
	static private att.grappa.Graph parseDotGraph(final URL thisURL)
	{
		try
		{
			// input stream
			final InputStream thisInputStream = thisURL.openStream();

			// make parser
			final Parser thisGrappaParser = new Parser(thisInputStream, System.err);

			// theGrappaParser.debug_parse(4);
			thisGrappaParser.parse();

			// graph
			return thisGrappaParser.getGraph();
		}
		catch (final Exception ex)
		{
			System.err.println("Dot graph parser: " + ex.getMessage());
			return null;
		}
	}

	/**
	 * Parse treebolic graph from dot graph
	 * 
	 * @return treebolic graph if successful
	 */
	static protected Graph parseGraph(final att.grappa.Graph thisDotGraph)
	{
		// treebolic graph
		final MutableGraph thisGraph = new MutableGraph();

		// nodes
		final Hashtable<String, MutableGraphNode> theNodesById = new Hashtable<String, MutableGraphNode>();
		for (final GraphEnumeration theseGraphNodes = thisDotGraph.elements(GrappaConstants.NODE); theseGraphNodes.hasMoreElements();)
		{
			final att.grappa.Node thisDotNode = (att.grappa.Node) theseGraphNodes.nextGraphElement();
			final String thisDotId = Integer.toString(thisDotNode.getId());
			final String thisId = thisDotNode.getName();

			// create node
			final MutableGraphNode thisNode = new MutableGraphNode(thisId);
			theNodesById.put(thisDotId, thisNode);
			thisGraph.add(thisNode);

			// label
			Attribute thisAttribute;
			thisAttribute = thisDotNode.getAttribute("label");
			if (thisAttribute != null)
			{
				thisNode.setLabel(thisAttribute.getStringValue());
			}

			// content
			thisAttribute = thisDotNode.getAttribute("content");
			if (thisAttribute != null)
			{
				thisNode.setContent(thisAttribute.getStringValue());
			}

			// content
			thisAttribute = thisDotNode.getAttribute("link_href");
			if (thisAttribute != null)
			{
				thisNode.setLink(thisAttribute.getStringValue());
			}

			// image
			thisAttribute = thisDotNode.getAttribute("img_src");
			if (thisAttribute != null)
			{
				thisNode.setImageFile(thisAttribute.getStringValue());
			}

			// mountpoint
			thisAttribute = thisDotNode.getAttribute("mountpoint_href");
			if (thisAttribute != null)
			{
				final MountPoint.Mounting thisMountPoint = new MountPoint.Mounting();
				thisMountPoint.theURL = thisAttribute.getStringValue();
				thisNode.setMountPoint(thisMountPoint);
			}

			// color
			thisAttribute = thisDotNode.getAttribute("backcolor");
			if (thisAttribute != null)
			{
				thisNode.setBackColor(Utils.stringToColor(thisAttribute.getStringValue()));
			}
			thisAttribute = thisDotNode.getAttribute("forecolor");
			if (thisAttribute != null)
			{
				thisNode.setForeColor(Utils.stringToColor(thisAttribute.getStringValue()));
			}
		}

		// edges
		for (final GraphEnumeration theseDotEdges = thisDotGraph.elements(GrappaConstants.EDGE); theseDotEdges.hasMoreElements();)
		{
			// get end ids
			final att.grappa.Edge thisDotEdge = (att.grappa.Edge) theseDotEdges.nextGraphElement();
			final String thisFromId = Integer.toString(thisDotEdge.getTail().getId());
			final String thisToId = Integer.toString(thisDotEdge.getHead().getId());

			// get end nodes
			final MutableGraphNode thisFromNode = theNodesById.get(thisFromId);
			if (thisFromNode == null)
			{
				System.err.println("Dot: Node not found with id=" + thisFromId);
				continue;
			}
			final MutableGraphNode thisToNode = theNodesById.get(thisToId);
			if (thisToNode == null)
			{
				System.err.println("Dot: Node not found with id=" + thisToId);
				continue;
			}

			Attribute thisAttribute;

			// whether this edge is tree edge
			Boolean isTreeEdge = null;
			thisAttribute = thisDotEdge.getAttribute("type");
			if (thisAttribute != null)
			{
				final String thisValue = thisAttribute.getStringValue();
				if (thisValue.equals("tree"))
				{
					isTreeEdge = true;
				}
				else if (thisValue.equals("nontree"))
				{
					isTreeEdge = false;
				}
			}

			// create edge
			final MutableEdge thisEdge = new MutableEdge(thisFromNode, thisToNode);
			final GraphEdge thisGraphEdge = new GraphEdge(thisFromNode, thisToNode, isTreeEdge);
			thisGraphEdge.setUserData(thisEdge);
			thisGraph.add(thisGraphEdge);

			// attributes
			thisAttribute = thisDotEdge.getAttribute("label");
			if (thisAttribute != null)
			{
				thisEdge.setLabel(thisAttribute.getStringValue());
			}
			thisAttribute = thisDotEdge.getAttribute("img_src");
			if (thisAttribute != null)
			{
				thisEdge.setImageFile(thisAttribute.getStringValue());
			}
			thisAttribute = thisDotEdge.getAttribute("edge_color");
			if (thisAttribute != null)
			{
				thisEdge.setColor(Utils.stringToColor(thisAttribute.getStringValue()));
			}
			thisEdge.setStyle(DotParser.getStyle(thisDotEdge.getAttribute("stroke"), thisDotEdge.getAttribute("fromterminator"), thisDotEdge.getAttribute("toterminator"), thisDotEdge.getAttribute("hidden")));
		}
		return thisGraph;
	}

	/**
	 * Parse tree from dot graph
	 * 
	 * @return tree if successful
	 */
	static private Tree parseTree(final att.grappa.Graph thisDotGraph)
	{
		final Graph thisGraph = DotParser.parseGraph(thisDotGraph);
		return new Converter<MutableGraphNode>().graphToTree(thisGraph);
	}

	/**
	 * Parse treebolic settings from dot file
	 * 
	 * @return treebolic settings if successful
	 */
	static protected Settings parseSettings(final att.grappa.Graph thisDotGraph)
	{
		final Settings theseSettings = new Settings();
		Attribute thisAttribute;

		// treebolic

		thisAttribute = thisDotGraph.getAttribute("treebolic_toolbar");
		if (thisAttribute != null)
		{
			theseSettings.theHasToolbarFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_statusbar");
		if (thisAttribute != null)
		{
			theseSettings.theHasStatusbarFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_popupmenu");
		if (thisAttribute != null)
		{
			theseSettings.theHasPopUpMenuFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_focus_on_hover");
		if (thisAttribute != null)
		{
			theseSettings.theFocusOnHoverFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_tooltip");
		if (thisAttribute != null)
		{
			theseSettings.theToolTipDisplaysContentFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_tooltip_displays_content");
		if (thisAttribute != null)
		{
			theseSettings.theHasToolTipFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_focus");
		if (thisAttribute != null)
		{
			theseSettings.theFocus = thisAttribute.getStringValue();
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_xmoveto");
		if (thisAttribute != null)
		{
			theseSettings.theXMoveTo = new Double(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_ymoveto");
		if (thisAttribute != null)
		{
			theseSettings.theYMoveTo = new Double(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_xshift");
		if (thisAttribute != null)
		{
			theseSettings.theXShift = new Double(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("treebolic_yshift");
		if (thisAttribute != null)
		{
			theseSettings.theYShift = new Double(thisAttribute.getStringValue());
		}

		// tree

		thisAttribute = thisDotGraph.getAttribute("tree_backcolor");
		if (thisAttribute != null)
		{
			theseSettings.theBackColor = Utils.stringToColor(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_forecolor");
		if (thisAttribute != null)
		{
			theseSettings.theForeColor = Utils.stringToColor(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_expansion");
		if (thisAttribute != null)
		{
			theseSettings.theExpansion = new Double(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_sweep");
		if (thisAttribute != null)
		{
			theseSettings.theSweep = new Double(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_orientation");
		if (thisAttribute != null)
		{
			theseSettings.theOrientation = thisAttribute.getStringValue();
		}

		thisAttribute = thisDotGraph.getAttribute("tree_preserve_orientation");
		if (thisAttribute != null)
		{
			theseSettings.thePreserveOrientationFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_fontface");
		if (thisAttribute != null)
		{
			theseSettings.theFontFace = thisAttribute.getStringValue();
		}

		thisAttribute = thisDotGraph.getAttribute("tree_fontsize");
		if (thisAttribute != null)
		{
			theseSettings.theFontSize = new Integer(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_fontsizestep");
		if (thisAttribute != null)
		{
			theseSettings.theFontSizeStep = new Integer(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("tree_img_src");
		if (thisAttribute != null)
		{
			theseSettings.theBackgroundImage = thisAttribute.getStringValue();
		}

		// nodes

		thisAttribute = thisDotGraph.getAttribute("nodes_backcolor");
		if (thisAttribute != null)
		{
			theseSettings.theNodeBackColor = Utils.stringToColor(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("nodes_forecolor");
		if (thisAttribute != null)
		{
			theseSettings.theNodeForeColor = Utils.stringToColor(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("nodes_img_src");
		if (thisAttribute != null)
		{
			theseSettings.theDefaultNodeImage = thisAttribute.getStringValue();
		}

		thisAttribute = thisDotGraph.getAttribute("nodes_default_treeedge_img_src");
		if (thisAttribute != null)
		{
			theseSettings.theDefaultTreeEdgeImage = thisAttribute.getStringValue();
		}

		thisAttribute = thisDotGraph.getAttribute("nodes_default_treeedge_color");
		if (thisAttribute != null)
		{
			theseSettings.theTreeEdgeColor = Utils.stringToColor(thisAttribute.getStringValue());
		}

		theseSettings.theTreeEdgeStyle = DotParser.getStyle(thisDotGraph.getAttribute("nodes_default_treeedge_stroke"), thisDotGraph.getAttribute("nodes_default_treeedge_fromterminator"), thisDotGraph.getAttribute("nodes_default_treeedge_toterminator"),
				thisDotGraph.getAttribute("nodes_default_treeedge_hidden"));

		// edges

		thisAttribute = thisDotGraph.getAttribute("edges_hyperbolic");
		if (thisAttribute != null)
		{
			theseSettings.theHyperbolicEdgesFlag = new Boolean(thisAttribute.getStringValue());
		}

		thisAttribute = thisDotGraph.getAttribute("edges_default_edge_img_src");
		if (thisAttribute != null)
		{
			theseSettings.theDefaultEdgeImage = thisAttribute.getStringValue();
		}

		thisAttribute = thisDotGraph.getAttribute("edges_default_edge_color");
		if (thisAttribute != null)
		{
			theseSettings.theEdgeColor = Utils.stringToColor(thisAttribute.getStringValue());
		}

		theseSettings.theEdgeStyle = DotParser.getStyle(thisDotGraph.getAttribute("edges_default_edge_stroke"), thisDotGraph.getAttribute("edges_default_edge_fromterminator"), thisDotGraph.getAttribute("edges_default_edge_toterminator"),
				thisDotGraph.getAttribute("edges_default_edge_hidden"));

		return theseSettings;
	}

	/**
	 * Style factory
	 * 
	 * @param thisStrokeAttribute
	 *        stroke attribute
	 * @param thisFromAttribute
	 *        from-terminator attribute
	 * @param thisToAttribute
	 *        to-terminator attribute
	 * @param thisHiddenAttribute
	 *        hidden attribute
	 * @return style
	 */
	static private Integer getStyle(final Attribute thisStrokeAttribute, final Attribute thisFromAttribute, final Attribute thisToAttribute, final Attribute thisHiddenAttribute)
	{
		String thisFromTerminator = null;
		String thisToTerminator = null;
		String thisStroke = null;
		String thisHiddenFlag = null;

		if (thisStrokeAttribute != null)
		{
			thisStroke = thisStrokeAttribute.getStringValue();
		}
		if (thisFromAttribute != null)
		{
			thisFromTerminator = thisFromAttribute.getStringValue();
		}
		if (thisToAttribute != null)
		{
			thisToTerminator = thisToAttribute.getStringValue();
		}
		if (thisHiddenAttribute != null)
		{
			thisHiddenFlag = thisHiddenAttribute.getStringValue();
		}
		return Utils.parseStyle(thisStroke, thisFromTerminator, thisToTerminator, thisHiddenFlag);
	}
}
