/**
 * Title: Treebolic GXL provider
 * Description: Treebolic GXL provider
 * Version:	2.x
 * Copyright: (c) 2001-2008
 * Terms of use: see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.gxl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import treebolic.model.IEdge;
import treebolic.model.INode;
import treebolic.model.MenuItem;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.MutableEdge;
import treebolic.model.MutableNode;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.Utils;
import treebolic.provider.xml.dom.Parser;

/**
 * GXL parser
 * 
 * @author Bernard Bou
 */
public class GxlParser
{
	/**
	 * Parse model
	 * 
	 * @return model if successful
	 */
	static public Model parseModel(URL thisUrl)
	{
		// DOM document
		Document thisDocument = new Parser().makeDocument(thisUrl, null);
		if (thisDocument == null)
			return null;

		Tree thisTree = parseTree(thisDocument);
		if (thisTree == null)
			return null;

		Settings theseSettings = parseSettings(thisDocument);
		return new Model(thisTree, theseSettings);
	}

	/**
	 * Parse tree
	 * 
	 * @return tree if successful
	 */
	static public Tree parseTree(URL thisUrl)
	{
		// DOM document
		Document thisDocument = new Parser().makeDocument(thisUrl, null);
		if (thisDocument == null)
			return null;

		return parseTree(thisDocument);
	}

	/**
	 * Parse tree
	 * 
	 * @return tree if successful
	 */
	static private Tree parseTree(Document thisDocument)
	{
		// map
		Hashtable<String, MutableNode> theseNodesById = new Hashtable<String, MutableNode>();

		// enumerate nodes
		NodeList theseNodeNodes = thisDocument.getElementsByTagName("node");
		for (int i = 0; i < theseNodeNodes.getLength(); i++)
		{
			Node thisNodeNode = theseNodeNodes.item(i);
			Element thisNodeElement = (Element) thisNodeNode;

			// get id
			String thisId = thisNodeElement.getAttribute("id");

			// create node
			MutableNode thisNode = new MutableNode(null, thisId);
			theseNodesById.put(thisId, thisNode);

			// attributes
			NodeList theseAttrNodes = thisNodeElement.getElementsByTagName("attr");
			for (int j = 0; j < theseAttrNodes.getLength(); j++)
			{
				Node thisAttrNode = theseAttrNodes.item(j);
				Element thisAttrElement = (Element) thisAttrNode;
				String thisAttributeName = thisAttrElement.getAttribute("name");

				if (thisAttributeName.equalsIgnoreCase("label"))
				{
					Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
						thisNode.setLabel(thisStringElement.getTextContent());
				}
				else if (thisAttributeName.equalsIgnoreCase("content"))
				{
					Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
						thisNode.setContent(thisStringElement.getTextContent());
				}
				else if (thisAttributeName.equalsIgnoreCase("link"))
				{
					Element thisLocatorElement = getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						thisNode.setLink(thisHRef);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("img-src"))
				{
					Element thisLocatorElement = getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						thisNode.setImageFile(thisHRef);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("mountpoint"))
				{
					Element thisLocatorElement = getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						MountPoint.Mounting thisMountPoint = new MountPoint.Mounting();
						thisMountPoint.theURL = thisHRef;
						thisNode.setMountPoint(thisMountPoint);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("mountpoint-weight"))
				{
					Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						MountPoint.Mounting thisMountPoint = (MountPoint.Mounting) thisNode.getMountPoint();
						if (thisMountPoint != null)
							thisNode.setWeight(Double.valueOf(thisStringElement.getTextContent()));
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("backcolor"))
				{
					Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
						thisNode.setBackColor(Utils.stringToColor(thisStringElement.getTextContent()));
				}
				else if (thisAttributeName.equalsIgnoreCase("forecolor"))
				{
					Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
						thisNode.setForeColor(Utils.stringToColor(thisStringElement.getTextContent()));
				}
			}
		}

		// enumerate edges
		List<IEdge> theseEdges = null;
		NodeList theseEdgeNodes = thisDocument.getElementsByTagName("edge");
		for (int i = 0; i < theseEdgeNodes.getLength(); i++)
		{
			Node thisEdgeNode = theseEdgeNodes.item(i);
			Element thisEdgeElement = (Element) thisEdgeNode;

			// get end ids
			String thisFromId = thisEdgeElement.getAttribute("from");
			if (thisFromId == null)
				continue;
			String thisToId = thisEdgeElement.getAttribute("to");
			if (thisToId == null)
				continue;

			// get ends
			MutableNode thisFromNode = theseNodesById.get(thisFromId);
			if (thisFromNode == null)
				continue;
			MutableNode thisToNode = theseNodesById.get(thisToId);
			if (thisToNode == null)
				continue;

			// type
			Element thisTypeElement = getFirstLevel1ElementByTagName(thisEdgeElement, "type");
			if (thisTypeElement != null)
			{
				String thisType = thisTypeElement.getAttribute("xlink:href");
				if (thisType != null && !thisType.isEmpty())
				{
					if (thisType.equals("schema.xml#TreeEdge"))
					{
						thisFromNode.getChildren().add(thisToNode);
						thisToNode.setParent(thisFromNode);

						// attributes
						NodeList theseAttrNodes = thisEdgeElement.getElementsByTagName("attr");
						for (int j = 0; j < theseAttrNodes.getLength(); j++)
						{
							Node thisAttrNode = theseAttrNodes.item(j);
							Element thisAttrElement = (Element) thisAttrNode;
							String thisAttributeName = thisAttrElement.getAttribute("name");
							if (thisAttributeName.equalsIgnoreCase("label"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
									thisToNode.setEdgeLabel(thisStringElement.getTextContent());
							}
							else if (thisAttributeName.equalsIgnoreCase("img-src"))
							{
								Element thisLocatorElement = getFirstLevel1ElementByTagName(thisAttrElement, "locator");
								if (thisLocatorElement != null)
								{
									String thisHRef = thisLocatorElement.getAttribute("xlink:href");
									thisToNode.setEdgeImageFile(thisHRef);
								}
							}
							else if (thisAttributeName.equalsIgnoreCase("stroke"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
								{
									Integer thisStyle = Utils.modifyStyle(thisToNode.getEdgeStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.STROKE);
									thisToNode.setEdgeStyle(thisStyle);
								}
							}
							else if (thisAttributeName.equalsIgnoreCase("fromterminator"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
								{
									Integer thisStyle = Utils.modifyStyle(thisToNode.getEdgeStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.FROMTERMINATOR);
									thisToNode.setEdgeStyle(thisStyle);
								}
							}
							else if (thisAttributeName.equalsIgnoreCase("toterminator"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
								{
									Integer thisStyle = Utils.modifyStyle(thisToNode.getEdgeStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.TOTERMINATOR);
									thisToNode.setEdgeStyle(thisStyle);
								}
							}
						}
					}
					else if (thisType.equals("schema.xml#NonTreeEdge"))
					{
						MutableEdge thisEdge = new MutableEdge(thisFromNode, thisToNode);
						if (theseEdges == null)
							theseEdges = new ArrayList<IEdge>();
						theseEdges.add(thisEdge);

						// attributes
						NodeList theseAttrNodes = thisEdgeElement.getElementsByTagName("attr");
						for (int j = 0; j < theseAttrNodes.getLength(); j++)
						{
							Node thisAttrNode = theseAttrNodes.item(j);
							Element thisAttrElement = (Element) thisAttrNode;
							String thisAttributeName = thisAttrElement.getAttribute("name");
							if (thisAttributeName.equalsIgnoreCase("label"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
									thisEdge.setLabel(thisStringElement.getTextContent());
							}
							else if (thisAttributeName.equalsIgnoreCase("img-src"))
							{
								Element thisLocatorElement = getFirstLevel1ElementByTagName(thisAttrElement, "locator");
								if (thisLocatorElement != null)
								{
									String thisHRef = thisLocatorElement.getAttribute("xlink:href");
									thisEdge.setImageFile(thisHRef);
								}
							}
							else if (thisAttributeName.equalsIgnoreCase("color"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
									thisEdge.setColor(Utils.stringToColor(thisStringElement.getTextContent()));
							}
							else if (thisAttributeName.equals("stroke"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
									thisEdge.setStyle(Utils.modifyStyle(thisEdge.getStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.STROKE));
							}
							else if (thisAttributeName.equals("fromterminator"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
									thisEdge.setStyle(Utils.modifyStyle(thisEdge.getStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.FROMTERMINATOR));
							}
							else if (thisAttributeName.equals("toterminator"))
							{
								Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
								if (thisStringElement != null)
									thisEdge.setStyle(Utils.modifyStyle(thisEdge.getStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.TOTERMINATOR));
							}
						}
					}
				}
			}
		}

		// get root : take node with 'root' id
		INode thisRoot = theseNodesById.get("root");
		if (thisRoot == null)
			return null;

		// return result
		return new Tree(thisRoot, theseEdges);
	}

	/**
	 * Parse
	 * 
	 * @return graph if successful
	 */
	static private Settings parseSettings(Document thisDocument)
	{
		Settings theseSettings = new Settings();
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theHasStatusbarFlag = true;

		Element thisGraphElement = getFirstElementByTagName(thisDocument.getDocumentElement(), "graph");
		if (thisGraphElement != null)
		{
			// enumerate attrs
			List<Element> theseAttrElements = getLevel1ChildElementsByTagName(thisGraphElement, "attr");
			for (Element thisAttrElement : theseAttrElements)
			{
				String thisName = thisAttrElement.getAttribute("name");

				// locator
				Element thisLocatorElement = getFirstLevel1ElementByTagName(thisAttrElement, "locator");
				if (thisLocatorElement != null)
				{
					String thisValue = thisLocatorElement.getAttribute("xlink:href");
					if (thisName.equals("nodes-img-src"))
						theseSettings.theDefaultNodeImage = thisValue;
					else if (thisName.equals("tree-img-src"))
						theseSettings.theBackgroundImage = thisValue;
					else if (thisName.equals("nodes-default-treeedge-img-src"))
						theseSettings.theDefaultTreeEdgeImage = thisValue;
					else if (thisName.equals("edges-default-edge-img-src"))
						theseSettings.theDefaultEdgeImage = thisValue;
				}

				// string
				Element thisStringElement = getFirstLevel1ElementByTagName(thisAttrElement, "string");
				if (thisStringElement != null)
				{
					String thisValue = thisStringElement.getTextContent();
					if (thisName.equals("treebolic-toolbar"))
						theseSettings.theHasToolbarFlag = Boolean.valueOf(thisValue);
					else if (thisName.equals("treebolic-statusbar"))
						theseSettings.theHasStatusbarFlag = Boolean.valueOf(thisValue);
					else if (thisName.equals("treebolic-popupmenu"))
						theseSettings.theHasPopUpMenuFlag = Boolean.valueOf(thisValue);
					else if (thisName.equals("treebolic-tooltip"))
						theseSettings.theHasToolTipFlag = Boolean.valueOf(thisValue);
					else if (thisName.equals("treebolic-focus-on-hover"))
						theseSettings.theFocusOnHoverFlag = Boolean.valueOf(thisValue);

					else if (thisName.equals("tree-backcolor"))
						theseSettings.theBackColor = Utils.stringToColor(thisValue);
					else if (thisName.equals("tree-forecolor"))
						theseSettings.theForeColor = Utils.stringToColor(thisValue);
					else if (thisName.equals("tree-fontface"))
						theseSettings.theFontFace = thisValue;
					else if (thisName.equals("tree-fontsize"))
						theseSettings.theFontSize = Integer.valueOf(thisValue);
					else if (thisName.equals("tree-fontsizestep"))
						theseSettings.theFontSizeStep = Integer.valueOf(thisValue);
					else if (thisName.equals("tree-expansion"))
						theseSettings.theExpansion = Double.valueOf(thisValue);
					else if (thisName.equals("tree-sweep"))
						theseSettings.theSweep = Double.valueOf(thisValue);
					else if (thisName.equals("tree-orientation"))
						theseSettings.theOrientation = thisValue;
					else if (thisName.equals("tree-preserve-orientation"))
						theseSettings.thePreserveOrientationFlag = Boolean.valueOf(thisValue);

					else if (thisName.equals("nodes-backcolor"))
						theseSettings.theNodeBackColor = Utils.stringToColor(thisValue);
					else if (thisName.equals("nodes-forecolor"))
						theseSettings.theNodeForeColor = Utils.stringToColor(thisValue);

					else if (thisName.equals("nodes-default-treeedge-color"))
						theseSettings.theTreeEdgeColor = Utils.stringToColor(thisValue);
					else if (thisName.equals("nodes-default-treeedge-stroke"))
						theseSettings.theTreeEdgeStyle = Utils.modifyStyle(theseSettings.theTreeEdgeStyle, thisValue, Utils.StyleComponent.STROKE);
					else if (thisName.equals("nodes-default-treeedge-fromterminator"))
						theseSettings.theTreeEdgeStyle = Utils.modifyStyle(theseSettings.theTreeEdgeStyle, thisValue, Utils.StyleComponent.FROMTERMINATOR);
					else if (thisName.equals("nodes-default-treeedge-toterminator"))
						theseSettings.theTreeEdgeStyle = Utils.modifyStyle(theseSettings.theTreeEdgeStyle, thisValue, Utils.StyleComponent.TOTERMINATOR);

					else if (thisName.equals("edges-hyperbolic"))
						theseSettings.theHyperbolicEdgesFlag = Boolean.valueOf(thisValue);
					else if (thisName.equals("edges-default-edge-color"))
						theseSettings.theEdgeColor = Utils.stringToColor(thisValue);
					else if (thisName.equals("edges-default-edge-fromterminator"))
						theseSettings.theEdgeStyle = Utils.modifyStyle(theseSettings.theEdgeStyle, thisValue, Utils.StyleComponent.FROMTERMINATOR);
					else if (thisName.equals("edges-default-edge-toterminator"))
						theseSettings.theEdgeStyle = Utils.modifyStyle(theseSettings.theEdgeStyle, thisValue, Utils.StyleComponent.TOTERMINATOR);
					else if (thisName.equals("edges-default-edge-stroke"))
						theseSettings.theEdgeStyle = Utils.modifyStyle(theseSettings.theEdgeStyle, thisValue, Utils.StyleComponent.STROKE);
					else
						System.err.println("Glx: Unhandled attribute: " + thisName);
				}

				// tuples
				Element thisTupElement = getFirstLevel1ElementByTagName(thisAttrElement, "tup");
				if (thisTupElement != null)
				{
					List<Element> theseTupleStringElements = getLevel1ChildElementsByTagName(thisTupElement, "string");
					if (theseTupleStringElements != null)
					{
						int k = 0;
						for (Element thisTupleStringElement : theseTupleStringElements)
						{
							if (thisTupleStringElement != null)
							{
								String thisValue = thisTupleStringElement.getTextContent();
								if (thisValue == null)
									continue;

								if (thisName.equals("menuitem-action"))
								{
									theseSettings.theMenu = allocate(theseSettings.theMenu, k);
									MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theAction = Utils.stringToAction(thisValue);
								}
								else if (thisName.equals("menuitem-label"))
								{
									theseSettings.theMenu = allocate(theseSettings.theMenu, k);
									MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theLabel = thisValue;

								}
								else if (thisName.equals("menuitem-match-target"))
								{
									theseSettings.theMenu = allocate(theseSettings.theMenu, k);
									MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theMatchTarget = thisValue;

								}
								else if (thisName.equals("menuitem-match-scope"))
								{
									theseSettings.theMenu = allocate(theseSettings.theMenu, k);
									MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theMatchScope = Utils.stringToScope(thisValue);
								}
								else if (thisName.equals("menuitem-match-mode"))
								{
									theseSettings.theMenu = allocate(theseSettings.theMenu, k);
									MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theMatchMode = Utils.stringToMode(thisValue);
								}
								k++;
							}
						}
					}
				}
			}
		}
		return theseSettings;
	}

	/**
	 * Allocate
	 * 
	 * @param thisList
	 *        menu item list
	 * @param i
	 *        index
	 * @return same list with ith slot having an allocated menu item
	 */
	static private List<MenuItem> allocate(List<MenuItem> thisList, int i)
	{
		if (thisList == null)
			thisList = new ArrayList<MenuItem>();
		if (i > thisList.size() - 1)
			thisList.add(new MenuItem());
		return thisList;
	}

	// H E L P E R S

	/**
	 * Find DOM element with given tag
	 * 
	 * @param thisElement
	 *        starting DOM element
	 * @param thisTagName
	 *        tag
	 * @return DOM element if found, null if none
	 */
	static private Element getFirstElementByTagName(Element thisElement, String thisTagName)
	{
		NodeList thisList = thisElement.getElementsByTagName(thisTagName);
		if (thisList.getLength() > 0)
			return (Element) thisList.item(0);
		return null;
	}

	/**
	 * Find DOM element with given tag among first level children
	 * 
	 * @param thisElement
	 *        starting DOM element
	 * @param thisTagName
	 *        tag
	 * @return DOM element if found, null if none
	 */
	static private Element getFirstLevel1ElementByTagName(Element thisElement, String thisTagName)
	{
		List<Element> theseChildElements = getLevel1ChildElementsByTagName(thisElement, thisTagName);
		if (!theseChildElements.isEmpty())
			return theseChildElements.get(0);
		return null;
	}

	/**
	 * Find DOM elements with given tag among first level children
	 * 
	 * @param thisElement
	 *        starting DOM element
	 * @param thisTagName
	 *        tag
	 * @return DOM element if found, null if none
	 */
	static private List<Element> getLevel1ChildElementsByTagName(Element thisElement, String thisTagName)
	{
		ArrayList<Element> thisList = new ArrayList<Element>();
		NodeList theseChildren = thisElement.getChildNodes();
		for (int i = 0; i < theseChildren.getLength(); i++)
		{
			Node thisNode = theseChildren.item(i);
			if (thisNode instanceof Element)
			{
				Element thisChildElement = (Element) thisNode;
				if (thisChildElement.getTagName().equals(thisTagName))
					thisList.add(thisChildElement);
			}
		}
		return thisList;
	}
}
