/**
 * Title: Treebolic OWL provider
 * Description: Treebolic OWL provider
 * Version:	2.x
 * Copyright: (c) 2001-2008
 * Terms of use: see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.owl;

import java.awt.Color;
import java.awt.Image;
import java.util.Properties;

import javax.swing.ImageIcon;

import treebolic.model.IEdge;
import treebolic.model.INode;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.MutableNode;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.Utils;

import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.ontology.OntProperty;
import com.hp.hpl.jena.ontology.OntResource;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.util.LocationMapper;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

/**
 * OWL parser
 * 
 * @author Bernard Bou
 */
public class OwlParser
{
	/**
	 * Whether to include individuals
	 */
	static boolean includeInstances = true;

	/**
	 * As tree
	 */
	static boolean asTree = false;

	/**
	 * Default class color
	 */
	static Color theDefaultClassBackColor = new Color(0xffffC0);

	/**
	 * Focus image
	 */
	private static final Image theRootImage = new ImageIcon(Provider.class.getResource("images/focus.png")).getImage();

	/**
	 * Class image
	 */
	private static final Image theClassImage = new ImageIcon(Provider.class.getResource("images/class.png")).getImage();

	/**
	 * Class with property image
	 */
	private static final Image theClassWithPropertiesImage = new ImageIcon(Provider.class.getResource("images/classwithprops.png")).getImage();

	/**
	 * Properties image
	 */
	private static final Image thePropertiesImage = new ImageIcon(Provider.class.getResource("images/properties.png")).getImage();

	/**
	 * Property image
	 */
	private static final Image thePropertyImage = new ImageIcon(Provider.class.getResource("images/property.png")).getImage();

	/**
	 * Instance image
	 */
	private static final Image theInstanceImage = new ImageIcon(Provider.class.getResource("images/instance.png")).getImage();

	/**
	 * Instance image
	 */
	private static final Image theInstanceClassImage = new ImageIcon(Provider.class.getResource("images/instanceclass.png")).getImage();

	/**
	 * Properties
	 */
	private final Properties theProperties;

	// CLASSES

	/**
	 * Class back color
	 */
	private Color theClassBackColor;

	/**
	 * Class fore color
	 */
	private Color theClassForeColor;

	/**
	 * Class image
	 */
	private String theClassImageFile;

	/**
	 * Class fore color
	 */
	private Color theClassWithPropertiesBackColor;

	/**
	 * Class fore color
	 */
	private Color theClassWithPropertiesForeColor;

	/**
	 * Class image
	 */
	private String theClassWithPropertiesImageFile;

	// ROOT

	/**
	 * Root label
	 */
	private String theRootLabel;

	/**
	 * Root fore color
	 */
	private Color theRootForeColor;

	/**
	 * Root fore color
	 */
	private Color theRootBackColor;

	/**
	 * Root image
	 */
	private String theRootImageFile;

	// PROPERTIES

	/**
	 * Properties label
	 */
	private String thePropertiesLabel;

	/**
	 * Properties fore color
	 */
	private Color thePropertiesForeColor;

	/**
	 * Properties fore color
	 */
	private Color thePropertiesBackColor;

	/**
	 * Properties image
	 */
	private String thePropertiesImageFile;

	// PROPERTY

	/**
	 * Property fore color
	 */
	private Color thePropertyForeColor;

	/**
	 * Property fore color
	 */
	private Color thePropertyBackColor;

	/**
	 * Property image
	 */
	private String thePropertyImageFile;

	/**
	 * Property edge color
	 */
	static private Color thePropertyEdgeColor;

	/**
	 * Property edge style
	 */
	static private Integer thePropertyEdgeStyle;

	/**
	 * Property edge image file
	 */
	static private String thePropertyEdgeImageFile;

	// INSTANCE

	/**
	 * Instance foreground color
	 */
	private Color theInstanceForeColor;

	/**
	 * Instance background color
	 */
	private Color theInstanceBackColor;

	/**
	 * Instance image file
	 */
	private String theInstanceImageFile;

	/**
	 * Instance edge style
	 */
	static private Integer theInstanceEdgeStyle;

	/**
	 * Instance edge color
	 */
	static private Color theInstanceEdgeColor;

	/**
	 * Instance image file
	 */
	static private String theInstanceEdgeImageFile;

	// INSTANCE CLASS

	/**
	 * InstanceClass foreground color
	 */
	private Color theInstanceClassForeColor;

	/**
	 * InstanceClass background color
	 */
	private Color theInstanceClassBackColor;

	/**
	 * InstanceClass image file
	 */
	private String theInstanceClassImageFile;

	// C O N S T R U C T O R

	public OwlParser(final Properties theseProperties)
	{
		this.theProperties = theseProperties;
	}

	/**
	 * Initialize from properties
	 */
	void initialize()
	{
		OwlParser.includeInstances = getBoolean("instances", false);

		this.theRootBackColor = getColor("root.backcolor", Color.ORANGE);
		this.theRootForeColor = getColor("root.forecolor", Color.BLACK);
		this.theRootImageFile = getImageFile("root.image");
		this.theRootLabel = getLabel("root.label", "Root");

		this.theClassBackColor = getColor("class.backcolor", OwlParser.theDefaultClassBackColor);
		this.theClassForeColor = getColor("class.forecolor", Color.BLACK);
		this.theClassImageFile = getImageFile("class.image");

		this.theClassWithPropertiesBackColor = getColor("class.withprops.backcolor", OwlParser.theDefaultClassBackColor);
		this.theClassWithPropertiesForeColor = getColor("class.withprops.forecolor", Color.BLUE);
		this.theClassWithPropertiesImageFile = getImageFile("class.withprops.image");

		this.thePropertiesLabel = getLabel("properties.label", "properties");
		this.thePropertiesBackColor = getColor("properties.backcolor", Color.WHITE);
		this.thePropertiesForeColor = getColor("properties.forecolor", Color.GRAY);
		this.thePropertiesImageFile = getImageFile("properties.image");

		this.thePropertyBackColor = getColor("property.backcolor", Color.WHITE);
		this.thePropertyForeColor = getColor("property.forecolor", Color.DARK_GRAY);
		this.thePropertyImageFile = getImageFile("property.image");

		this.theInstanceBackColor = getColor("instance.backcolor", Color.WHITE);
		this.theInstanceForeColor = getColor("instance.forecolor", Color.DARK_GRAY);
		this.theInstanceImageFile = getImageFile("instance.image");

		this.theInstanceClassBackColor = getColor("instance.class.backcolor", OwlParser.theDefaultClassBackColor.darker());
		this.theInstanceClassForeColor = getColor("instance.class.forecolor", Color.BLACK);
		this.theInstanceClassImageFile = getImageFile("instance.class.image");
	}

	// P A R S E

	/**
	 * Parse model
	 * 
	 * @return model if successful
	 */
	public Model parseModel(final String thisUrlString)
	{
		final Tree thisTree = parseTree(thisUrlString);
		if (thisTree == null)
			return null;

		final Settings theSettings = new Settings();
		theSettings.theHasToolbarFlag = true;
		theSettings.theHasStatusbarFlag = true;
		theSettings.theOrientation = OwlParser.asTree ? "south" : "radial";
		theSettings.theHasToolbarFlag = true;
		theSettings.theBackColor = new Color(0xffffe0);
		theSettings.theFontFace = "Dialog";
		theSettings.theFontSize = 20;
		theSettings.theFontSizeStep = 2;
		theSettings.theExpansion = .9;
		theSettings.theSweep = 1.2;
		theSettings.theHasStatusbarFlag = true;
		theSettings.theFocusOnHoverFlag = false;
		theSettings.theTreeEdgeColor = Color.BLACK;
		theSettings.theTreeEdgeStyle = IEdge.SOLID | IEdge.TOTRIANGLE | IEdge.TOSOLID;
		if (OwlParser.asTree)
		{
			theSettings.theYMoveTo = -0.3;
		}

		// override
		if (this.theProperties != null)
		{
			try
			{
				theSettings.load(this.theProperties);
			}
			catch (final Exception thisException)
			{
				System.err.println("SETTING " + thisException.toString());
			}
		}

		// cache property features from settings
		OwlParser.thePropertyEdgeStyle = theSettings.theEdgeStyle;
		OwlParser.thePropertyEdgeColor = theSettings.theEdgeColor;
		OwlParser.thePropertyEdgeImageFile = theSettings.theDefaultEdgeImage;

		OwlParser.theInstanceEdgeStyle = theSettings.theTreeEdgeStyle;
		OwlParser.theInstanceEdgeColor = theSettings.theTreeEdgeColor;
		OwlParser.theInstanceEdgeImageFile = theSettings.theDefaultTreeEdgeImage;

		return new Model(thisTree, theSettings);
	}

	/**
	 * Parse tree
	 * 
	 * @param thatUrlString
	 *        url string
	 * @return tree if successful
	 */
	public Tree parseTree(final String thatUrlString)
	{
		String thisUrlString = thatUrlString;

		initialize();

		// LocationMapper
		LocationMapper.setGlobalLocationMapper(new LocationMapper());

		// ontology model
		final OntModelSpec thisSpec = new OntModelSpec(OntModelSpec.OWL_DL_MEM);
		final OntModel thisModel = ModelFactory.createOntologyModel(thisSpec, null);

		// parameter and url
		String theseParameters = null;
		final int thisIndex = thisUrlString.indexOf('?');
		if (thisIndex != -1)
		{
			theseParameters = thisUrlString.substring(thisIndex + 1);
			thisUrlString = thisUrlString.substring(0, thisIndex);
		}

		// load document
		try
		{
			thisModel.read(thisUrlString);
		}
		catch (final Exception e)
		{
			System.err.println("Owl:" + e.toString());
			return null;
		}

		// root
		final MutableNode thisRoot = new MutableNode(null, "root");

		if (theseParameters != null)
		{
			String thisClassUri = null;
			String thisTarget = null;
			final String[] theseFields = theseParameters.split("&");
			for (final String thisField : theseFields)
			{
				final String[] thisNameValue = thisField.split("=");
				if (thisNameValue.length > 1)
				{
					if (thisNameValue[0].equals("class"))
					{
						thisClassUri = thisNameValue[1];
					}
					else if (thisNameValue[0].equals("target"))
					{
						thisTarget = thisNameValue[1];
					}
				}
			}
			if (thisClassUri != null)
			{
				if (thisTarget != null)
				{
					if (thisTarget.equals("properties"))
					{
						// root
						thisRoot.setLabel(OwlParser.getName(thisClassUri));
						thisRoot.setBackColor(this.theClassWithPropertiesBackColor);
						thisRoot.setForeColor(this.theClassWithPropertiesForeColor);
						setImage(thisRoot, this.theClassWithPropertiesImageFile, OwlParser.theClassWithPropertiesImage);

						// properties root
						final MutableNode thisPropertiesNode = new MutableNode(thisRoot, thisClassUri + "-prop");
						thisPropertiesNode.setLabel(this.thePropertiesLabel);
						thisPropertiesNode.setBackColor(this.thePropertiesBackColor);
						thisPropertiesNode.setForeColor(this.thePropertiesForeColor);
						setImage(thisPropertiesNode, this.thePropertiesImageFile, OwlParser.thePropertiesImage);

						// properties
						final OntClass thisClass = thisModel.getOntClass(thisClassUri);
						if (thisClass != null)
						{
							walkProperties(thisPropertiesNode, thisClass.listDeclaredProperties(true));
							return new Tree(thisRoot, null);
						}
					}
				}
			}
			return null;
		}
		else
		{
			thisRoot.setLabel(this.theRootLabel);
			thisRoot.setBackColor(this.theRootBackColor);
			thisRoot.setForeColor(this.theRootForeColor);
			setImage(thisRoot, this.theRootImageFile, OwlParser.theRootImage);

			// walk classes
			final ExtendedIterator<OntClass> theseRootClasses = thisModel.listHierarchyRootClasses();
			walkClasses(thisRoot, theseRootClasses, thisUrlString);
		}
		return new Tree(thisRoot, null);
	}

	/**
	 * Walk classes in iterator
	 * 
	 * @param thisParentClassNode
	 *        treebolic parent node to attach to
	 * @param theseClasses
	 *        class iterator
	 * @param thisUrlString
	 *        URL string
	 */
	public void walkClasses(final INode thisParentClassNode, final ExtendedIterator<OntClass> theseClasses, final String thisUrlString)
	{
		for (final OntClass thisClass : theseClasses.toList())
		{
			if (thisClass.isIndividual() && !thisClass.hasSuperClass())
			{
				continue;
			}
			walkClass(thisParentClassNode, thisClass, thisUrlString);
		}
	}

	/**
	 * Walk class in iterator
	 * 
	 * @param thisParentClassNode
	 *        treebolic parent node to attach to
	 * @param thisClass
	 *        ontology class
	 * @param thisUrlString
	 *        URL string
	 */
	public void walkClass(final INode thisParentClassNode, final OntClass thisClass, final String thisUrlString)
	{
		final String thisUri = thisClass.getURI();
		if (thisUri == null)
			return;

		final String thisId = OwlParser.getName(thisUri);

		// node
		final MutableNode thisClassNode = new MutableNode(thisParentClassNode, thisId);
		thisClassNode.setLabel(thisId);
		thisClassNode.setContent(thisUri + "\n" + thisClass.getComment(null));
		thisClassNode.setBackColor(this.theClassBackColor);
		thisClassNode.setForeColor(this.theClassForeColor);
		setImage(thisClassNode, this.theClassImageFile, OwlParser.theClassImage);

		// individuals
		walkInstances(thisClassNode, thisClass, thisUrlString);

		// node mountpoint
		if (thisClass.listDeclaredProperties(true).hasNext())
		{
			final MountPoint.Mounting thisMountingPoint = new MountPoint.Mounting();
			thisMountingPoint.theURL = thisUrlString + "?class=" + thisUri + "&target=properties";
			thisClassNode.setMountPoint(thisMountingPoint);
		}

		// recurse
		final ExtendedIterator<OntClass> theseSubClasses = thisClass.listSubClasses(true);
		walkClasses(thisClassNode, theseSubClasses, thisUrlString);
	}

	/**
	 * Walk individuals in iterator
	 * 
	 * @param thisParentNode
	 *        treebolic parent node to attach to
	 * @param thisClass
	 *        ontology class
	 */
	public void walkInstances(final INode thisParentNode, final OntClass thisClass, final String thisUrlString)
	{
		if (!OwlParser.includeInstances)
			return;

		final ExtendedIterator<? extends OntResource> theseInstances = thisClass.listInstances(true);
		for (final OntResource thisInstance : theseInstances.toList())
		{
			final String thisUri = thisInstance.getURI();
			final String thisId = OwlParser.getName(thisUri);

			final MutableNode thisInstanceNode = new MutableNode(thisParentNode, thisId);
			thisInstanceNode.setLabel(thisId);
			thisInstanceNode.setContent(thisUri + "\n" + thisInstance.getComment(null));
			thisInstanceNode.setForeColor(this.theInstanceForeColor);
			thisInstanceNode.setBackColor(this.theInstanceBackColor);
			setImage(thisInstanceNode, this.theInstanceImageFile, OwlParser.theInstanceImage);
			thisInstanceNode.setEdgeStyle(OwlParser.theInstanceEdgeStyle);
			thisInstanceNode.setEdgeColor(OwlParser.theInstanceEdgeColor);
			thisInstanceNode.setEdgeImageFile(OwlParser.theInstanceEdgeImageFile);

			// recurse
			if (thisInstance.isClass())
			{
				thisInstanceNode.setBackColor(this.theInstanceClassBackColor);
				thisInstanceNode.setForeColor(this.theInstanceClassForeColor);
				setImage(thisInstanceNode, this.theInstanceClassImageFile, OwlParser.theInstanceClassImage);

				final OntClass thisInstanceClass = thisInstance.asClass();
				final ExtendedIterator<OntClass> theseSubClasses = thisInstanceClass.listSubClasses(true);
				walkClasses(thisInstanceNode, theseSubClasses, thisUrlString);
				walkInstances(thisInstanceNode, thisInstanceClass, thisUrlString);
			}
		}
	}

	/**
	 * Walk properties in iterator
	 * 
	 * @param thisParentNode
	 *        treebolic parent node to attach to
	 * @param theseProperties
	 *        property iterator
	 */
	public void walkProperties(final INode thisParentNode, final ExtendedIterator<? extends OntProperty> theseProperties)
	{
		for (final OntProperty thisProperty : theseProperties.toList())
		{
			final String thisUri = thisProperty.getURI();
			if (thisUri == null)
			{
				continue;
			}

			final String thisId = OwlParser.getName(thisUri);

			// node
			final MutableNode thisPropertyNode = new MutableNode(thisParentNode, thisId);
			thisPropertyNode.setLabel(thisId);
			thisPropertyNode.setForeColor(this.thePropertyForeColor);
			thisPropertyNode.setBackColor(this.thePropertyBackColor);
			setImage(thisPropertyNode, this.thePropertyImageFile, OwlParser.thePropertyImage);
			thisPropertyNode.setContent(thisUri + "\n" + thisProperty.getComment(null));
			thisPropertyNode.setEdgeStyle(OwlParser.thePropertyEdgeStyle);
			thisPropertyNode.setEdgeColor(OwlParser.thePropertyEdgeColor);
			thisPropertyNode.setEdgeImageFile(OwlParser.thePropertyEdgeImageFile);

			// recurse
			final ExtendedIterator<? extends OntProperty> theseSubProperties = thisProperty.listSubProperties(true);
			walkProperties(thisPropertyNode, theseSubProperties);
		}
	}

	/**
	 * Get boolean
	 * 
	 * @param thisBooleanKey
	 *        boolean key
	 * @param thisBooleanValue
	 *        boolean value
	 */
	private boolean getBoolean(final String thisBooleanKey, final boolean thisBooleanValue)
	{
		return this.theProperties == null ? thisBooleanValue : Boolean.parseBoolean(this.theProperties.getProperty(thisBooleanKey, Boolean.toString(thisBooleanValue)));
	}

	/**
	 * Get node label
	 * 
	 * @param thisLabelKey
	 *        label key
	 * @param thisLabelValue
	 *        label value
	 */
	private String getLabel(final String thisLabelKey, final String thisLabelValue)
	{
		return this.theProperties == null ? thisLabelValue : this.theProperties.getProperty(thisLabelKey, thisLabelValue);
	}

	/**
	 * Get node color
	 * 
	 * @param thisColorKey
	 *        forecolor key
	 * @param thisColorValue
	 *        forecolor value
	 * @return color
	 */
	private Color getColor(final String thisColorKey, final Color thisColorValue)
	{
		final String thisColorString = this.theProperties == null ? null : this.theProperties.getProperty(thisColorKey);
		return thisColorString == null ? thisColorValue : Utils.stringToColor(thisColorString);
	}

	/**
	 * Set node image file
	 * 
	 * @param thisImageKey
	 *        image key
	 * @return image file or null
	 */
	private String getImageFile(final String thisImageKey)
	{
		return this.theProperties == null ? null : this.theProperties.getProperty(thisImageKey);
	}

	/**
	 * Set node image
	 * 
	 * @param thisNode
	 *        node
	 * @param thisImageFile
	 *        image file (can be null to default)
	 * @param thisImageDefault
	 *        image value
	 */
	private void setImage(final MutableNode thisNode, final String thisImageFile, final Image thisImageDefault)
	{
		if (thisImageFile != null)
		{
			thisNode.setImageFile(thisImageFile);
		}
		else
		{
			thisNode.setImage(thisImageDefault);
		}
	}

	// N A M E

	/**
	 * Get name
	 * 
	 * @param thisUriString
	 *        uri string
	 * @return name
	 */
	static public String getName(final String thisUriString)
	{
		if (thisUriString != null)
		{
			final int thisIndex = thisUriString.lastIndexOf('#');
			if (thisIndex != -1)
				return thisUriString.substring(thisIndex + 1);
		}
		return thisUriString;
	}
}
