/**
 * Title: Treebolic SQL provider
 * Description: Treebolic SQL provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.sql;

import java.awt.Color;
import java.net.URL;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import treebolic.model.IEdge;
import treebolic.model.INode;
import treebolic.model.IProvider;
import treebolic.model.IProviderContext;
import treebolic.model.MenuItem;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.MutableEdge;
import treebolic.model.MutableNode;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.TreeMutableNode;
import treebolic.model.Utils;

/**
 * Provider for SQL
 * 
 * @author Bernard Bou
 */
public class Provider implements IProvider
{
	// S T A T I C S

	/**
	 * Property files
	 */
	static private String thePropertyFile = "TreebolicSql.properties";

	/**
	 * Get nodes SQL statement
	 */
	static private String theDefaultNodesSql = "SELECT * FROM %nodes%;";

	/**
	 * Get tree edges SQL statement
	 */
	static private String theDefaultTreeEdgesSql = "SELECT * FROM %edges% WHERE %istree% = TRUE;";

	/**
	 * Get edges SQL statement
	 */
	static private String theDefaultEdgesSql = "SELECT * FROM %edges% WHERE %istree% = FALSE;";

	/**
	 * Get settings SQL statement
	 */
	static private String theDefaultSettingsSql = "SELECT * FROM %settings%;";

	/**
	 * Get menu item SQL statement
	 */
	static private String theDefaultMenuSql = "SELECT * FROM %menu% WHERE %menuid% = 0;";

	// D A T A

	/**
	 * Provider context
	 */
	private IProviderContext theContext;

	/**
	 * Node map by id
	 */
	private final Hashtable<String, TreeMutableNode> theIdToNodeMap = new Hashtable<String, TreeMutableNode>();

	/**
	 * Properties
	 */
	protected Properties theProperties;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 */
	public Provider()
	{
		// do nothing
	}

	/**
	 * Initialize
	 * 
	 * @param thisSource
	 *        source
	 * @return true if successful
	 */
	private boolean initialize(final String thisSource)
	{
		// get property file
		String thisSqlFile = thisSource;
		if (thisSqlFile == null)
		{
			thisSqlFile = this.theContext.getParameter("sql");
		}
		if (thisSqlFile == null)
		{
			thisSqlFile = Provider.thePropertyFile;
		}

		// parse property file
		if (thisSqlFile != null)
		{
			// give the context a chance
			final URL thisUrl = this.theContext.makeURL(thisSqlFile);
			if (thisUrl == null)
				return false;

			this.theProperties = SqlProperties.load(thisUrl);
			if (this.theProperties != null)
				return true;
		}
		return false;
	}

	// I N T E R F A C E

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#set(treebolic.model.IProviderContext)
	 */
	public void connect(final IProviderContext thisContext)
	{
		this.theContext = thisContext;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#makeModel(java.lang.String)
	 */
	public Model makeModel(final String thisSource)
	{
		if (initialize(thisSource))
			return queryModel();
		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#makeTree(java.lang.String, boolean)
	 */
	public Tree makeTree(final String thisSource, final boolean checkRecursion)
	{
		if (initialize(thisSource))
			return queryTree();
		return null;
	}

	// H E L P E R

	/**
	 * Make URL
	 * 
	 * @param theseProperties
	 *        properties
	 * @return url string
	 */
	private String makeUrl(final Properties theseProperties)
	{
		String thisURL = theseProperties.getProperty("url");
		if (thisURL == null || thisURL.isEmpty())
		{
			final String thisProtocol = theseProperties.getProperty("protocol");
			String thisServer = theseProperties.getProperty("server");
			if (thisServer == null || thisServer.isEmpty())
			{
				thisServer = "localhost";
			}
			final String thisDatabase = theseProperties.getProperty("db");
			thisURL = thisProtocol + "://" + thisServer + "/" + thisDatabase;
		}
		return thisURL;
	}

	// Q U E R Y

	/**
	 * Query model
	 * 
	 * @return model
	 */
	private Model queryModel()
	{
		// url/user/password
		final String thisUrl = makeUrl(this.theProperties);
		final String thisUser = this.theProperties.getProperty("user");
		final String thisPasswd = this.theProperties.getProperty("passwd");

		try
		{
			// connect
			final Connection thisConnection = DriverManager.getConnection(thisUrl, thisUser, thisPasswd);

			// tree
			final Tree thisTree = queryTree(thisConnection);

			// settings
			final Settings theseSettings = querySettings(thisConnection);

			// close connection
			thisConnection.close();

			// result
			if (thisTree == null)
				return null;
			return new Model(thisTree, theseSettings);
		}
		catch (final SQLException thisException)
		{
			System.err.println("Sql: " + thisException.getMessage());
		}
		return null;
	}

	/**
	 * Query tree
	 * 
	 * @return tree
	 */
	private Tree queryTree()
	{
		// url/user/password
		final String thisUrl = makeUrl(this.theProperties);
		final String thisUser = this.theProperties.getProperty("user");
		final String thisPasswd = this.theProperties.getProperty("passwd");

		try
		{
			// connect
			final Connection thisConnection = DriverManager.getConnection(thisUrl, thisUser, thisPasswd);

			// tree
			final Tree thisTree = queryTree(thisConnection);

			// close connection
			thisConnection.close();

			// result
			return thisTree;
		}
		catch (final SQLException thisException)
		{
			System.err.println("Sql: " + thisException.getMessage());
		}
		return null;
	}

	/**
	 * Query tree
	 * 
	 * @param thisConnection
	 *        connection
	 * @return tree
	 * @throws SQLException
	 */
	private Tree queryTree(final Connection thisConnection) throws SQLException
	{
		final INode thisRoot = queryNodes(thisConnection);
		final List<IEdge> theseEdges = queryEdges(thisConnection);
		return new Tree(thisRoot, theseEdges);
	}

	/**
	 * Query nodes
	 * 
	 * @param thisConnection
	 *        connection
	 * @return tree root node
	 * @throws SQLException
	 */
	private MutableNode queryNodes(final Connection thisConnection) throws SQLException
	{
		// sql
		String thisNodesSql = this.theProperties.getProperty("nodesSql", Provider.theDefaultNodesSql);
		thisNodesSql = macroExpand(thisNodesSql);

		String thisTreeEdgesSql = this.theProperties.getProperty("treeEdgesSql", Provider.theDefaultTreeEdgesSql);
		thisTreeEdgesSql = macroExpand(thisTreeEdgesSql);

		// field names
		final String thisIdName = getName("nodes.id");
		final String thisLabelName = getName("nodes.label");
		final String thisContentName = getName("nodes.content");
		final String thisBackcolorName = getName("nodes.backcolor");
		final String thisForecolorName = getName("nodes.forecolor");
		final String thisImageName = getName("nodes.image");
		final String thisLinkName = getName("nodes.link");
		final String thisTargetName = getName("nodes.target");
		final String thisWeightName = getName("nodes.weight");
		final String thisMountpointName = getName("nodes.mountpoint");

		final String thisFromName = getName("edges.from");
		final String thisToName = getName("edges.to");
		final String thisEdgeLabelName = getName("edges.label");
		final String thisEdgeImageName = getName("edges.image");
		final String thisEdgeColorName = getName("edges.color");
		final String thisEdgeHiddenName = getName("edges.hidden");
		final String thisEdgeStrokeName = getName("edges.stroke");
		final String thisEdgeFromTerminatorName = getName("edges.fromterminator");
		final String thisEdgeToTerminatorName = getName("edges.toterminator");

		// id to node map
		this.theIdToNodeMap.clear();

		// N O D E S
		final Statement thisNodesStatement = thisConnection.createStatement();
		final ResultSet thisNodesResultSet = thisNodesStatement.executeQuery(thisNodesSql);
		while (thisNodesResultSet.next())
		{
			// read id
			final String thisId = thisNodesResultSet.getString(thisIdName);
			if (thisId == null)
			{
				System.err.println("Sql: " + thisNodesResultSet.getRow() + "th edge record with null id");
				continue;
			}

			// make node
			final TreeMutableNode thisNode = new TreeMutableNode(null, thisId);
			this.theIdToNodeMap.put(thisId, thisNode);

			// data
			thisNode.setLabel(readString(thisNodesResultSet, thisLabelName));
			thisNode.setContent(readString(thisNodesResultSet, thisContentName));
			thisNode.setBackColor(readColor(thisNodesResultSet, thisBackcolorName));
			thisNode.setForeColor(readColor(thisNodesResultSet, thisForecolorName));
			thisNode.setImageFile(readString(thisNodesResultSet, thisImageName));
			thisNode.setLink(readString(thisNodesResultSet, thisLinkName));
			thisNode.setTarget(readString(thisNodesResultSet, thisTargetName));
			final Double thisWeight = readDouble(thisNodesResultSet, thisWeightName);
			if (thisWeight != null)
			{
				thisNode.setWeight(-thisWeight);
			}

			// mountpoint
			final String thisValue = readString(thisNodesResultSet, thisMountpointName);
			if (thisValue != null)
			{
				final MountPoint.Mounting thisMountPoint = new MountPoint.Mounting();
				thisMountPoint.theURL = thisValue;
				thisNode.setMountPoint(thisMountPoint);
				final Boolean thisBooleanValue = readBoolean(thisNodesResultSet, "now");
				if (thisBooleanValue != null && thisBooleanValue)
				{
					System.err.println("SQL: Recursive mounting not implemented");
				}
			}
		}
		thisNodesResultSet.close();
		thisNodesStatement.close();

		// T R E E . E D G E S
		final List<TreeMutableNode> theseParentLessNodes = new ArrayList<TreeMutableNode>();
		final Statement thisTreeEdgesStatement = thisConnection.createStatement();
		final ResultSet thisTreeEdgesResultSet = thisTreeEdgesStatement.executeQuery(thisTreeEdgesSql);
		while (thisTreeEdgesResultSet.next())
		{
			// from/to ids
			final String thisFromId = thisTreeEdgesResultSet.getString(thisFromName);
			final String thisToId = thisTreeEdgesResultSet.getString(thisToName);
			if ((thisFromId == null || thisFromId.isEmpty()) && (thisToId == null || thisToId.isEmpty()))
			{
				System.err.println("Sql: " + thisTreeEdgesResultSet.getRow() + "th tree edge record with null ids : from-id = <" + thisFromId + "> and to-id=<" + thisToId + ">");
				continue;
			}

			// to node
			final TreeMutableNode thisToNode = this.theIdToNodeMap.get(thisToId);

			// root candidate
			if ((thisFromId == null || thisFromId.isEmpty()) && thisToNode != null)
			{
				theseParentLessNodes.add(thisToNode);
				continue;
			}

			// from node
			final TreeMutableNode thisFromNode = this.theIdToNodeMap.get(thisFromId);

			// ill-formed nodes
			if (thisFromNode == null || thisToNode == null)
			{
				System.err.println("Sql: " + thisTreeEdgesResultSet.getRow() + "th tree edge record with null end nodes : from-id = <" + thisFromId + "> and to-id=<" + thisToId + ">");
				continue;
			}

			// make tree
			List<INode> theseChildren = thisFromNode.getChildren();
			if (theseChildren == null)
			{
				theseChildren = new ArrayList<INode>();
				thisFromNode.setChildren(theseChildren);
			}
			theseChildren.add(thisToNode);
			thisToNode.setParent(thisFromNode);

			// data
			thisToNode.setEdgeLabel(readString(thisTreeEdgesResultSet, thisEdgeLabelName));
			thisToNode.setEdgeImageFile(readString(thisTreeEdgesResultSet, thisEdgeImageName));
			thisToNode.setEdgeColor(readColor(thisTreeEdgesResultSet, thisEdgeColorName));
			final Boolean thisBoolean = readBoolean(thisTreeEdgesResultSet, thisEdgeHiddenName);
			thisToNode.setEdgeStyle(Utils.parseStyle(readString(thisTreeEdgesResultSet, thisEdgeStrokeName), readString(thisTreeEdgesResultSet, thisEdgeFromTerminatorName), readString(thisTreeEdgesResultSet, thisEdgeToTerminatorName),
					thisBoolean == null ? null : thisBoolean.toString()));
		}
		thisTreeEdgesResultSet.close();
		thisTreeEdgesStatement.close();

		if (theseParentLessNodes.size() == 1)
			return theseParentLessNodes.get(0);
		return this.theIdToNodeMap.get(getName("id.root"));
	}

	/**
	 * Query edges
	 * 
	 * @param thisConnection
	 *        connection
	 * @return edge list
	 * @throws SQLException
	 */
	private List<IEdge> queryEdges(final Connection thisConnection) throws SQLException
	{
		// sql
		String thisEdgesSql = this.theProperties.getProperty("edgesSql", Provider.theDefaultEdgesSql);
		thisEdgesSql = macroExpand(thisEdgesSql);

		// field names
		final String thisFromIdName = getName("edges.from");
		final String thisToIdName = getName("edges.to");
		final String thisEdgeLabelName = getName("edges.label");
		final String thisEdgeImageName = getName("edges.image");
		final String thisEdgeColorName = getName("edges.color");
		final String thisEdgeHiddenName = getName("edges.hidden");
		final String thisEdgeStrokeName = getName("edges.stroke");
		final String thisEdgeFromTerminatorName = getName("edges.fromterminator");
		final String thisEdgeToTerminatorName = getName("edges.toterminator");

		List<IEdge> thisEdgeList = null;

		// EDGES
		final Statement thisEdgesStatement = thisConnection.createStatement();
		final ResultSet thisEdgesRecordSet = thisEdgesStatement.executeQuery(thisEdgesSql);
		while (thisEdgesRecordSet.next())
		{
			// read from/to ids
			final String thisFromId = thisEdgesRecordSet.getString(thisFromIdName);
			final String thisToId = thisEdgesRecordSet.getString(thisToIdName);
			if (thisFromId == null || thisToId == null)
			{
				System.err.println("Sql: " + thisEdgesRecordSet.getRow() + "th edge record with null from/to id");
				continue;
			}

			// check from/to
			final MutableNode thisFromNode = this.theIdToNodeMap.get(thisFromId);
			final MutableNode thisToNode = this.theIdToNodeMap.get(thisToId);
			if (thisFromNode == null || thisToNode == null)
			{
				System.err.println("Sql: " + thisEdgesRecordSet.getRow() + "th edge record with null from/to node");
				continue;
			}

			// make edge
			final MutableEdge thisEdge = new MutableEdge(thisFromNode, thisToNode);
			thisEdge.setColor(readColor(thisEdgesRecordSet, "color"));
			if (thisEdgeList == null)
			{
				thisEdgeList = new ArrayList<IEdge>();
			}
			thisEdgeList.add(thisEdge);

			// attributes
			thisEdge.setLabel(readString(thisEdgesRecordSet, thisEdgeLabelName));
			thisEdge.setImageFile(readString(thisEdgesRecordSet, thisEdgeImageName));
			thisEdge.setColor(readColor(thisEdgesRecordSet, thisEdgeColorName));
			final Boolean thisBoolean = readBoolean(thisEdgesRecordSet, thisEdgeHiddenName);
			thisEdge.setStyle(Utils.parseStyle(thisEdgesRecordSet.getString(thisEdgeStrokeName), thisEdgesRecordSet.getString(thisEdgeFromTerminatorName), thisEdgesRecordSet.getString(thisEdgeToTerminatorName),
					thisBoolean == null ? null : thisBoolean.toString()));
		}
		thisEdgesRecordSet.close();
		thisEdgesStatement.close();
		return thisEdgeList;
	}

	/**
	 * Query settings
	 * 
	 * @param thisConnection
	 *        connection
	 * @return settings
	 */
	private Settings querySettings(final Connection thisConnection)
	{
		// sql
		String thisSettingsSql = this.theProperties.getProperty("settingsSql", Provider.theDefaultSettingsSql);
		thisSettingsSql = macroExpand(thisSettingsSql);
		String thisMenuSql = this.theProperties.getProperty("menuSql", Provider.theDefaultMenuSql);
		thisMenuSql = macroExpand(thisMenuSql);

		// field names
		final String thisBackImageName = getName("settings.backimage");
		final String thisBackcolorName = getName("settings.backcolor");
		final String thisForecolorName = getName("settings.forecolor");
		final String thisFontFaceName = getName("settings.fontface");
		final String thisFontSizeName = getName("settings.fontsize");
		final String thisFontSizeStepName = getName("settings.fontsizestep");
		final String thisOrientationName = getName("settings.orientation");
		final String thisExpansionName = getName("settings.expansion");
		final String thisSweepName = getName("settings.sweep");
		final String thisPreserveOrientationName = getName("settings.preserveorientation");
		final String thisHyperbolicName = getName("settings.hyperbolic");
		final String thisHasToolbarName = getName("settings.hastoolbar");
		final String thisHasStatusbarName = getName("settings.hasstatusbar");
		final String thisHasPopupMenuName = getName("settings.haspopupmenu");
		final String thisHasTooltipName = getName("settings.hastooltip");
		final String thisTooltipDisplaysContentName = getName("settings.tooltipdisplayscontent");
		final String thisFocusOnHoverName = getName("settings.focusonhover");
		final String thisFocusName = getName("settings.focus");
		final String thisXMoveToName = getName("settings.xmoveto");
		final String thisYMoveToName = getName("settings.ymoveto");
		final String thisXShiftName = getName("settings.xshift");
		final String thisYShiftName = getName("settings.yshift");
		final String thisNodeBackcolorName = getName("settings.nodebackcolor");
		final String thisNodeForecolorName = getName("settings.nodeforecolor");
		final String thisNodeImageName = getName("settings.nodeimage");
		final String thisTreeEdgeColorName = getName("settings.treeedgecolor");
		final String thisTreeEdgeHiddenName = getName("settings.treeedgehidden");
		final String thisTreeEdgeStrokeName = getName("settings.treeedgestroke");
		final String thisTreeEdgeFromTerminatorName = getName("settings.treeedgefromterminator");
		final String thisTreeEdgeToTerminatorName = getName("settings.treeedgetoterminator");
		final String thisTreeEdgeImageName = getName("settings.treeedgeimage");
		final String thisEdgeColorName = getName("settings.edgecolor");
		final String thisEdgeHiddenName = getName("settings.edgehidden");
		final String thisEdgeStrokeName = getName("settings.edgestroke");
		final String thisEdgeFromTerminatorName = getName("settings.edgefromterminator");
		final String thisEdgeToTerminatorName = getName("settings.edgetoterminator");
		final String thisEdgeImageName = getName("settings.edgeimage");

		final String thisMenuActionName = getName("menu.action");
		final String thisMenuLabelName = getName("menu.label");
		final String thisMenuTargetName = getName("menu.target");
		final String thisMenuScopeName = getName("menu.scope");
		final String thisMenuModeName = getName("menu.mode");
		final String thisMenuLinkName = getName("menu.link");

		try
		{
			final Settings theseSettings = new Settings();

			// first record
			final Statement thisEdgesStatement = thisConnection.createStatement();
			final ResultSet thisSettingsResultSetSet = thisEdgesStatement.executeQuery(thisSettingsSql);
			while (thisSettingsResultSetSet.next())
			{
				// boolean
				theseSettings.theHasToolbarFlag = readBoolean(thisSettingsResultSetSet, thisHasToolbarName);
				theseSettings.theHasStatusbarFlag = readBoolean(thisSettingsResultSetSet, thisHasStatusbarName);
				theseSettings.theHasPopUpMenuFlag = readBoolean(thisSettingsResultSetSet, thisHasPopupMenuName);
				theseSettings.theHasToolTipFlag = readBoolean(thisSettingsResultSetSet, thisHasTooltipName);
				theseSettings.theToolTipDisplaysContentFlag = readBoolean(thisSettingsResultSetSet, thisTooltipDisplaysContentName);
				theseSettings.theFocusOnHoverFlag = readBoolean(thisSettingsResultSetSet, thisFocusOnHoverName);
				theseSettings.thePreserveOrientationFlag = readBoolean(thisSettingsResultSetSet, thisPreserveOrientationName);
				theseSettings.theHyperbolicEdgesFlag = readBoolean(thisSettingsResultSetSet, thisHyperbolicName);

				// integer
				theseSettings.theFontSize = readInteger(thisSettingsResultSetSet, thisFontSizeName);
				theseSettings.theFontSizeStep = readInteger(thisSettingsResultSetSet, thisFontSizeStepName);

				// double
				theseSettings.theXMoveTo = readDouble(thisSettingsResultSetSet, thisXMoveToName);
				theseSettings.theYMoveTo = readDouble(thisSettingsResultSetSet, thisYMoveToName);
				theseSettings.theXShift = readDouble(thisSettingsResultSetSet, thisXShiftName);
				theseSettings.theYShift = readDouble(thisSettingsResultSetSet, thisYShiftName);
				theseSettings.theExpansion = readDouble(thisSettingsResultSetSet, thisExpansionName);
				theseSettings.theSweep = readDouble(thisSettingsResultSetSet, thisSweepName);

				// colors
				theseSettings.theBackColor = readColor(thisSettingsResultSetSet, thisBackcolorName);
				theseSettings.theForeColor = readColor(thisSettingsResultSetSet, thisForecolorName);
				theseSettings.theNodeBackColor = readColor(thisSettingsResultSetSet, thisNodeBackcolorName);
				theseSettings.theNodeForeColor = readColor(thisSettingsResultSetSet, thisNodeForecolorName);
				theseSettings.theTreeEdgeColor = readColor(thisSettingsResultSetSet, thisTreeEdgeColorName);
				theseSettings.theEdgeColor = readColor(thisSettingsResultSetSet, thisEdgeColorName);

				// styles
				Boolean thisBoolean = readBoolean(thisSettingsResultSetSet, thisTreeEdgeHiddenName);
				theseSettings.theTreeEdgeStyle = Utils.parseStyle(readString(thisSettingsResultSetSet, thisTreeEdgeStrokeName), readString(thisSettingsResultSetSet, thisTreeEdgeFromTerminatorName),
						readString(thisSettingsResultSetSet, thisTreeEdgeToTerminatorName), thisBoolean == null ? null : thisBoolean.toString());
				thisBoolean = readBoolean(thisSettingsResultSetSet, thisEdgeHiddenName);
				theseSettings.theEdgeStyle = Utils.parseStyle(readString(thisSettingsResultSetSet, thisEdgeStrokeName), readString(thisSettingsResultSetSet, thisEdgeFromTerminatorName), readString(thisSettingsResultSetSet, thisEdgeToTerminatorName),
						thisBoolean == null ? null : thisBoolean.toString());

				// string
				theseSettings.theOrientation = readString(thisSettingsResultSetSet, thisOrientationName);
				theseSettings.theFocus = readString(thisSettingsResultSetSet, thisFocusName);
				theseSettings.theFontFace = readString(thisSettingsResultSetSet, thisFontFaceName);
				theseSettings.theOrientation = readString(thisSettingsResultSetSet, thisOrientationName);
				theseSettings.theBackgroundImage = readString(thisSettingsResultSetSet, thisBackImageName);
				theseSettings.theDefaultNodeImage = readString(thisSettingsResultSetSet, thisNodeImageName);
				theseSettings.theDefaultTreeEdgeImage = readString(thisSettingsResultSetSet, thisTreeEdgeImageName);
				theseSettings.theDefaultEdgeImage = readString(thisSettingsResultSetSet, thisEdgeImageName);

				// take first if many are returned
				break;
			}
			thisSettingsResultSetSet.close();
			thisEdgesStatement.close();

			// first record
			final Statement thisMenuStatement = thisConnection.createStatement();
			final ResultSet thisMenuResultSet = thisMenuStatement.executeQuery(thisMenuSql);
			while (thisMenuResultSet.next())
			{
				final MenuItem thisMenuItem = new MenuItem();
				thisMenuItem.theLabel = readString(thisMenuResultSet, thisMenuLabelName);
				thisMenuItem.theMatchTarget = readString(thisMenuResultSet, thisMenuTargetName);
				thisMenuItem.theLink = readString(thisMenuResultSet, thisMenuLinkName);
				final String thisActionString = readString(thisMenuResultSet, thisMenuActionName);
				final String thisScopeString = readString(thisMenuResultSet, thisMenuScopeName);
				final String thisModeString = readString(thisMenuResultSet, thisMenuModeName);
				Utils.parseMenuItem(thisMenuItem, thisActionString, thisScopeString, thisModeString);

				// add
				if (theseSettings.theMenu == null)
				{
					theseSettings.theMenu = new ArrayList<MenuItem>();
				}
				theseSettings.theMenu.add(thisMenuItem);
			}
			thisMenuResultSet.close();
			thisMenuStatement.close();

			return theseSettings;
		}
		catch (final SQLException e)
		{
			System.err.println("Sql: Cannot read Settings." + e.getMessage());
		}
		// return default
		return new Settings();
	}

	// R E S U L T S E T . H E L P E R S

	/**
	 * Read string
	 * 
	 * @param thisRecordSet
	 *        record set
	 * @param thisName
	 *        field name
	 * @return String value
	 */
	private String readString(final ResultSet thisRecordSet, final String thisName)
	{
		try
		{
			return thisRecordSet.getString(thisName);
		}
		catch (final SQLException e)
		{
			// do nothing
		}
		return null;
	}

	/**
	 * Read integer
	 * 
	 * @param thisRecordSet
	 *        record set
	 * @param thisName
	 *        field name
	 * @return Integer value
	 */
	private Integer readInteger(final ResultSet thisRecordSet, final String thisName)
	{
		try
		{
			return thisRecordSet.getInt(thisName);
		}
		catch (final SQLException e)
		{
			// do nothing
		}
		return null;
	}

	/**
	 * Read double
	 * 
	 * @param thisRecordSet
	 *        record set
	 * @param thisName
	 *        field name
	 * @return Double value
	 */
	private Double readDouble(final ResultSet thisRecordSet, final String thisName)
	{
		try
		{
			return thisRecordSet.getDouble(thisName);
		}
		catch (final SQLException e)
		{
			// do nothing
		}
		return null;
	}

	/**
	 * Read boolean
	 * 
	 * @param thisRecordSet
	 *        record set
	 * @param thisName
	 *        field name
	 * @return Boolean value
	 */
	private Boolean readBoolean(final ResultSet thisRecordSet, final String thisName)
	{
		try
		{
			return thisRecordSet.getBoolean(thisName);
		}
		catch (final SQLException e)
		{
			// do nothing
		}
		return null;
	}

	/**
	 * Read color
	 * 
	 * @param thisRecordSet
	 *        record set
	 * @param thisName
	 *        field name
	 * @return Color value
	 */
	private Color readColor(final ResultSet thisRecordSet, final String thisName)
	{
		try
		{
			final String thisValue = thisRecordSet.getString(thisName);
			if (thisValue != null)
				return Utils.stringToColor(thisValue);
		}
		catch (final SQLException e)
		{
			// do nothing
		}
		return null;
	}

	// P R O P E R T I E S

	/**
	 * Get name
	 * 
	 * @param thisKey
	 *        key
	 * @return value if not null or key after last '.'
	 */
	private String getName(final String thisKey)
	{
		final String thisValue = (String) this.theProperties.get(thisKey);
		if (thisValue != null)
			return thisValue;

		final String thisStringKey = (String) thisKey;
		final int thisIndex = thisStringKey.lastIndexOf('.');
		return thisIndex == -1 ? thisStringKey : thisStringKey.substring(thisIndex + 1);
	}

	/**
	 * Expand macros
	 * 
	 * @param thisString
	 *        string to expand macro in
	 * @return string with expanded macros
	 */
	private String macroExpand(final String thisString)
	{
		final String thisRegEx = "%[^%]*%";
		final Pattern thePattern = Pattern.compile(thisRegEx);

		// percents
		int thisPercentCount = 0;
		for (int p = 0; (p = thisString.indexOf('"', p + 1)) != -1;)
		{
			thisPercentCount++;
		}
		if (thisPercentCount % 2 != 0)
			return null;

		// macro map
		final Map<String, String> thisMacroMap = new HashMap<String, String>();
		for (final Matcher thisMatcher = thePattern.matcher(thisString); thisMatcher.find();)
		{
			final String thisMatch = thisMatcher.group();
			final String thisKey = thisMatch.substring(1, thisMatch.length() - 1);
			if (!thisMacroMap.containsKey(thisKey))
			{
				final String thisValue = (String) this.theProperties.get(thisKey);
				thisMacroMap.put(thisKey, thisValue);
			}
		}

		// macro substitution
		String thisResult = thisString;
		for (final String thisKey : thisMacroMap.keySet())
		{
			String thisValue = thisMacroMap.get(thisKey);
			if (thisValue == null)
			{
				thisValue = thisKey;
			}
			thisResult = thisResult.replaceAll("%" + thisKey + "%", thisValue);
		}
		return thisResult;
	}
}
