/**
 * Title: Treebolic Text provider
 * Description: Treebolic Text provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.text.indent;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;

import treebolic.model.IEdge;
import treebolic.model.INode;
import treebolic.model.IProvider;
import treebolic.model.IProviderContext;
import treebolic.model.Model;
import treebolic.model.MutableNode;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.Utils;

/**
 * Provider for file system
 * 
 * @author Bernard Bou
 */
public class Provider implements IProvider
{
	static private Color theBackgroundColor = Color.WHITE;

	/**
	 * Stack entry
	 */
	private class StackEntry
	{
		/**
		 * Node
		 */
		public MutableNode theNode;

		/**
		 * Node's level
		 */
		public int theLevel;

		/**
		 * Constructor
		 * 
		 * @param thisNode
		 *        node
		 * @param thisLevel
		 *        itslevel
		 */
		public StackEntry(final MutableNode thisNode, final int thisLevel)
		{
			this.theNode = thisNode;
			this.theLevel = thisLevel;
		}
	}

	/**
	 * Provider context
	 */
	private IProviderContext theContext;

	/**
	 * Constructor
	 */
	public Provider()
	{
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#set(treebolic.model.IProviderContext)
	 */
	@Override
	public void connect(final IProviderContext thisContext)
	{
		this.theContext = thisContext;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#makeModel(java.lang.String)
	 */
	@Override
	public Model makeModel(final String thisSource)
	{
		final Tree thisTree = makeTree(thisSource, false);
		if (thisTree == null)
			return null;

		// settings
		final boolean asTree = thisTree.getRoot().getChildren().size() < 3;
		final Settings theseSettings = new Settings();
		theseSettings.theBackColor = Provider.theBackgroundColor;
		theseSettings.theNodeBackColor = Color.WHITE;
		theseSettings.theNodeForeColor = Color.BLACK;

		theseSettings.theOrientation = asTree ? "south" : "radial";
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theBackColor = new Color(0xf5f5f0);
		theseSettings.theFontFace = "Dialog";
		theseSettings.theFontSize = 20;
		theseSettings.theFontSizeStep = 2;
		theseSettings.theExpansion = .9;
		theseSettings.theSweep = 1.2;
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theHasStatusbarFlag = true;
		theseSettings.theFocusOnHoverFlag = true;
		theseSettings.theTreeEdgeColor = Color.GRAY;
		theseSettings.theTreeEdgeStyle = IEdge.SOLID | IEdge.TOTRIANGLE | IEdge.TOSOLID;
		return new Model(thisTree, theseSettings);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#makeTree(java.lang.String, boolean)
	 */
	@Override
	public Tree makeTree(final String thisSource, final boolean checkRecursion)
	{
		// get text file
		String thisTextFile = thisSource;
		if (thisTextFile == null)
		{
			thisTextFile = this.theContext.getParameter("text");
		}

		if (thisTextFile != null)
		{
			// URL
			final URL thisURL = this.theContext.makeURL(thisTextFile);
			if (thisURL != null)
			{
				// parse
				this.theContext.putProgress("Loading ..." + thisTextFile, false);
				final Tree thisTree = parseTree(thisURL);
				if (thisTree != null)
				{
					this.theContext.putProgress("Loaded " + thisTextFile, false);
					return thisTree;
				}
				else
				{
					this.theContext.putMessage("Cannot load text file <" + thisTextFile + ">");
				}
			}
			else
			{
				this.theContext.putMessage("Bad URL " + thisTextFile);
			}
		}
		return null;
	}

	/**
	 * Parse tree
	 * 
	 * @param thisUrl
	 *        text file URL
	 * @return tree
	 */
	private Tree parseTree(final URL thisUrl)
	{
		// root
		MutableNode thisRootNode = new MutableNode(null, "root");
		thisRootNode.setLabel("root");
		thisRootNode.setBackColor(Color.RED);
		thisRootNode.setForeColor(Color.WHITE);

		final Deque<StackEntry> thisStack = new ArrayDeque<StackEntry>();
		thisStack.push(new StackEntry(thisRootNode, -1));

		try
		{
			// input stream
			final InputStream thisInputStream = thisUrl.openStream();
			final BufferedReader thisReader = new BufferedReader(new InputStreamReader(thisInputStream));

			// parse lines
			String thisLine;
			for (int l = 1; (thisLine = thisReader.readLine()) != null; l++)
			{
				if (thisLine.matches("^\\s*$") || thisLine.startsWith("#"))
				{
					continue;
				}
				processLine(thisLine, l, thisStack);
			}

			// graph
			final List<INode> theseChildren = thisRootNode.getChildren();
			if (theseChildren.size() == 1)
			{
				thisRootNode = (MutableNode) theseChildren.get(0);
				thisRootNode.setParent(null);
			}
			return new Tree(thisRootNode, null);
		}
		catch (final Exception ex)
		{
			System.err.println("Text graph parser: " + ex.toString());
			return null;
		}
	}

	/**
	 * Process line
	 * 
	 * @param thisLine
	 *        line
	 * @param thisLineNumber
	 *        line number
	 * @param thisStack
	 *        node stack
	 */
	private void processLine(final String thisLine, final int thisLineNumber, final Deque<StackEntry> thisStack)
	{
		final int thisLevel = getLevel(thisLine);
		if (thisLevel == -1)
			return;

		// stack
		final StackEntry thisEntry = thisStack.peek();
		int thatLevel = thisEntry.theLevel;

		// parent
		MutableNode thisParent = null;
		if (thisLevel == thatLevel)
		{
			thisStack.pop();
			thisParent = thisStack.peek().theNode;
		}
		else if (thisLevel > thatLevel)
		{
			thisParent = thisStack.peek().theNode;
		}
		else
		{
			while (thisLevel <= thatLevel)
			{
				thisStack.pop();
				thatLevel = thisStack.peek().theLevel;
			}
			thisParent = thisStack.peek().theNode;
		}

		// parse
		// label:id:backcolor:forecolor:img:link:comment
		final String[] theseFields = thisLine.substring(thisLevel).split(":");
		final String thisId = theseFields.length > 1 ? theseFields[1] : "N" + thisLineNumber;

		// new node
		final MutableNode thisNode = new MutableNode(thisParent, thisId);
		thisStack.push(new StackEntry(thisNode, thisLevel));

		// node data
		thisNode.setLabel(theseFields[0]);
		if (theseFields.length > 2)
		{
			thisNode.setBackColor(Utils.stringToColor(theseFields[2]));
		}
		if (theseFields.length > 3)
		{
			thisNode.setForeColor(Utils.stringToColor(theseFields[3]));
		}
		if (theseFields.length > 4)
		{
			thisNode.setImageFile(theseFields[4]);
		}
		if (theseFields.length > 5)
		{
			thisNode.setLink(theseFields[5]);
		}
		if (theseFields.length > 6)
		{
			thisNode.setContent(theseFields[6]);
		}
	}

	/**
	 * Get level (as per indentation)
	 * 
	 * @param thisLine
	 *        line
	 * @return level
	 */
	private int getLevel(final String thisLine)
	{
		for (int i = 0; i < thisLine.length(); i++)
		{
			final char thisChar = thisLine.charAt(i);
			if (thisChar != ' ' && thisChar != '\t')
				return i;
		}
		return -1;
	}
}
