/**
 * Title: Treebolic Text provider
 * Description: Treebolic Text provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.text.pair;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import treebolic.model.IEdge;
import treebolic.model.IProvider;
import treebolic.model.IProviderContext;
import treebolic.model.Model;
import treebolic.model.MutableEdge;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.TreeMutableNode;
import treebolic.model.graph.Converter;
import treebolic.model.graph.GraphEdge;
import treebolic.model.graph.GraphNode;
import treebolic.model.graph.MutableGraph;

/**
 * Provider for file system
 * 
 * @author Bernard Bou
 */
public class Provider implements IProvider
{
	static private Color theBackgroundColor = Color.WHITE;

	private class MutableGraphNode extends TreeMutableNode implements GraphNode
	{
		/**
		 * Constructor
		 * 
		 * @param thisId
		 *        id
		 */
		public MutableGraphNode(final String thisId)
		{
			super(null, thisId);
		}
	}

	/**
	 * Provider context
	 */
	private IProviderContext theContext;

	/**
	 * Constructor
	 */
	public Provider()
	{
		// do nothing
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#set(treebolic.model.IProviderContext)
	 */
	@Override
	public void connect(final IProviderContext thisContext)
	{
		this.theContext = thisContext;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#makeModel(java.lang.String)
	 */
	@Override
	public Model makeModel(final String thisSource)
	{
		final Tree thisTree = makeTree(thisSource, false);
		if (thisTree == null)
			return null;

		// settings
		final boolean asTree = thisTree.getRoot().getChildren().size() < 3;
		final Settings theseSettings = new Settings();
		theseSettings.theBackColor = Provider.theBackgroundColor;
		theseSettings.theNodeBackColor = Color.WHITE;
		theseSettings.theNodeForeColor = Color.BLACK;

		theseSettings.theOrientation = asTree ? "south" : "radial";
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theBackColor = new Color(0xf5f5f0);
		theseSettings.theFontFace = "Dialog";
		theseSettings.theFontSize = 20;
		theseSettings.theFontSizeStep = 2;
		theseSettings.theExpansion = .9;
		theseSettings.theSweep = 1.2;
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theHasStatusbarFlag = true;
		theseSettings.theFocusOnHoverFlag = true;
		theseSettings.theTreeEdgeColor = Color.GRAY;
		theseSettings.theTreeEdgeStyle = IEdge.SOLID | IEdge.TOTRIANGLE | IEdge.TOSOLID;
		return new Model(thisTree, theseSettings);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#makeTree(java.lang.String, boolean)
	 */
	@Override
	public Tree makeTree(final String thisSource, final boolean checkRecursion)
	{
		// get text file
		String thisTextFile = thisSource;
		if (thisTextFile == null)
		{
			thisTextFile = this.theContext.getParameter("text");
		}

		if (thisTextFile != null)
		{
			// URL
			final URL thisURL = this.theContext.makeURL(thisTextFile);
			if (thisURL != null)
			{
				// parse
				this.theContext.putProgress("Loading ..." + thisTextFile, false);
				final Tree thisTree = parseTree(thisURL);
				if (thisTree != null)
				{
					this.theContext.putProgress("Loaded " + thisTextFile, false);
					return thisTree;
				}
				else
				{
					this.theContext.putMessage("Cannot load text file <" + thisTextFile + ">");
				}
			}
			else
			{
				this.theContext.putMessage("Bad URL " + thisTextFile);
			}
		}
		return null;
	}

	/**
	 * Parse tree
	 * 
	 * @param thisUrl
	 *        text file URL
	 * @return tree
	 */
	private Tree parseTree(final URL thisUrl)
	{
		final MutableGraph thisGraph = new MutableGraph();
		final Map<String, MutableGraphNode> thisMap = new HashMap<String, MutableGraphNode>();
		try
		{
			// input stream
			final InputStream thisInputStream = thisUrl.openStream();
			final BufferedReader thisReader = new BufferedReader(new InputStreamReader(thisInputStream));

			// parse lines
			String thisLine;
			for (int l = 1; (thisLine = thisReader.readLine()) != null; l++)
			{
				if (thisLine.matches("^\\s*$") || thisLine.startsWith("#"))
				{
					continue;
				}
				processLine(thisLine, l, thisGraph, thisMap);
			}
			return new Converter<MutableGraphNode>().graphToTree(thisGraph);
		}
		catch (final Exception ex)
		{
			System.err.println("Text graph parser: " + ex.toString());
			return null;
		}
	}

	/**
	 * Process line
	 * 
	 * @param thisLine
	 *        line
	 * @param thisLineNumber
	 *        line number
	 * @param thisMap
	 *        node stack
	 */
	private void processLine(final String thisLine, final int thisLineNumber, final MutableGraph thisGraph, final Map<String, MutableGraphNode> thisMap)
	{
		// parse
		// parent\tchild
		final String[] theseFields = thisLine.split("\t");
		if (theseFields.length < 2)
			return;
		final String thisParentId = theseFields[0].trim();
		if (thisParentId == null || thisParentId.isEmpty())
		{
			System.err.println("ERR " + thisLineNumber);
		}
		final String thisChildId = theseFields[1].trim();
		if (thisChildId == null || thisChildId.isEmpty())
		{
			System.err.println("ERR " + thisLineNumber);
		}

		// nodes
		MutableGraphNode thisParent = thisMap.get(thisParentId);
		if (thisParent == null)
		{
			thisParent = new MutableGraphNode(thisParentId);
			thisParent.setLabel(thisParentId);
			thisMap.put(thisParentId, thisParent);
			thisGraph.add(thisParent);
		}
		MutableGraphNode thisChild = thisMap.get(thisChildId);
		if (thisChild == null)
		{
			thisChild = new MutableGraphNode(thisChildId);
			thisChild.setLabel(thisChildId);
			thisMap.put(thisChildId, thisChild);
			thisGraph.add(thisChild);
		}

		// edge
		final GraphEdge thisGraphEdge = new GraphEdge(thisParent, thisChild, true);
		thisGraphEdge.setUserData(new MutableEdge(thisParent, thisChild));
		thisGraph.add(thisGraphEdge);
	}
}
