/**
 * Title: Treebolic WordNet provider
 * Description: Treebolic WordNet Provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.wn;

import java.awt.Color;
import java.awt.Image;
import java.util.ArrayList;

import javax.swing.ImageIcon;

import treebolic.control.Controller;
import treebolic.model.IEdge;
import treebolic.model.IProvider;
import treebolic.model.IProviderContext;
import treebolic.model.MenuItem;
import treebolic.model.Model;
import treebolic.model.Settings;
import treebolic.model.Tree;
import edu.princeton.wordnet.Index;
import edu.princeton.wordnet.Synset;
import edu.princeton.wordnet.WordNet;

/**
 * Provider for WordNet
 * 
 * @author Bernard Bou
 */
public class Provider implements IProvider
{
	/**
	 * Focus image
	 */
	private static final Image theFocusImage = new ImageIcon(Provider.class.getResource("images/focus.gif")).getImage();

	/**
	 * Provider client
	 */
	private IProviderContext theContext;

	/**
	 * Extended WordNet interface
	 */
	private XWordNet theWn;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 */
	public Provider()
	{
		//System.out.println("java.library.path=" + System.getProperty("java.library.path"));
		//System.out.println("WNHOME=" + System.getenv("WNHOME"));
		theWn = new XWordNet();
	}

	// I N T E R F A C E

	/*
	 * (non-Javadoc)
	 * 
	 * @see treebolic.model.IProvider#set(treebolic.model.IProviderContext)
	 */
	public void connect(IProviderContext thisContext)
	{
		theContext = thisContext;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see treebolic.model.IProvider#makeModel(java.lang.String)
	 */
	public Model makeModel(String thisSource)
	{
		Tree thisTree = makeTree(thisSource, false);
		if (thisTree == null)
			return null;

		// settings
		boolean asTree = thisTree.getRoot().getChildren().size() < 5;
		Settings theseSettings = new Settings();
		theseSettings.theOrientation = asTree ? "south" : "radial";
		theseSettings.theBackColor = new Color(0xf3f1e2);
		theseSettings.theFontFace = "Dialog";
		theseSettings.theFontSize = 20;
		theseSettings.theFontSizeStep = 2;
		theseSettings.theExpansion = .9;
		theseSettings.theSweep = 1.2;
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theHasStatusbarFlag = true;
		theseSettings.theFocusOnHoverFlag = false;
		theseSettings.theTreeEdgeColor = Color.GRAY;
		theseSettings.theTreeEdgeStyle = IEdge.DASH | IEdge.TOTRIANGLE | IEdge.TOSOLID;
		if (asTree)
			theseSettings.theYMoveTo = -0.3;

		// menu
		theseSettings.theMenu = new ArrayList<MenuItem>();

		MenuItem thisInputMenuItem = new MenuItem();
		thisInputMenuItem.theAction = Controller.Action.GOTO;
		thisInputMenuItem.theLabel = "Input (=%e)";
		thisInputMenuItem.theLink = "%e";
		theseSettings.theMenu.add(thisInputMenuItem);

		MenuItem thisContextMenuItem = new MenuItem();
		thisContextMenuItem.theAction = Controller.Action.GOTO;
		thisContextMenuItem.theLabel = "WordNet %u";
		thisContextMenuItem.theLink = "%u";
		theseSettings.theMenu.add(thisContextMenuItem);

		// result
		return new Model(thisTree, theseSettings);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see treebolic.model.IProvider#makeTree(java.lang.String, boolean)
	 */
	public Tree makeTree(String thisSource, boolean checkRecursion)
	{
		String thisWord = thisSource;
		if (thisWord == null)
			thisWord = theContext.getParameter("word");
		if (thisWord == null)
			return null;

		if (thisWord.startsWith("wordnet:"))
			thisWord = thisWord.substring(8);
		thisWord = thisWord.replace(' ', '_');
		//System.out.println("word=" + thisWord);

		// walk
		Node thisRootNode = walkLinks(thisWord, true);

		// result
		return new Tree(thisRootNode, null);
	}

	// H E L P E R S

	/**
	 * Walk links
	 * 
	 * @param thisWord
	 *        target word
	 * @param recurse
	 *        whether to recurse
	 * @return result node
	 */
	public Node walkLinks(String thisWord, boolean recurse)
	{
		Node thisRootNode = new Node(null, thisWord.replace('_', ' '), theFocusImage, Color.BLUE, Color.WHITE);

		// get maps
		int thisPosMask = theWn.getPosMask(thisWord);
		long theseLinks[] = theWn.getPos2LinksMap(thisPosMask, thisWord);

		// iterate on parts of speech
		for (int i = 0; i <= WordNet.LASTPOS - WordNet.FIRSTPOS; i++)
		{
			int thisPos = WordNet.FIRSTPOS + i;

			// valid link types
			int thisPosBit = 1 << (WordNet.FIRSTPOS + i);
			if ((thisPosMask & thisPosBit) != 0)
			{
				// pos node
				Node thisPosNode = new PosNode(thisRootNode, i);
				Index thisIndex = theWn.getindex(thisWord, thisPos);

				if (thisIndex != null)
					for (int j = 0; j < thisIndex.offset.length; j++)
					{
						// read synset
						Synset thisSynset = theWn.read_synset(thisPos, thisIndex.offset[j], thisWord);

						// sense node
						Node thisSenseNode = new SenseNode(thisPosNode, j + 1, thisSynset);

						// synset nodes
						walkSynsets(thisSenseNode, thisSynset, 0);

						// links
						for (int k = 0; k <= WordNet.LASTLINKTYPE - WordNet.FIRSTLINKTYPE; k++)
						{
							int thisSearch = WordNet.FIRSTLINKTYPE + k;

							long thisSearchBit = 1 << thisSearch;
							if ((theseLinks[i] & thisSearchBit & WordNet.thePosLinkMask[i]) != 0)
							{
								// recursion
								if (recurse && XWordNet.canRecurse(thisSearch))
									thisSearch = -thisSearch;

								// domain and domain terms redirected to category
								if (thisSearch == WordNet.CLASSIFICATION)
									thisSearch = WordNet.CLASSIF_CATEGORY;
								else if (thisSearch == WordNet.CLASS)
									thisSearch = WordNet.CLASS_CATEGORY;

								// search
								Synset thisLinkSynset = theWn.findtheinfo_ds(thisWord, thisPos, thisSearch, j + 1);
								if (thisLinkSynset.ptrlist != null)
								{
									// link
									walkLinks(thisSenseNode, thisLinkSynset, k);
								}
							}
						}
					}
			}
		}
		return thisRootNode;
	}

	/**
	 * Walk synsets
	 * 
	 * @param thisParentNode
	 *        parent node
	 * @param thisSynset
	 *        synset
	 * @param thisLevel
	 *        recursion level
	 * @return parent node
	 */
	public Node walkSynsets(Node thisParentNode, Synset thisSynset, int thisLevel)
	{
		for (int i = 0; i < thisSynset.words.length; i++)
		{
			String thisString = thisSynset.words[i];
			thisString = thisString.replace('_', ' ');
			new WordNode(thisParentNode, thisString, thisSynset, thisLevel);
		}
		return thisParentNode;
	}

	/**
	 * Walk links
	 * 
	 * @param thisParentNode
	 *        parent node
	 * @param thisSynset
	 *        synset
	 * @param thisIdx
	 *        index
	 * @return parent node
	 */
	public Node walkLinks(Node thisParentNode, Synset thisSynset, int thisIdx)
	{
		Node thisNode = thisParentNode;
		int i = 1;
		for (Synset thisLink = thisSynset.ptrlist; thisLink != null; thisLink = thisLink.ptrlist, i++)
		{
			thisNode = new LinkNode(thisNode, thisIdx);
			walkSynsets(thisNode, thisLink, i);
			walkHorizontalLinks(thisNode, thisLink, i);
		}
		return thisParentNode;
	}

	/**
	 * Walk horizontal links
	 * 
	 * @param thisParentNode
	 *        parent node
	 * @param thisSynset
	 *        synset
	 * @param thisLevel
	 *        recursion level
	 * @return parent node
	 */
	public Node walkHorizontalLinks(Node thisParentNode, Synset thisSynset, int thisLevel)
	{
		for (Synset thisLink = thisSynset.nextss; thisLink != null; thisLink = thisLink.nextss)
			walkSynsets(thisParentNode, thisLink, thisLevel);
		return thisParentNode;
	}
}
