/**
 * Title: Treebolic WordNet provider
 * Description: Treebolic WordNet Provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.wn;

import edu.princeton.wordnet.WordNet;

/**
 * Extended WordNet querier
 * 
 * @author Bernard Bou
 */
class XWordNet extends WordNet
{
	/**
	 * Link type strings
	 */
	public static final String theSearchTypeXString[] = { //
	"%s is opposed to ...(antonyms)", //
			"%s is a kind of ... (hypernyms)", //
			"... is a kind of %s (hyponyms)", //
			"%s entails ...", "%s is similar to ...", //

			"%s is a member of ... (member holonyms)", //
			"%s is a substance of ... (substance holonyms)", //
			"%s is a part of ... (part holonyms)", //

			"... is a member of %s (member meronyms)", //
			"... is a substance of %s (substance meronyms)", //
			"... is a part of %s (part meronyms)", //

			"... is a part of %s (all meronyms)", //
			"%s is a part of ... (all holonyms)", "%s causes ...", //

			"participle", "see also", //
			"pertains to noun or derived from adjective", //
			"attribute", //
			"verb group", //

			"is derivationally related to ...", //
			"domain", "domain term", //

			"find synonyms", //
			"polysemy", //
			"verb example sentences and generic frames", //
			"noun coordinates", //
			"group related senses", //
			"hierarchical meronym search", //
			"hierarchical holonym search", //
			"find keywords by substring", "show all synsets for word" //
	};

	/**
	 * Link type string indexed on pos-1
	 */
	public static final String thePosLinkTypeString[][] = {
	// NOUN
			{ null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null },

			// VERB
			{ "to %s is opposed to ...(antonyms)", "to %s is one way to ... (hypernyms)", "... is is a particular way to %s (troponyms)", "to %s entails ...", null, null, null, null, null, null, null, null, null, "to %s causes ...", null, "see also",
					null, null, "to %s 's verb group", null, null, null },

			// ADJ
			{ null, null, null, null, null, null, null, null, null, null, null, null, null, null, "participle of verb ...", null, "pertains to noun ...", null, null, null, null, null },

			// ADV
			{ null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, "derived from adjective ...", null, null, null, null, null } };

	/**
	 * Constructor
	 */
	public XWordNet()
	{
		// initialize
		int thisInitResult = this.wninit();
		if (thisInitResult != 0)
			System.err.println("Wn: wninit:" + thisInitResult);

		// initialize morphology functions
		int thisMorphInitResult = this.morphinit();
		if (thisMorphInitResult != 0)
			System.err.println("Wn: morphinit:" + thisMorphInitResult);

		if (thisInitResult != 0 || thisMorphInitResult != 0)
		{
			System.err.println("Wn: java.library.path=" + System.getProperty("java.library.path"));
			System.err.println("Wn: WNHOME=" + System.getenv("WNHOME"));
		}
	}

	/**
	 * Get whether search type can recurse
	 * 
	 * @param thisSearch
	 *        search type
	 * @return true if search type can recurse
	 */
	public static boolean canRecurse(int thisSearch)
	{
		switch (thisSearch)
		{
		case WordNet.ANTPTR:
		case WordNet.HYPERPTR:
		case WordNet.HYPOPTR:
		case WordNet.ENTAILPTR:
		case WordNet.ISMEMBERPTR:
		case WordNet.ISSTUFFPTR:
		case WordNet.ISPARTPTR:
		case WordNet.HASMEMBERPTR:
		case WordNet.HASSTUFFPTR:
		case WordNet.HASPARTPTR:
		case WordNet.MERONYM:
		case WordNet.HOLONYM:
		case WordNet.CAUSETO:
			return true;

		case WordNet.SIMPTR:
		case WordNet.PPLPTR:
		case WordNet.SEEALSOPTR:
		case WordNet.PERTPTR:
		case WordNet.ATTRIBUTE:
		case WordNet.VERBGROUP:
		default:
			return false;
		}
	}

	/**
	 * Get search type string
	 * 
	 * @param thisPos
	 *        part of speech
	 * @param thisIndex
	 *        index
	 * @param thisWord
	 *        target word
	 * @return search type string
	 */
	public static String getSearchType(int thisPos, int thisIndex, String thisWord)
	{
		String thisFormat = thePosLinkTypeString[thisPos - WordNet.FIRSTPOS][thisIndex];
		if (thisFormat == null)
			thisFormat = theSearchTypeXString[thisIndex];

		int thisInsertPos = thisFormat.indexOf("%s");
		if (thisInsertPos == -1)
			return thisFormat;
		else
			return thisFormat.substring(0, thisInsertPos) + thisWord + thisFormat.substring(thisInsertPos + 2);
	}

	/**
	 * Get part of speech mask
	 * 
	 * @param thisWord
	 *        target word
	 * @return part of speech mask for target word
	 */
	public int getPosMask(String thisWord)
	{
		return this.in_wn(thisWord, WordNet.ALLPOS);
	}

	/**
	 * Get links for part of speech and target word
	 * 
	 * @param thisPosMask
	 *        part of speech mask
	 * @param thisWord
	 *        target word
	 * @return links
	 */
	public long[] getPos2LinksMap(int thisPosMask, String thisWord)
	{
		// create array
		int thisNPos = WordNet.LASTPOS - WordNet.FIRSTPOS + 1;
		long theLinks[] = new long[thisNPos];

		// iterate on parts of speech
		int thisBit = 1;
		for (int i = WordNet.FIRSTPOS; i <= WordNet.LASTPOS; i++)
		{
			// available links
			thisBit <<= 1;
			theLinks[i - WordNet.FIRSTPOS] = (thisPosMask & thisBit) == 0 ? 0 : this.is_defined(thisWord, i);
		}
		return theLinks;
	}

	/**
	 * Get lemmas for part of speech and target word
	 * 
	 * @param thisPosMask
	 *        part of speech mask
	 * @param thisWord
	 *        target word
	 * @return lemmas
	 */
	public String[] getPos2LemmaMap(int thisPosMask, String thisWord)
	{
		// create array
		int thisNPos = WordNet.LASTPOS - WordNet.FIRSTPOS + 1;
		String theseLemmas[] = new String[thisNPos];

		// iterate on parts of speech
		for (int i = WordNet.FIRSTPOS; i <= WordNet.LASTPOS; i++)
		{
			// lemma
			String thisLemma = this.morphword(thisWord, i);
			if (thisLemma != null && !thisLemma.equalsIgnoreCase(thisWord))
				theseLemmas[i - WordNet.FIRSTPOS] = thisLemma.replace('_', ' ');
			else
				theseLemmas[i - WordNet.FIRSTPOS] = null;
		}
		return theseLemmas;
	}
}
