/**
 * Title :		Treebolic WordNet provider
 * Description: Treebolic WordNet Provider
 * Version:		1.10.x
 * Copyright:	(c) 2001-2005
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author:		Bernard Bou
 * Company:		bsys
 * Update:      Sat May 6 07:22:33 CEST 2006
 */

package treebolic.provider.wn.ejb;

import java.io.IOException;
import java.net.URL;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.rmi.PortableRemoteObject;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;

import org.w3c.dom.Document;

import treebolic.model.IProvider;
import treebolic.model.IProviderContext;
import treebolic.model.Model;
import treebolic.model.Tree;
import treebolic.provider.xml.dom.DocumentAdapter;
import edu.princeton.wordnet.interfaces.WordNet;
import edu.princeton.wordnet.interfaces.WordNetHome;

/**
 * Provider factory implementation for WordNet
 * 
 * @author Bernard Bou
 */
public class Provider implements IProvider
{
	// D A T A

	/**
	 * Context used to query for Url
	 */
	protected IProviderContext theContext;

	/**
	 * Recursive queries
	 */
	static private final boolean recurse = false;

	/**
	 * Recursive queries
	 */
	static private boolean hardContext = false;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 */
	public Provider()
	{
		// do nothing
	}

	// M A K E

	/*
	 * (non-Javadoc)
	 * @see treebolic.model.IProvider#set(treebolic.model.IProviderContext)
	 */
	@Override
	public void connect(final IProviderContext thisContext)
	{
		this.theContext = thisContext;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.provider.xml.dom.Provider#makeModel(java.lang.String)
	 */
	@Override
	public Model makeModel(final String thisSource)
	{
		String thisWord = thisSource;
		if (thisWord == null || thisWord.isEmpty())
		{
			thisWord = this.theContext.getParameter("word");
		}
		if (thisWord == null || thisWord.isEmpty())
		{
			this.theContext.putProgress("Null query <" + thisWord + ">", true);
			return null;
		}

		this.theContext.putProgress("Retrieving ..." + thisWord, false);
		final Document thisDocument = makeDocument(thisWord);
		if (thisDocument == null)
		{
			this.theContext.putProgress("Failed to make document for <" + thisWord + ">", true);
			return null;
		}
		this.theContext.putProgress("Retrieved ..." + thisWord, false);

		this.theContext.putProgress("Adapting ..." + thisWord, false);
		final Model thisModel = new DocumentAdapter(this).makeModel(thisDocument);
		this.theContext.putProgress("Adapted ..." + thisWord, false);
		return thisModel;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.provider.xml.dom.Provider#makeTree(java.lang.String, boolean)
	 */
	@Override
	public Tree makeTree(final String thisSource, final boolean checkRecursion)
	{
		final String thisWord = thisSource;
		if (thisWord == null || thisWord.isEmpty())
		{
			this.theContext.getParameter("word");
		}
		if (thisWord == null || thisWord.isEmpty())
			return null;

		this.theContext.putProgress("Retrieving ..." + thisWord, false);
		final Document thisDocument = makeDocument(thisWord);
		this.theContext.putProgress("Retrieved ..." + thisWord, false);

		this.theContext.putProgress("Adapting ..." + thisWord, false);
		final Tree thisTree = new DocumentAdapter(this).makeTree(thisDocument);
		this.theContext.putProgress("Adapted ..." + thisWord, false);
		return thisTree;
	}

	// D O C U M E N T . F A C T O R Y ( T R A N S F O R M )

	/**
	 * Make treebolic document. Get WordNet docuemnt and transform it.
	 * 
	 * @param thisWord
	 *        queried word
	 * @return DOM document in treebolic format
	 */
	protected Document makeDocument(final String thisWord)
	{
		final Document thisDocument = makeWordNetDocument(thisWord);
		if (thisDocument != null)
		{
			try
			{
				final URL thisXslt = Provider.class.getResource("wordnet2treebolic.xsl");
				final StreamSource thisStyleSource = new StreamSource(thisXslt.openStream());

				// source
				final DOMSource thisSource = new DOMSource(thisDocument);

				// result
				final DOMResult thisResult = new DOMResult();

				// transform
				final TransformerFactory thisTransformerFactory = TransformerFactory.newInstance();
				final Transformer thisTransformer = thisTransformerFactory.newTransformer(thisStyleSource);
				thisTransformer.setOutputProperty(javax.xml.transform.OutputKeys.DOCTYPE_SYSTEM, "WordNet.dtd");
				thisTransformer.transform(thisSource, thisResult);
				return (Document) thisResult.getNode();
			}
			catch (final TransformerException thisException)
			{
				System.err.println("WnEjb: " + thisException.toString());
			}
			catch (final IOException thisException)
			{
				System.err.println("WnEjb: " + thisException.toString());
			}
		}
		return null;
	}

	// W O R D N E T . D O C U M E N T

	/**
	 * Get the initial naming context
	 */
	protected Context getInitialContext() throws Exception
	{
		final Hashtable<String, String> theseProperties = new Hashtable<String, String>();
		theseProperties.put(Context.INITIAL_CONTEXT_FACTORY, "org.jnp.interfaces.NamingContextFactory");
		theseProperties.put(Context.URL_PKG_PREFIXES, "org.jboss.naming:org.jnp.interfaces");
		theseProperties.put(Context.PROVIDER_URL, "jnp://localhost:1099");
		return new InitialContext(theseProperties);
	}

	/**
	 * Get WordNet document
	 * 
	 * @param thisWord
	 *        queried word
	 * @return DOM document in WordNet format
	 */
	protected Document makeWordNetDocument(final String thisWord)
	{
		try
		{
			// context
			final Context thisContext = Provider.hardContext ? getInitialContext() : new InitialContext();

			// lookup
			// Object thisRef = thisContext.lookup("java:/comp/env/ejb/WordNet");
			final Object thisRef = thisContext.lookup("ejb/WordNet");

			// ejb home
			final WordNetHome thisHome = (WordNetHome) PortableRemoteObject.narrow(thisRef, WordNetHome.class);

			// create bean instance
			final WordNet thisBean = thisHome.create();

			// invoke bean method
			final Document thisDocument = thisBean.queryDoc(thisWord, true, Provider.recurse);

			// remove bean
			thisBean.remove();

			return thisDocument;
		}
		catch (final Exception thisException)
		{
			System.err.println(thisException.toString());
		}
		finally
		{
			// do nothing
		}
		return null;
	}
}
