/**
 * Title: Treebolic provider
 * Description: Treebolic Provider
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.provider.xml.dom;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import treebolic.control.Mounter;
import treebolic.model.IEdge;
import treebolic.model.IProvider;
import treebolic.model.MenuItem;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.MutableEdge;
import treebolic.model.MutableNode;
import treebolic.model.Settings;
import treebolic.model.Tree;
import treebolic.model.Utils;

/**
 * Document adapter to model/graph
 * 
 * @author Bernard Bou
 */
public class DocumentAdapter
{
	// D A T A

	/**
	 * Mount task (protracted)
	 * 
	 * @author Bernard Bou
	 */
	class MountTask
	{
		/**
		 * Mount point
		 */
		public MountPoint.Mounting theMountPoint;

		/**
		 * Node to mount graph at
		 */
		public MutableNode theMountingNode;

		/**
		 * Constructor
		 * 
		 * @param thisMountPoint
		 *        mount point
		 * @param thisMountingNode
		 *        Node to mount graph at
		 */
		public MountTask(final MountPoint.Mounting thisMountPoint, final MutableNode thisMountingNode)
		{
			this.theMountPoint = thisMountPoint;
			this.theMountingNode = thisMountingNode;
		}

		/**
		 * Perform task
		 * 
		 * @param theseEdges
		 *        edges in grafting tree
		 */
		@SuppressWarnings("synthetic-access")
		public void run(final List<IEdge> theseEdges)
		{
			if (DocumentAdapter.this.theProvider == null)
			{
				System.out.println("Mount not performed: " + this.theMountPoint + " @ " + this.theMountingNode);
				return;
			}
			final Tree thisTree = DocumentAdapter.this.theProvider.makeTree(this.theMountPoint.theURL, true);
			if (thisTree != null)
			{
				Mounter.graft(this.theMountingNode, thisTree.getRoot(), theseEdges, thisTree.getEdges());
			}
		}
	}

	/**
	 * Protracted mount tasks
	 */
	private List<MountTask> theMountTasks = null;

	/**
	 * Provider (used to generate mounted trees)
	 */
	private final IProvider theProvider;

	/**
	 * Id to node map
	 */
	private Map<String, MutableNode> theIdToNodeMap;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 * 
	 * @param thisProvider
	 *        provider (used in recursion)
	 */
	public DocumentAdapter(final IProvider thisProvider)
	{
		this.theProvider = thisProvider;
	}

	/**
	 * Constructor
	 */
	public DocumentAdapter()
	{
		this(null);
	}

	// A C C E S S

	/**
	 * Get id to node map
	 * 
	 * @return id to node map
	 */
	public Map<String, MutableNode> getIdToNodeMap()
	{
		return this.theIdToNodeMap;
	}

	// M A K E

	/**
	 * Make model from document
	 * 
	 * @param thisDocument
	 *        document
	 * @return model
	 */
	public Model makeModel(final Document thisDocument)
	{
		this.theIdToNodeMap = new TreeMap<String, MutableNode>();
		final Tree thisTree = toTree(thisDocument);
		if (thisTree == null)
			return null;
		final Settings theseSettings = DocumentAdapter.toSettings(thisDocument);
		return new Model(thisTree, theseSettings);
	}

	/**
	 * Make tree from document
	 * 
	 * @param thisDocument
	 *        document
	 * @return tree
	 */
	public Tree makeTree(final Document thisDocument)
	{
		this.theIdToNodeMap = new Hashtable<String, MutableNode>();
		final Tree thisTree = toTree(thisDocument);
		return thisTree;
	}

	// P A R S E

	/**
	 * Make model node
	 * 
	 * @param thisParent
	 *        model parent
	 * @param thisId
	 *        id
	 * @return model node
	 */
	protected MutableNode makeNode(final MutableNode thisParent, final String thisId)
	{
		return new MutableNode(thisParent, thisId);
	}

	/**
	 * Make model edge
	 * 
	 * @param thisFromNode
	 *        model from-node end
	 * @param thisToNode
	 *        model to-node end
	 * @return model edge
	 */
	protected MutableEdge makeEdge(final MutableNode thisFromNode, final MutableNode thisToNode)
	{
		return new MutableEdge(thisFromNode, thisToNode);
	}

	/**
	 * Make graph
	 * 
	 * @param thisDocument
	 *        document
	 * @return graph
	 */
	private Tree toTree(final Document thisDocument)
	{
		// nodes
		final Element thisRootElement = DocumentAdapter.getFirstElementByTagName(thisDocument.getDocumentElement(), "node");
		if (thisRootElement == null)
			return null;
		final MutableNode thisRoot = toNode(thisRootElement, null);

		// edges
		final List<IEdge> theseEdges = toEdges(thisDocument);

		// run protracted mount tasks (had to be protracted until edges become available)
		if (this.theMountTasks != null)
		{
			for (final MountTask thisTask : this.theMountTasks)
			{
				thisTask.run(theseEdges);
			}
			this.theMountTasks.clear();
			this.theMountTasks = null;
		}

		return new Tree(thisRoot, theseEdges);
	}

	/**
	 * Make model node
	 * 
	 * @param thisNodeElement
	 *        starting DOM element
	 * @param thisParent
	 *        model parent node
	 * @return model node
	 */
	private MutableNode toNode(final Element thisNodeElement, final MutableNode thisParent)
	{
		// id
		final String thisId = thisNodeElement.getAttribute("id");

		// make
		final MutableNode thisNode = makeNode(thisParent, thisId);
		this.theIdToNodeMap.put(thisId, thisNode);

		// colors
		final Color thisBackColor = Utils.stringToColor(thisNodeElement.getAttribute("backcolor"));
		thisNode.setBackColor(thisBackColor);
		final Color thisForeColor = Utils.stringToColor(thisNodeElement.getAttribute("forecolor"));
		thisNode.setForeColor(thisForeColor);

		// weight
		final String thisWeight = thisNodeElement.getAttribute("weight");
		if (thisWeight != null && !thisWeight.isEmpty())
		{
			thisNode.setWeight(-Double.parseDouble(thisWeight));
		}

		// label
		Element thisElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisNodeElement, "label");
		if (thisElement != null)
		{
			final String thisLabel = thisElement.getTextContent();
			if (thisLabel != null && !thisLabel.isEmpty())
			{
				thisNode.setLabel(thisLabel);
			}
		}

		// content
		thisElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisNodeElement, "content");
		if (thisElement != null)
		{
			final String thisContent = thisElement.getTextContent();
			if (thisContent != null && !thisContent.isEmpty())
			{
				thisNode.setContent(thisContent);
			}
		}

		// image
		thisElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisNodeElement, "img");
		if (thisElement != null)
		{
			final String thisImageSrc = thisElement.getAttribute("src");
			if (thisImageSrc != null && !thisImageSrc.isEmpty())
			{
				thisNode.setImageFile(thisImageSrc);
			}
		}

		// tree.edge
		thisElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisNodeElement, "treeedge");
		if (thisElement != null)
		{
			// label
			final Element thisLabelElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisElement, "label");
			if (thisLabelElement != null)
			{
				final String thisLabel = thisLabelElement.getTextContent();
				if (thisLabel != null && !thisLabel.isEmpty())
				{
					thisNode.setEdgeLabel(thisLabel);
				}
			}

			// image
			final Element thisImageElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisElement, "img");
			if (thisImageElement != null)
			{
				final String thisImageSrc = thisImageElement.getAttribute("src");
				if (thisImageSrc != null && !thisImageSrc.isEmpty())
				{
					thisNode.setEdgeImageFile(thisImageSrc);
				}
			}

			// color
			final Color thisColor = Utils.stringToColor(thisElement.getAttribute("color"));
			if (thisColor != null)
			{
				thisNode.setEdgeColor(thisColor);
			}

			// style
			final Integer thisStyle = Utils.parseStyle(thisElement.getAttribute("stroke"), thisElement.getAttribute("fromterminator"), thisElement.getAttribute("toterminator"), thisElement.getAttribute("hidden"));
			if (thisStyle != null)
			{
				thisNode.setEdgeStyle(thisStyle);
			}
		}

		// link
		thisElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisNodeElement, "a");
		if (thisElement != null)
		{
			final String thisHref = thisElement.getAttribute("href");
			if (thisHref != null && !thisHref.isEmpty())
			{
				thisNode.setLink(thisHref);
			}
		}

		// mount
		thisElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisNodeElement, "mountpoint");
		if (thisElement != null)
		{
			final Element thisAElement = DocumentAdapter.getFirstElementByTagName(thisElement, "a");
			if (thisAElement != null)
			{
				final String thisHref = thisAElement.getAttribute("href");
				if (thisHref != null && !thisHref.isEmpty())
				{
					final MountPoint.Mounting thisMountPoint = new MountPoint.Mounting();
					thisMountPoint.theURL = thisHref;

					// mount now ?
					final String thisValue = thisElement.getAttribute("now");
					if (thisValue != null && !thisValue.isEmpty() && Boolean.valueOf(thisValue))
					{
						thisMountPoint.now = true;

						final MountTask thisTask = new MountTask(thisMountPoint, thisNode);
						if (this.theMountTasks == null)
						{
							this.theMountTasks = new ArrayList<MountTask>();
						}
						this.theMountTasks.add(thisTask);
					}
					thisNode.setMountPoint(thisMountPoint);
				}
			}
		}

		// recurse to children
		final List<Element> theseChildElements = DocumentAdapter.getLevel1ChildElementsByTagName(thisNodeElement, "node");
		for (final Element thisChildElement : theseChildElements)
		{
			toNode(thisChildElement, thisNode);
		}

		return thisNode;
	}

	/**
	 * Make model edge
	 * 
	 * @param thisEdgeElement
	 *        edge DOM element
	 * @return edge
	 */
	private MutableEdge toEdge(final Element thisEdgeElement)
	{
		final String thisFromId = thisEdgeElement.getAttribute("from");
		final String thisToId = thisEdgeElement.getAttribute("to");
		final MutableNode thisFromNode = this.theIdToNodeMap.get(thisFromId);
		final MutableNode thisToNode = this.theIdToNodeMap.get(thisToId);
		final MutableEdge thisEdge = makeEdge(thisFromNode, thisToNode);

		// label
		final Element thisLabelElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisEdgeElement, "label");
		if (thisLabelElement != null)
		{
			final String thisLabel = thisLabelElement.getTextContent();
			if (thisLabel != null && !thisLabel.isEmpty())
			{
				thisEdge.setLabel(thisLabel);
			}
		}

		// image
		final Element thisImageElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisEdgeElement, "img");
		if (thisImageElement != null)
		{
			final String thisImageSrc = thisImageElement.getAttribute("src");
			if (thisImageSrc != null && !thisImageSrc.isEmpty())
			{
				thisEdge.setImageFile(thisImageSrc);
			}
		}

		// color
		final Color thisColor = Utils.stringToColor(thisEdgeElement.getAttribute("color"));
		if (thisColor != null)
		{
			thisEdge.setColor(thisColor);
		}

		// style
		final Integer thisStyle = Utils.parseStyle(thisEdgeElement.getAttribute("stroke"), thisEdgeElement.getAttribute("fromterminator"), thisEdgeElement.getAttribute("toterminator"), thisEdgeElement.getAttribute("hidden"));
		if (thisStyle != null)
		{
			thisEdge.setStyle(thisStyle);
		}
		return thisEdge;
	}

	/**
	 * Make list of model edges
	 * 
	 * @param thisDocument
	 *        DOM document
	 * @return list of edges
	 */
	private List<IEdge> toEdges(final Document thisDocument)
	{
		List<IEdge> thisEdgeList = null;
		final NodeList theseChildren = thisDocument.getElementsByTagName("edge");
		for (int i = 0; i < theseChildren.getLength(); i++)
		{
			final Node thisNode = theseChildren.item(i);
			final Element thisEdgeElement = (Element) thisNode;
			final MutableEdge thisEdge = toEdge(thisEdgeElement);
			if (thisEdgeList == null)
			{
				thisEdgeList = new ArrayList<IEdge>();
			}
			thisEdgeList.add(thisEdge);
		}
		return thisEdgeList;
	}

	/**
	 * Make settings
	 * 
	 * @param thisDocument
	 *        DOM document
	 * @return settings
	 */
	static private Settings toSettings(final Document thisDocument)
	{
		final Settings theseSettings = new Settings();

		// T O P
		Element thisElement = thisDocument.getDocumentElement();
		if (thisElement != null && thisElement.getNodeName().equals("treebolic"))
		{
			String thisAttribute = thisElement.getAttribute("toolbar");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theHasToolbarFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("statusbar");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theHasStatusbarFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("popupmenu");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theHasPopUpMenuFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("tooltip");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theHasToolTipFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("tooltip-displays-content");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theToolTipDisplaysContentFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("focus");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theFocus = thisAttribute;
			}
			thisAttribute = thisElement.getAttribute("focus-on-hover");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theFocusOnHoverFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("xmoveto");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theXMoveTo = Double.valueOf(thisAttribute).doubleValue();
			}
			thisAttribute = thisElement.getAttribute("ymoveto");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theYMoveTo = Double.valueOf(thisAttribute).doubleValue();
			}
			thisAttribute = thisElement.getAttribute("xshift");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theXShift = Double.valueOf(thisAttribute).doubleValue();
			}
			thisAttribute = thisElement.getAttribute("yshift");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theYShift = Double.valueOf(thisAttribute).doubleValue();
			}
		}

		// T R E E
		thisElement = DocumentAdapter.getFirstElementByTagName(thisDocument.getDocumentElement(), "tree");
		if (thisElement != null)
		{
			// img
			final Element thisImageElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisElement, "img");
			if (thisImageElement != null)
			{
				final String thisSrc = thisImageElement.getAttribute("src");
				if (thisSrc != null && !thisSrc.isEmpty())
				{
					theseSettings.theBackgroundImage = thisSrc;
				}
			}

			// colors
			Color thisColor = Utils.stringToColor(thisElement.getAttribute("backcolor"));
			if (thisColor != null)
			{
				theseSettings.theBackColor = thisColor;
			}
			thisColor = Utils.stringToColor(thisElement.getAttribute("forecolor"));
			if (thisColor != null)
			{
				theseSettings.theForeColor = thisColor;
			}

			// attributes
			String thisAttribute;
			thisAttribute = thisElement.getAttribute("orientation");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theOrientation = thisAttribute;
			}
			thisAttribute = thisElement.getAttribute("expansion");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theExpansion = Double.valueOf(thisAttribute).doubleValue();
			}
			thisAttribute = thisElement.getAttribute("sweep");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theSweep = Double.valueOf(thisAttribute).doubleValue();
			}
			thisAttribute = thisElement.getAttribute("preserve-orientation");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.thePreserveOrientationFlag = Boolean.valueOf(thisAttribute);
			}
			thisAttribute = thisElement.getAttribute("fontface");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theFontFace = thisAttribute;
			}
			thisAttribute = thisElement.getAttribute("fontsize");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theFontSize = Integer.valueOf(thisAttribute).intValue();
			}
			thisAttribute = thisElement.getAttribute("fontsizestep");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theFontSizeStep = Integer.valueOf(thisAttribute).intValue();
			}
		}

		// N O D E S
		thisElement = DocumentAdapter.getFirstElementByTagName(thisDocument.getDocumentElement(), "nodes");
		if (thisElement != null)
		{
			// img
			final Element thisImageElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisElement, "img");
			if (thisImageElement != null)
			{
				final String thisSrc = thisImageElement.getAttribute("src");
				if (thisSrc != null && !thisSrc.isEmpty())
				{
					theseSettings.theDefaultNodeImage = thisSrc;
				}
			}

			// colors
			Color thisColor = Utils.stringToColor(thisElement.getAttribute("backcolor"));
			if (thisColor != null)
			{
				theseSettings.theNodeBackColor = thisColor;
			}
			thisColor = Utils.stringToColor(thisElement.getAttribute("forecolor"));
			if (thisColor != null)
			{
				theseSettings.theNodeForeColor = thisColor;
			}
		}

		// E D G E S
		thisElement = DocumentAdapter.getFirstElementByTagName(thisDocument.getDocumentElement(), "edges");
		if (thisElement != null)
		{
			// img
			final Element thisImageElement = DocumentAdapter.getFirstLevel1ElementByTagName(thisElement, "img");
			if (thisImageElement != null)
			{
				final String thisSrc = thisImageElement.getAttribute("src");
				if (thisSrc != null && !thisSrc.isEmpty())
				{
					theseSettings.theDefaultEdgeImage = thisSrc;
				}
			}

			// hyperbolic edges
			final String thisAttribute = thisElement.getAttribute("hyperbolic");
			if (thisAttribute != null && !thisAttribute.isEmpty())
			{
				theseSettings.theHyperbolicEdgesFlag = Boolean.valueOf(thisAttribute);
			}
		}

		// D E F A U L T . T R E E . E D G E
		thisElement = DocumentAdapter.getFirstElementByTagName(thisDocument.getDocumentElement(), "default.treeedge");
		if (thisElement != null)
		{
			// img
			final Element thisImageElement = DocumentAdapter.getFirstElementByTagName(thisElement, "img");
			if (thisImageElement != null)
			{
				final String thisSrc = thisImageElement.getAttribute("src");
				if (thisSrc != null && !thisSrc.isEmpty())
				{
					theseSettings.theDefaultTreeEdgeImage = thisSrc;
				}
			}

			// color
			final Color thisColor = Utils.stringToColor(thisElement.getAttribute("color"));
			if (thisColor != null)
			{
				theseSettings.theTreeEdgeColor = thisColor;
			}

			// style
			final Integer thisStyle = Utils.parseStyle(thisElement.getAttribute("stroke"), thisElement.getAttribute("fromterminator"), thisElement.getAttribute("toterminator"), thisElement.getAttribute("hidden"));
			if (thisStyle != null)
			{
				theseSettings.theTreeEdgeStyle = thisStyle;
			}
		}

		// D E F A U L T . E D G E
		thisElement = DocumentAdapter.getFirstElementByTagName(thisDocument.getDocumentElement(), "default.edge");
		if (thisElement != null)
		{
			// img
			final Element thisImageElement = DocumentAdapter.getFirstElementByTagName(thisElement, "img");
			if (thisImageElement != null)
			{
				final String thisSrc = thisImageElement.getAttribute("src");
				if (thisSrc != null && !thisSrc.isEmpty())
				{
					theseSettings.theDefaultEdgeImage = thisSrc;
				}
			}

			// color
			final Color thisColor = Utils.stringToColor(thisElement.getAttribute("color"));
			if (thisColor != null)
			{
				theseSettings.theEdgeColor = thisColor;
			}

			// style
			final Integer thisStyle = Utils.parseStyle(thisElement.getAttribute("stroke"), thisElement.getAttribute("fromterminator"), thisElement.getAttribute("toterminator"), thisElement.getAttribute("hidden"));
			if (thisStyle != null)
			{
				theseSettings.theEdgeStyle = thisStyle;
			}
		}

		// D E F A U L T . E D G E
		List<MenuItem> thisMenuItemList = null;
		final NodeList theseChildren = thisDocument.getElementsByTagName("menuitem");
		for (int i = 0; i < theseChildren.getLength(); i++)
		{
			final Node thisNode = theseChildren.item(i);
			thisElement = (Element) thisNode;
			final MenuItem thisMenuItem = DocumentAdapter.toMenuItem(thisElement);

			if (thisMenuItemList == null)
			{
				thisMenuItemList = new ArrayList<MenuItem>();
			}
			thisMenuItemList.add(thisMenuItem);
		}
		theseSettings.theMenu = thisMenuItemList;

		return theseSettings;
	}

	/**
	 * Make menu item
	 * 
	 * @param thisElement
	 *        menu item DOM element
	 * @return menu item
	 */
	static private MenuItem toMenuItem(final Element thisElement)
	{
		final MenuItem thisMenuItem = new MenuItem();
		Utils.parseMenuItem(thisMenuItem, thisElement.getAttribute("action"), thisElement.getAttribute("match-scope"), thisElement.getAttribute("match-mode"));

		// match target
		final String thisAttribute = thisElement.getAttribute("match-target");
		thisMenuItem.theMatchTarget = thisAttribute;

		// label
		final Element thisLabelElement = DocumentAdapter.getFirstElementByTagName(thisElement, "label");
		if (thisLabelElement != null)
		{
			thisMenuItem.theLabel = thisLabelElement.getTextContent();
		}

		// link
		final Element thisLinkElement = DocumentAdapter.getFirstElementByTagName(thisElement, "a");
		if (thisLinkElement != null)
		{
			thisMenuItem.theLink = thisLinkElement.getAttribute("href");
			thisMenuItem.theTarget = thisLinkElement.getAttribute("target");
		}

		return thisMenuItem;
	}

	// H E L P E R S

	/**
	 * Find DOM element with given tag
	 * 
	 * @param thisElement
	 *        starting DOM element
	 * @param thisTagName
	 *        tag
	 * @return DOM element if found, null if none
	 */
	static private Element getFirstElementByTagName(final Element thisElement, final String thisTagName)
	{
		if (thisElement != null)
		{
			final NodeList thisList = thisElement.getElementsByTagName(thisTagName);
			if (thisList.getLength() > 0)
				return (Element) thisList.item(0);
		}
		return null;
	}

	/**
	 * Find DOM element with given tag among first level children
	 * 
	 * @param thisElement
	 *        starting DOM element
	 * @param thisTagName
	 *        tag
	 * @return DOM element if found, null if none
	 */
	static private Element getFirstLevel1ElementByTagName(final Element thisElement, final String thisTagName)
	{
		final List<Element> theseChildElements = DocumentAdapter.getLevel1ChildElementsByTagName(thisElement, thisTagName);
		if (!theseChildElements.isEmpty())
			return theseChildElements.get(0);
		return null;
	}

	/**
	 * Find DOM elements with given tag among first level children
	 * 
	 * @param thisElement
	 *        starting DOM element
	 * @param thisTagName
	 *        tag
	 * @return DOM element if found, null if none
	 */
	static private List<Element> getLevel1ChildElementsByTagName(final Element thisElement, final String thisTagName)
	{
		final ArrayList<Element> thisList = new ArrayList<Element>();
		final NodeList theseChildren = thisElement.getChildNodes();
		for (int i = 0; i < theseChildren.getLength(); i++)
		{
			final Node thisNode = theseChildren.item(i);
			if (thisNode instanceof Element)
			{
				final Element thisChildElement = (Element) thisNode;
				if (thisChildElement.getTagName().equals(thisTagName))
				{
					thisList.add(thisChildElement);
				}
			}
		}
		return thisList;
	}
}
