/**
 * Title: Treebolic provider
 * Description: Treebolic Provider
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */

package treebolic.provider.xml.fast;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Stack;

import treebolic.control.Mounter;
import treebolic.model.IEdge;
import treebolic.model.MenuItem;
import treebolic.model.Model;
import treebolic.model.MountPoint;
import treebolic.model.MutableEdge;
import treebolic.model.MutableNode;
import treebolic.model.Node;
import treebolic.model.Tree;
import treebolic.model.Utils;

/**
 * Light-weight XML parser for treebolic
 * 
 * @author Bernard Bou
 */
class Parser
{
	// D A T A

	/**
	 * Source URL
	 */
	private URL theURL;

	/**
	 * Provider
	 */
	private Provider theProvider;

	// model

	/**
	 * Root node
	 */
	private MutableNode theRoot = null;

	/**
	 * Edge list
	 */
	private List<IEdge> theEdges = null;

	/**
	 * Id to node map
	 */
	private final Hashtable<String, MutableNode> theNodesById = new Hashtable<String, MutableNode>();

	// settings

	/**
	 * Settings
	 */
	private treebolic.model.Settings theSettings = new treebolic.model.Settings();

	/**
	 * Menu
	 */
	private List<MenuItem> theMenu;

	/**
	 * Menu item
	 */
	private MenuItem theMenuItem;

	// reader

	/**
	 * Reader
	 */
	private BufferedReader theReader;

	// scope

	/**
	 * Scope types
	 */
	private static enum Scope
	{
		NONE, TREEBOLIC, TREE, NODES, NODE, TREEEDGE, EDGES, EDGE, DEFAULTTREEEDGE, DEFAULTEDGE, MOUNTPOINT, MENU
	};

	// C O N S T R U C T O R

	/**
	 * Constructor
	 * 
	 * @param thisURL
	 *        source url
	 * @param thisProvider
	 *        provider
	 */
	public Parser(final URL thisURL, final Provider thisProvider)
	{
		this.theRoot = null;
		this.theEdges = null;
		this.theSettings = null;

		this.theURL = thisURL;
		this.theProvider = thisProvider;

		try
		{
			this.theReader = new BufferedReader(new InputStreamReader(this.theURL.openStream(), "UTF8"));
		}
		catch (final SecurityException e)
		{
			System.err.println("Xml parser: Access refused for " + thisURL.toString() + ":" + e.getMessage());
			e.printStackTrace();
		}
		catch (final FileNotFoundException e)
		{
			System.err.println("Xml parser: File not found on " + thisURL.toString() + ":" + e.getMessage());
			return;
		}
		catch (final Throwable e)
		{
			System.err.println("Xml parser: Exception on " + thisURL.toString() + ":" + e.getMessage());
			e.printStackTrace();
			return;
		}
		return;
	}

	/**
	 * Whether reader is valid
	 * 
	 * @return true if reader is valid
	 */
	public boolean isValid()
	{
		return this.theReader != null;
	}

	// R E A D E R
	// implements basic PushbackReader with one-character lookahead capabilities

	/**
	 * Character lookahead
	 */
	private int lookahead = -1;

	/**
	 * Read character
	 * 
	 * @return read character
	 */
	private int read()
	{
		int c;
		if (this.lookahead != -1)
		{
			c = this.lookahead;
			this.lookahead = -1;
			return c;
		}
		try
		{
			c = this.theReader.read();
			return c;
		}
		catch (final IOException e)
		{
			return -1;
		}
	}

	/**
	 * Unread (=push back) character
	 * 
	 * @param c
	 *        character
	 */
	private void unread(final int c)
	{
		this.lookahead = c;
	}

	// T A G . A N D . T E X T . E X T R A C T I O N

	/**
	 * Whether more tags are available
	 * 
	 * @return true if more tags are available
	 */
	private boolean hasMoreTags()
	{
		int c;
		do
		{
			c = read();
			if (c == -1)
				return false;
		}
		while (c != '<');
		unread(c);
		return true;
		// out assert : next read yields '<'
	}

	/**
	 * Get next tag string
	 * 
	 * @return next tag string
	 */
	private String getNextTag()
	{
		// in assert : next read yields '<'
		// eat '<'
		int c = read();
		if (c != '<')
			// unexpected
			return null;

		// accumulate
		final StringBuffer thisBuffer = new StringBuffer();
		while ((c = read()) != -1)
		{
			if (c == '>')
				// out assert : next read is past '>'
				return thisBuffer.toString();
			thisBuffer.append((char) c);

			// CDATA
			if (thisBuffer.length() == 8 && thisBuffer.indexOf("![CDATA[", 0) != -1)
			{
				while ((c = read()) != -1)
				{
					thisBuffer.append((char) c);
					if (c == ']')
					{
						break;
					}
				}
				if (c == -1)
				{
					System.err.println("Xml parser: Unterminated CDATA: " + thisBuffer.toString());
				}
			}
		}
		// unexpected EOF
		return null;
	}

	/**
	 * Get tag text
	 * 
	 * @return tag text
	 */
	private String getText()
	{
		// between thisTag.opening and closing tag
		final StringBuffer thisBuffer = new StringBuffer();
		for (int c; (c = read()) != -1;)
		{
			if (c == '<')
			{
				// out assert : next read yields '<'
				unread(c);
				return Entities.unescapeXml(thisBuffer.toString());
			}
			thisBuffer.append((char) c);
		}
		// unexpected EOF
		return null;
	}

	// P A R S E . T A G S

	/**
	 * Parse <treebolic> tag
	 * 
	 * @param thisTag
	 *        tag string
	 */
	private void parseTreebolicTag(final Tag thisTag)
	{
		String thisParam;
		thisParam = thisTag.getAttribute("toolbar");
		if (thisParam != null)
		{
			this.theSettings.theHasToolbarFlag = Boolean.valueOf(thisParam);
		}
		thisParam = thisTag.getAttribute("statusbar");
		if (thisParam != null)
		{
			this.theSettings.theHasStatusbarFlag = Boolean.valueOf(thisParam);
		}
		thisParam = thisTag.getAttribute("popupmenu");
		if (thisParam != null)
		{
			this.theSettings.theHasPopUpMenuFlag = Boolean.valueOf(thisParam);
		}
		thisParam = thisTag.getAttribute("tooltip");
		if (thisParam != null)
		{
			this.theSettings.theHasToolTipFlag = Boolean.valueOf(thisParam);
		}
		thisParam = thisTag.getAttribute("tooltip-displays-content");
		if (thisParam != null)
		{
			this.theSettings.theToolTipDisplaysContentFlag = Boolean.valueOf(thisParam);
		}
		thisParam = thisTag.getAttribute("focus");
		if (thisParam != null)
		{
			this.theSettings.theFocus = thisParam;
		}
		thisParam = thisTag.getAttribute("focus-on-hover");
		if (thisParam != null)
		{
			this.theSettings.theFocusOnHoverFlag = Boolean.valueOf(thisParam);
		}
		thisParam = thisTag.getAttribute("xmoveto");
		if (thisParam != null)
		{
			this.theSettings.theXMoveTo = Double.valueOf(thisParam).doubleValue();
		}
		thisParam = thisTag.getAttribute("ymoveto");
		if (thisParam != null)
		{
			this.theSettings.theYMoveTo = Double.valueOf(thisParam).doubleValue();
		}
		thisParam = thisTag.getAttribute("xshift");
		if (thisParam != null)
		{
			this.theSettings.theXShift = Double.valueOf(thisParam).doubleValue();
		}
		thisParam = thisTag.getAttribute("yshift");
		if (thisParam != null)
		{
			this.theSettings.theYShift = Double.valueOf(thisParam).doubleValue();
		}

	}

	/**
	 * Parse <tree> tag
	 * 
	 * @param thisTag
	 *        tag string
	 */
	private void parseTreeTag(final Tag thisTag)
	{
		Color thisColor;
		thisColor = Utils.stringToColor(thisTag.getAttribute("backcolor"));
		if (thisColor != null)
		{
			this.theSettings.theBackColor = thisColor;
		}
		thisColor = Utils.stringToColor(thisTag.getAttribute("forecolor"));
		if (thisColor != null)
		{
			this.theSettings.theForeColor = thisColor;
		}
		String thisParameter;
		thisParameter = thisTag.getAttribute("orientation");
		if (thisParameter != null)
		{
			this.theSettings.theOrientation = thisParameter;
		}
		thisParameter = thisTag.getAttribute("expansion");
		if (thisParameter != null)
		{
			this.theSettings.theExpansion = Double.valueOf(thisParameter).doubleValue();
		}
		thisParameter = thisTag.getAttribute("sweep");
		if (thisParameter != null)
		{
			this.theSettings.theSweep = Double.valueOf(thisParameter).doubleValue();
		}
		thisParameter = thisTag.getAttribute("preserve-orientation");
		if (thisParameter != null)
		{
			this.theSettings.thePreserveOrientationFlag = Boolean.valueOf(thisParameter);
		}
		thisParameter = thisTag.getAttribute("fontface");
		if (thisParameter != null)
		{
			this.theSettings.theFontFace = thisParameter;
		}
		thisParameter = thisTag.getAttribute("fontsize");
		if (thisParameter != null)
		{
			this.theSettings.theFontSize = Integer.valueOf(thisParameter).intValue();
		}
		thisParameter = thisTag.getAttribute("fontsizestep");
		if (thisParameter != null)
		{
			this.theSettings.theFontSizeStep = Integer.valueOf(thisParameter).intValue();
		}
	}

	/**
	 * Parse <nodes> tag
	 * 
	 * @param thisTag
	 *        tag string
	 */
	private void parseNodesTag(final Tag thisTag)
	{
		Color thisColor;
		thisColor = Utils.stringToColor(thisTag.getAttribute("backcolor"));
		if (thisColor != null)
		{
			this.theSettings.theNodeBackColor = thisColor;
		}
		thisColor = Utils.stringToColor(thisTag.getAttribute("forecolor"));
		if (thisColor != null)
		{
			this.theSettings.theNodeForeColor = thisColor;
		}
	}

	/**
	 * Parse <nodesc> tag
	 * 
	 * @param thisTag
	 *        tag string
	 */
	private void parseEdgesTag(final Tag thisTag)
	{
		String thisParam;
		thisParam = thisTag.getAttribute("hyperbolic");
		if (thisParam != null)
		{
			this.theSettings.theHyperbolicEdgesFlag = Boolean.valueOf(thisParam);
		}
	}

	/**
	 * Parse <edge> tag
	 * 
	 * @param thisTag
	 *        tag string
	 * @return edge
	 */
	private MutableEdge parseEdgeTag(final Tag thisTag)
	{
		final String thisFromId = thisTag.getAttribute("from");
		final MutableNode thisFrom = this.theNodesById.get(thisFromId);

		final String thisToId = thisTag.getAttribute("to");
		final MutableNode thisTo = this.theNodesById.get(thisToId);

		// make new edge
		if (thisFrom != null && thisTo != null)
		{
			final MutableEdge thisEdge = new MutableEdge(thisFrom, thisTo);

			final Color thisColor = Utils.stringToColor(thisTag.getAttribute("color"));
			if (thisColor != null)
			{
				thisEdge.setColor(thisColor);
			}
			final Integer thisStyle = Utils.parseStyle(thisTag.getAttribute("stroke"), thisTag.getAttribute("toterminator"), thisTag.getAttribute("fromterminator"), thisTag.getAttribute("hidden"));
			if (thisStyle != null && thisStyle != 0)
			{
				thisEdge.setStyle(thisStyle);
			}
			return thisEdge;
		}
		else
		{
			System.err.println("Xml parser: Malformed edge:" + thisTag);
		}
		return null;
	}

	// P A R S E

	/**
	 * Parse model
	 * 
	 * @return model, null if fail
	 */
	public Model parseModel()
	{
		final Tree thisTree = parse(true);
		return thisTree != null ? new Model(thisTree, this.theSettings) : null;
	}

	/**
	 * Parse tree
	 * 
	 * @return tree, null if fail
	 */
	public Tree parseTree()
	{
		return parse(false);
	}

	/**
	 * Parse. Settings are built as a side-effect.
	 * 
	 * @param doSettings
	 *        whether to consider or discard settings
	 * @return graph if successful
	 */
	private Tree parse(final boolean doSettings)
	{
		this.theSettings = doSettings ? new treebolic.model.Settings() : null;
		final Stack<MutableNode> thisNodeStack = new Stack<MutableNode>();
		MutableEdge thisEdge = null;
		Scope thisScope = Scope.NONE;
		String thisCData = null;
		boolean mountNow = false;

		while (hasMoreTags())
		{
			// tag string
			final String thisTagString = getNextTag();
			if (thisTagString == null)
			{
				break;
			}

			// parse tag
			final Tag thisTag = Tag.makeTag(thisTagString);
			if (thisTag == null)
			{
				System.err.println("UNPARSED TAG:" + thisTagString);
				continue;
			}

			// N O D E
			// (changes scope)
			if (thisTag.theName.equals("node"))
			{
				if (thisTag.open)
				{
					final String thisId = thisTag.getAttribute("id");
					MutableNode thisNode;
					if (!thisNodeStack.empty())
					{
						final Node thisParent = thisNodeStack.peek();
						thisNode = new MutableNode(thisParent, thisId);
					}
					else
					{
						thisNode = new MutableNode(null, thisId);
						this.theRoot = thisNode;
					}
					this.theNodesById.put(thisId, thisNode);

					// attributes
					final Color thisForeColor = Utils.stringToColor(thisTag.getAttribute("forecolor"));
					thisNode.setForeColor(thisForeColor);
					final Color thisBackColor = Utils.stringToColor(thisTag.getAttribute("backcolor"));
					thisNode.setBackColor(thisBackColor);
					final String thisWeight = thisTag.getAttribute("weight");
					if (thisWeight != null && thisWeight.isEmpty())
					{
						thisNode.setWeight(-Double.valueOf(thisWeight).doubleValue());
					}

					// tracking
					if (thisTag.unclosed)
					{
						thisNodeStack.push(thisNode);
						thisScope = Scope.NODE;
					}
					else
					{
						// standalone
						thisScope = thisNodeStack.isEmpty() ? Scope.NONE : Scope.NODE;
					}
				}
				else
				{
					// close
					thisNodeStack.pop();
					thisScope = thisNodeStack.isEmpty() ? Scope.NONE : Scope.NODE;
				}
			}

			// L A B E L
			// (does not change scope)
			// (scope dependent)
			else if (thisTag.theName.equals("label"))
			{
				if (thisTag.open)
				{
					switch (thisScope)
					{
					case MENU:
						if (doSettings)
						{
							final String thisLabel = getText();
							this.theMenuItem.theLabel = thisLabel != null ? thisLabel : "?";
						}
						break;

					case NODE:
						if (!thisNodeStack.empty())
						{
							final String thisLabel = getText();
							if (thisLabel != null)
							{
								final MutableNode thisNode = thisNodeStack.peek();
								thisNode.setLabel(thisLabel);
							}
							break;
						}
					case TREEEDGE:
						if (!thisNodeStack.empty())
						{
							final String thisLabel = getText();
							if (thisLabel != null)
							{
								final MutableNode thisNode = thisNodeStack.peek();
								thisNode.setEdgeLabel(thisLabel);
							}
							break;
						}

					case EDGE:
						final String thisLabel = getText();
						if (thisLabel != null && thisEdge != null)
						{
							thisEdge.setLabel(thisLabel);
						}
						break;

					default:
						System.err.println("Xml parser: Tag error: " + thisTag + " scope=" + thisScope);
					}
				}
				else
				{
					// get embedded cdata text
					if (thisCData != null)
					{
						if (!thisNodeStack.empty())
						{
							final MutableNode thisNode = thisNodeStack.peek();
							thisNode.setLabel(thisCData);
						}
					}
					// invalidate cdata
					thisCData = null;
				}

			}

			// C O N T E N T
			// (does not change scope)
			else if (thisTag.theName.equals("content"))
			{
				if (thisTag.open)
				{
					if (!thisNodeStack.empty())
					{
						// get content text
						final String thisContent = getText();
						if (thisContent != null)
						{
							final MutableNode thisNode = thisNodeStack.peek();
							thisNode.setContent(thisContent);
						}
					}
				}
				else
				{
					// get embedded cdata text
					if (thisCData != null)
					{
						if (!thisNodeStack.empty())
						{
							final MutableNode thisNode = thisNodeStack.peek();
							thisNode.setContent(thisCData);
						}
					}
					// invalidate cdata
					thisCData = null;
				}
			}

			// I M A G E
			// (does not change scope)
			// (scope dependent)
			else if (thisTag.theName.equals("img"))
			{
				if (thisTag.open)
				{
					final String thisParam = thisTag.getAttribute("src");
					if (thisParam != null)
					{
						switch (thisScope)
						{
						case NODE:
							if (!thisNodeStack.empty())
							{
								final MutableNode thisNode = thisNodeStack.peek();
								thisNode.setImageFile(thisParam);
							}
							break;

						case TREEEDGE:
							if (!thisNodeStack.empty())
							{
								final MutableNode thisNode = thisNodeStack.peek();
								thisNode.setEdgeImageFile(thisParam);
							}
							break;

						case EDGE:
							if (thisEdge != null)
							{
								thisEdge.setImageFile(thisParam);
							}
							break;

						case DEFAULTTREEEDGE:
							if (doSettings)
							{
								this.theSettings.theDefaultTreeEdgeImage = thisParam;
							}
							break;

						case DEFAULTEDGE:
							if (doSettings)
							{
								this.theSettings.theDefaultEdgeImage = thisParam;
							}
							break;

						case NODES:
							if (doSettings)
							{
								this.theSettings.theDefaultNodeImage = thisParam;
							}
							break;

						case TREE:
							if (doSettings)
							{
								this.theSettings.theBackgroundImage = thisParam;
							}
							break;

						default:
							System.err.println("Xml parser: Tag error: " + thisTag + " scope=" + thisScope);
						}
					}
				}
			}

			// L I N K
			// (does not change scope)
			// (scope dependent)
			else if (thisTag.theName.equals("a"))
			{
				if (thisTag.open)
				{
					switch (thisScope)
					{
					case MOUNTPOINT:
						if (!thisNodeStack.empty())
						{
							final String thisHrefParam = thisTag.getAttribute("href");
							if (thisHrefParam != null)
							{
								final Node thisNode = thisNodeStack.peek();
								final MountPoint.Mounting thisMountPoint = (MountPoint.Mounting) thisNode.getMountPoint();
								thisMountPoint.theURL = thisHrefParam;
							}
						}
						break;

					case MENU:
						if (doSettings)
						{
							this.theMenuItem.theLink = thisTag.getAttribute("href");
							this.theMenuItem.theTarget = thisTag.getAttribute("target");
						}
						break;

					case NODE:
						if (!thisNodeStack.empty())
						{
							final String thisHrefParam = thisTag.getAttribute("href");
							final String thisTargetParam = thisTag.getAttribute("target");
							if (thisHrefParam != null)
							{
								final MutableNode thisNode = thisNodeStack.peek();
								thisNode.setLink(thisHrefParam);
								thisNode.setTarget(thisTargetParam);
							}
						}
						break;

					default:
						System.err.println("Xml parser: Tag error: " + thisTag + " scope=" + thisScope);
					}
				}
			}

			// T R E E . E D G E
			// (changes scope)
			else if (thisTag.theName.equals("treeedge"))
			{
				if (thisTag.open)
				{
					if (!thisNodeStack.empty())
					{
						final MutableNode thisNode = thisNodeStack.peek();

						Color thisColor;
						thisColor = Utils.stringToColor(thisTag.getAttribute("color"));
						if (thisColor != null)
						{
							thisNode.setEdgeColor(thisColor);
						}

						final Integer thisStyle = Utils.parseStyle(thisTag.getAttribute("stroke"), thisTag.getAttribute("toterminator"), thisTag.getAttribute("fromterminator"), thisTag.getAttribute("hidden"));
						if (thisStyle != null && thisStyle != 0)
						{
							thisNode.setEdgeStyle(thisStyle);
						}
					}
					thisScope = thisTag.unclosed ? Scope.TREEEDGE : Scope.NODE;
				}
				else
				{
					thisScope = Scope.NODE;
				}
			}

			// N O D E S
			// (changes scope)
			else if (thisTag.theName.equals("nodes"))
			{
				if (thisTag.open)
				{
					if (doSettings)
					{
						parseNodesTag(thisTag);
					}
					thisScope = thisTag.unclosed ? Scope.NODES : Scope.NONE;
				}
				else
				{
					thisScope = Scope.NONE;
				}
			}

			// E D G E
			// (changes scope)
			else if (thisTag.theName.equals("edge"))
			{
				if (thisTag.open)
				{
					thisEdge = parseEdgeTag(thisTag);
					if (thisEdge != null)
					{
						if (this.theEdges == null)
						{
							this.theEdges = new ArrayList<IEdge>();
						}
						this.theEdges.add(thisEdge);
					}
					thisScope = thisTag.unclosed ? Scope.EDGE : Scope.NONE;
				}
				else
				{
					thisScope = Scope.NONE;
				}
			}

			// E D G E S
			// (changes scope)
			else if (thisTag.theName.equals("edges"))
			{
				if (thisTag.open)
				{
					if (doSettings)
					{
						parseEdgesTag(thisTag);
					}
					thisScope = thisTag.unclosed ? Scope.EDGES : Scope.NONE;
				}
				else
				{
					thisScope = Scope.NONE;
				}
			}

			// T R E E
			// (changes scope)
			else if (thisTag.theName.equals("tree"))
			{
				if (thisTag.open)
				{
					if (doSettings)
					{
						parseTreeTag(thisTag);
					}
					thisScope = thisTag.unclosed ? Scope.TREE : Scope.NONE;
				}
				else
				{
					thisScope = Scope.NONE;
				}
			}

			// T R E E B O L I C
			// (changes scope)
			else if (thisTag.theName.equals("treebolic"))
			{
				if (thisTag.open)
				{
					if (doSettings)
					{
						parseTreebolicTag(thisTag);
					}
					thisScope = thisTag.unclosed ? Scope.TREEBOLIC : Scope.NONE;
				}
				else
				{
					thisScope = Scope.NONE;
					break;
				}
			}

			// M O U N T P O I N T
			// (changes scope)
			else if (thisTag.theName.equals("mountpoint"))
			{
				if (thisTag.open)
				{
					thisScope = thisTag.unclosed ? Scope.MOUNTPOINT : Scope.NODE;
					if (!thisNodeStack.empty())
					{
						final Node thisNode = thisNodeStack.peek();
						final MountPoint.Mounting thisMountPoint = new MountPoint.Mounting();
						thisNode.setMountPoint(thisMountPoint);

						// mount now
						final String thisParam = thisTag.getAttribute("now");
						mountNow = thisParam != null && thisParam.equalsIgnoreCase("true");
					}
				}
				else
				{
					if (!thisNodeStack.empty())
					{
						final Node thisNode = thisNodeStack.peek();
						final MountPoint.Mounting thisMountingPoint = (MountPoint.Mounting) thisNode.getMountPoint();
						if (thisMountingPoint != null)
						{
							// mount now ?
							if (mountNow)
							{
								// mount
								final Tree thisTree = this.theProvider.makeTree(thisMountingPoint.theURL, true);
								if (thisTree != null)
								{
									Mounter.graft(thisNode, thisTree.getRoot(), this.theEdges, thisTree.getEdges());
								}
							}
							mountNow = false;
						}
					}
					thisScope = Scope.NODE;
				}
			}

			// D E F A U L T . T R E E E D G E
			// (changes scope)
			else if (thisTag.theName.equals("default.treeedge"))
			{
				if (thisTag.open)
				{
					thisScope = Scope.DEFAULTTREEEDGE;
					if (doSettings)
					{
						final Integer thisStyle = Utils.parseStyle(thisTag.getAttribute("stroke"), thisTag.getAttribute("toterminator"), thisTag.getAttribute("fromterminator"), thisTag.getAttribute("hidden"));
						if (thisStyle != null && thisStyle != 0)
						{
							this.theSettings.theTreeEdgeStyle = thisStyle;
						}
						final Color thisColor = Utils.stringToColor(thisTag.getAttribute("color"));
						if (thisColor != null)
						{
							this.theSettings.theTreeEdgeColor = thisColor;
						}
					}
				}
				else
				{
					thisScope = Scope.NONE;
				}
			}

			// D E F A U L T . E D G E
			// (changes scope)
			else if (thisTag.theName.equals("default.edge"))
			{
				if (thisTag.open)
				{
					thisScope = Scope.DEFAULTEDGE;
					if (doSettings)
					{
						final Integer thisStyle = Utils.parseStyle(thisTag.getAttribute("stroke"), thisTag.getAttribute("toterminator"), thisTag.getAttribute("fromterminator"), thisTag.getAttribute("hidden"));
						if (thisStyle != null && thisStyle != 0)
						{
							this.theSettings.theEdgeStyle = thisStyle;
						}
						final Color thisColor = Utils.stringToColor(thisTag.getAttribute("color"));
						if (thisColor != null)
						{
							this.theSettings.theEdgeColor = thisColor;
						}
					}
				}
				else
				{
					thisScope = Scope.NONE;
				}
			}

			// M E N U
			// (changes scope)
			else if (thisTag.theName.equals("menu"))
			{
				if (thisTag.open)
				{
					thisScope = Scope.MENU;
					if (doSettings)
					{
						this.theMenu = new ArrayList<MenuItem>();
					}
				}
				else
				{
					thisScope = Scope.NONE;
					if (doSettings)
					{
						this.theSettings.theMenu = this.theMenu;
					}
				}
			}

			// M E N U . I T E M
			// (does not change scope)
			else if (thisTag.theName.equals("menuitem"))
			{
				if (thisTag.open)
				{
					if (doSettings)
					{
						this.theMenuItem = new MenuItem();
						Utils.parseMenuItem(this.theMenuItem, thisTag.getAttribute("action"), thisTag.getAttribute("match-scope"), thisTag.getAttribute("match-mode"));

						// match target
						final String thisAttribute = thisTag.getAttribute("match-target");
						this.theMenuItem.theMatchTarget = thisAttribute;
					}
					else if (doSettings)
					{
						this.theMenu.add(this.theMenuItem);
					}
				}
			}
			else if (thisTag.theName.equals("tools"))
			{
				// do nothing
			}

			// O T H E R S
			else if (thisTag.theName.startsWith("!"))
			{
				if (thisTagString.startsWith("![CDATA["))
				{
					thisCData = thisTagString.substring(8, thisTagString.length() - 2);
				}
			}
			else if (thisTag.theName.startsWith("?"))
			{
				System.out.println("[?]" + thisTag);
			}
			else
			{
				System.err.println("Xml parser: Unhandled tag: " + thisTag);
				return null;
			}
		}
		return new Tree(this.theRoot, this.theEdges);
	}
}
