/**
 * Title: Treebolic GXL provider
 * Description: Treebolic GXL graph provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use: see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.xml.graph.gxl;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import treebolic.model.MenuItem;
import treebolic.model.MountPoint;
import treebolic.model.MutableEdge;
import treebolic.model.Settings;
import treebolic.model.Utils;
import treebolic.model.graph.Graph;
import treebolic.model.graph.GraphEdge;
import treebolic.model.graph.MutableGraph;
import treebolic.model.graph.MutableGraphNode;
import treebolic.provider.xml.graph.AbstractParser;

/**
 * GXL parser
 * 
 * @author Bernard Bou
 */
public class GxlParser extends treebolic.provider.xml.graph.AbstractParser
{
	/**
	 * Parse graph
	 * 
	 * @return graph if successful
	 */
	@Override
	protected Graph parseGraph(final Document thisDocument)
	{
		final MutableGraph thisGraph = new MutableGraph();

		// map
		final Hashtable<String, MutableGraphNode> theseGraphNodesById = new Hashtable<String, MutableGraphNode>();

		// enumerate nodes
		final NodeList theseNodeNodes = thisDocument.getElementsByTagName("node");
		for (int i = 0; i < theseNodeNodes.getLength(); i++)
		{
			final Node thisNodeNode = theseNodeNodes.item(i);
			final Element thisNodeElement = (Element) thisNodeNode;

			// get id
			final String thisId = thisNodeElement.getAttribute("id");
			if (thisId == null || thisId.isEmpty())
			{
				continue;
			}

			// create node
			final MutableGraphNode thisGraphNode = new MutableGraphNode(thisId);
			theseGraphNodesById.put(thisId, thisGraphNode);

			// add to graph
			thisGraph.add(thisGraphNode);

			// attributes
			final NodeList theseAttrNodes = thisNodeElement.getElementsByTagName("attr");
			for (int j = 0; j < theseAttrNodes.getLength(); j++)
			{
				final Node thisAttrNode = theseAttrNodes.item(j);
				final Element thisAttrElement = (Element) thisAttrNode;
				final String thisAttributeName = thisAttrElement.getAttribute("name");

				if (thisAttributeName.equalsIgnoreCase("label"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisGraphNode.setLabel(thisStringElement.getTextContent());
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("content"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisGraphNode.setContent(thisStringElement.getTextContent());
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("link"))
				{
					final Element thisLocatorElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						final String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						thisGraphNode.setLink(thisHRef);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("img-src"))
				{
					final Element thisLocatorElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						final String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						thisGraphNode.setImageFile(thisHRef);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("mountpoint"))
				{
					final Element thisLocatorElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						final String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						final MountPoint.Mounting thisMountPoint = new MountPoint.Mounting();
						thisMountPoint.theURL = thisHRef;
						thisGraphNode.setMountPoint(thisMountPoint);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("mountpoint-weight"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						final MountPoint.Mounting thisMountPoint = (MountPoint.Mounting) thisGraphNode.getMountPoint();
						if (thisMountPoint != null)
						{
							thisGraphNode.setWeight(Double.valueOf(thisStringElement.getTextContent()));
						}
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("backcolor"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisGraphNode.setBackColor(Utils.stringToColor(thisStringElement.getTextContent()));
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("forecolor"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisGraphNode.setForeColor(Utils.stringToColor(thisStringElement.getTextContent()));
					}
				}
			}
		}

		// EDGES
		final NodeList theseEdgeNodes = thisDocument.getElementsByTagName("edge");
		for (int i = 0; i < theseEdgeNodes.getLength(); i++)
		{
			final Node thisEdgeNode = theseEdgeNodes.item(i);
			final Element thisEdgeElement = (Element) thisEdgeNode;

			// get end ids
			final String thisFromId = thisEdgeElement.getAttribute("from");
			if (thisFromId == null)
			{
				continue;
			}
			final String thisToId = thisEdgeElement.getAttribute("to");
			if (thisToId == null)
			{
				continue;
			}

			// get ends
			final MutableGraphNode thisFromNode = theseGraphNodesById.get(thisFromId);
			if (thisFromNode == null)
			{
				continue;
			}
			final MutableGraphNode thisToNode = theseGraphNodesById.get(thisToId);
			if (thisToNode == null)
			{
				continue;
			}

			// try to determine whether this edge is a tree edge
			Boolean isTreeEdge = null;
			final Element thisTypeElement = AbstractParser.getFirstLevel1ElementByTagName(thisEdgeElement, "type");
			if (thisTypeElement != null)
			{
				final String thisType = thisTypeElement.getAttribute("xlink:href");
				if (thisType != null && !thisType.isEmpty())
				{
					if (thisType.equals("schema.xml#TreeEdge"))
					{
						isTreeEdge = true;
					}
					else if (thisType.equals("schema.xml#TreeEdge"))
					{
						isTreeEdge = false;
					}
				}
			}

			// create edge
			final MutableEdge thisEdge = new MutableEdge(thisFromNode, thisToNode);

			final GraphEdge thisGraphEdge = new GraphEdge(thisFromNode, thisToNode, isTreeEdge);
			thisGraphEdge.setUserData(thisEdge);
			thisGraph.add(thisGraphEdge);

			// attributes
			final NodeList theseAttrNodes = thisEdgeElement.getElementsByTagName("attr");
			for (int j = 0; j < theseAttrNodes.getLength(); j++)
			{
				final Node thisAttrNode = theseAttrNodes.item(j);
				final Element thisAttrElement = (Element) thisAttrNode;
				final String thisAttributeName = thisAttrElement.getAttribute("name");
				if (thisAttributeName.equalsIgnoreCase("label"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisEdge.setLabel(thisStringElement.getTextContent());
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("img-src"))
				{
					final Element thisLocatorElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "locator");
					if (thisLocatorElement != null)
					{
						final String thisHRef = thisLocatorElement.getAttribute("xlink:href");
						thisEdge.setImageFile(thisHRef);
					}
				}
				else if (thisAttributeName.equalsIgnoreCase("color"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisEdge.setColor(Utils.stringToColor(thisStringElement.getTextContent()));
					}
				}
				else if (thisAttributeName.equals("stroke"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisEdge.setStyle(Utils.modifyStyle(thisEdge.getStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.STROKE));
					}
				}
				else if (thisAttributeName.equals("fromterminator"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisEdge.setStyle(Utils.modifyStyle(thisEdge.getStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.FROMTERMINATOR));
					}
				}
				else if (thisAttributeName.equals("toterminator"))
				{
					final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
					if (thisStringElement != null)
					{
						thisEdge.setStyle(Utils.modifyStyle(thisEdge.getStyle(), thisStringElement.getTextContent(), Utils.StyleComponent.TOTERMINATOR));
					}
				}
			}
		}
		return thisGraph;
	}

	/**
	 * Parse settings
	 * 
	 * @return graph if successful
	 */
	@Override
	protected Settings parseSettings(final Document thisDocument)
	{
		final Settings theseSettings = new Settings();
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theHasStatusbarFlag = true;

		final Element thisGraphElement = AbstractParser.getFirstElementByTagName(thisDocument.getDocumentElement(), "graph");
		if (thisGraphElement != null)
		{
			// enumerate attrs
			final List<Element> theseAttrElements = AbstractParser.getLevel1ChildElementsByTagName(thisGraphElement, "attr");
			for (final Element thisAttrElement : theseAttrElements)
			{
				final String thisName = thisAttrElement.getAttribute("name");

				// locator
				final Element thisLocatorElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "locator");
				if (thisLocatorElement != null)
				{
					final String thisValue = thisLocatorElement.getAttribute("xlink:href");
					if (thisName.equals("nodes-img-src"))
					{
						theseSettings.theDefaultNodeImage = thisValue;
					}
					else if (thisName.equals("tree-img-src"))
					{
						theseSettings.theBackgroundImage = thisValue;
					}
					else if (thisName.equals("nodes-default-treeedge-img-src"))
					{
						theseSettings.theDefaultTreeEdgeImage = thisValue;
					}
					else if (thisName.equals("edges-default-edge-img-src"))
					{
						theseSettings.theDefaultEdgeImage = thisValue;
					}
				}

				// string
				final Element thisStringElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "string");
				if (thisStringElement != null)
				{
					final String thisValue = thisStringElement.getTextContent();
					if (thisName.equals("treebolic-toolbar"))
					{
						theseSettings.theHasToolbarFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("treebolic-statusbar"))
					{
						theseSettings.theHasStatusbarFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("treebolic-popupmenu"))
					{
						theseSettings.theHasPopUpMenuFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("treebolic-tooltip"))
					{
						theseSettings.theHasToolTipFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("treebolic-focus-on-hover"))
					{
						theseSettings.theFocusOnHoverFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("tree-backcolor"))
					{
						theseSettings.theBackColor = Utils.stringToColor(thisValue);
					}
					else if (thisName.equals("tree-forecolor"))
					{
						theseSettings.theForeColor = Utils.stringToColor(thisValue);
					}
					else if (thisName.equals("tree-fontface"))
					{
						theseSettings.theFontFace = thisValue;
					}
					else if (thisName.equals("tree-fontsize"))
					{
						theseSettings.theFontSize = Integer.valueOf(thisValue);
					}
					else if (thisName.equals("tree-fontsizestep"))
					{
						theseSettings.theFontSizeStep = Integer.valueOf(thisValue);
					}
					else if (thisName.equals("tree-expansion"))
					{
						theseSettings.theExpansion = Double.valueOf(thisValue);
					}
					else if (thisName.equals("tree-sweep"))
					{
						theseSettings.theSweep = Double.valueOf(thisValue);
					}
					else if (thisName.equals("tree-orientation"))
					{
						theseSettings.theOrientation = thisValue;
					}
					else if (thisName.equals("tree-preserve-orientation"))
					{
						theseSettings.thePreserveOrientationFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("nodes-backcolor"))
					{
						theseSettings.theNodeBackColor = Utils.stringToColor(thisValue);
					}
					else if (thisName.equals("nodes-forecolor"))
					{
						theseSettings.theNodeForeColor = Utils.stringToColor(thisValue);
					}
					else if (thisName.equals("nodes-default-treeedge-color"))
					{
						theseSettings.theTreeEdgeColor = Utils.stringToColor(thisValue);
					}
					else if (thisName.equals("nodes-default-treeedge-stroke"))
					{
						theseSettings.theTreeEdgeStyle = Utils.modifyStyle(theseSettings.theTreeEdgeStyle, thisValue, Utils.StyleComponent.STROKE);
					}
					else if (thisName.equals("nodes-default-treeedge-fromterminator"))
					{
						theseSettings.theTreeEdgeStyle = Utils.modifyStyle(theseSettings.theTreeEdgeStyle, thisValue, Utils.StyleComponent.FROMTERMINATOR);
					}
					else if (thisName.equals("nodes-default-treeedge-toterminator"))
					{
						theseSettings.theTreeEdgeStyle = Utils.modifyStyle(theseSettings.theTreeEdgeStyle, thisValue, Utils.StyleComponent.TOTERMINATOR);
					}
					else if (thisName.equals("edges-hyperbolic"))
					{
						theseSettings.theHyperbolicEdgesFlag = Boolean.valueOf(thisValue);
					}
					else if (thisName.equals("edges-default-edge-color"))
					{
						theseSettings.theEdgeColor = Utils.stringToColor(thisValue);
					}
					else if (thisName.equals("edges-default-edge-fromterminator"))
					{
						theseSettings.theEdgeStyle = Utils.modifyStyle(theseSettings.theEdgeStyle, thisValue, Utils.StyleComponent.FROMTERMINATOR);
					}
					else if (thisName.equals("edges-default-edge-toterminator"))
					{
						theseSettings.theEdgeStyle = Utils.modifyStyle(theseSettings.theEdgeStyle, thisValue, Utils.StyleComponent.TOTERMINATOR);
					}
					else if (thisName.equals("edges-default-edge-stroke"))
					{
						theseSettings.theEdgeStyle = Utils.modifyStyle(theseSettings.theEdgeStyle, thisValue, Utils.StyleComponent.STROKE);
					}
					else
					{
						System.err.println("Glx: Unhandled attribute: " + thisName);
					}
				}

				// tup
				final Element thisTupElement = AbstractParser.getFirstLevel1ElementByTagName(thisAttrElement, "tup");
				if (thisTupElement != null)
				{
					final List<Element> theseTupleStringElements = AbstractParser.getLevel1ChildElementsByTagName(thisTupElement, "string");
					if (theseTupleStringElements != null)
					{
						int k = 0;
						for (final Element thisTupleStringElement : theseTupleStringElements)
						{
							if (thisTupleStringElement != null)
							{
								final String thisValue = thisTupleStringElement.getTextContent();
								if (thisValue == null)
								{
									continue;
								}

								if (thisName.equals("menuitem-action"))
								{
									theseSettings.theMenu = GxlParser.allocate(theseSettings.theMenu, k);
									final MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theAction = Utils.stringToAction(thisValue);
								}
								else if (thisName.equals("menuitem-label"))
								{
									theseSettings.theMenu = GxlParser.allocate(theseSettings.theMenu, k);
									final MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theLabel = thisValue;

								}
								else if (thisName.equals("menuitem-match-target"))
								{
									theseSettings.theMenu = GxlParser.allocate(theseSettings.theMenu, k);
									final MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theMatchTarget = thisValue;

								}
								else if (thisName.equals("menuitem-match-scope"))
								{
									theseSettings.theMenu = GxlParser.allocate(theseSettings.theMenu, k);
									final MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theMatchScope = Utils.stringToScope(thisValue);
								}
								else if (thisName.equals("menuitem-match-mode"))
								{
									theseSettings.theMenu = GxlParser.allocate(theseSettings.theMenu, k);
									final MenuItem thisMenuItem = theseSettings.theMenu.get(k);
									thisMenuItem.theMatchMode = Utils.stringToMode(thisValue);
								}
								k++;
							}
						}
					}
				}
			}
		}
		return theseSettings;
	}

	/**
	 * Allocate menu item
	 * 
	 * @param thatList
	 *        current list of menu items
	 * @param i
	 *        ith item to ensure allocation of
	 * @return menu item list on which get(k) will succeed
	 */
	static private List<MenuItem> allocate(final List<MenuItem> thatList, final int i)
	{
		List<MenuItem> thisList = thatList;
		if (thisList == null)
		{
			thisList = new ArrayList<MenuItem>();
		}
		if (i > thisList.size() - 1)
		{
			thisList.add(new MenuItem());
		}
		return thisList;
	}
}
