/**
 * Title: Treebolic GXML provider
 * Description: Treebolic GXML graph provider
 * Version: 2.x
 * Copyright: (c) 2001-2008
 * Terms of use: see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.provider.xml.graph.gxml;

import java.awt.Color;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import treebolic.model.IEdge;
import treebolic.model.MutableEdge;
import treebolic.model.Settings;
import treebolic.model.Utils;
import treebolic.model.graph.Graph;
import treebolic.model.graph.GraphEdge;
import treebolic.model.graph.MutableGraph;
import treebolic.model.graph.MutableGraphNode;
import treebolic.provider.xml.graph.AbstractParser;

/**
 * GXL parser
 * 
 * @author Bernard Bou
 */
public class GxmlParser extends treebolic.provider.xml.graph.AbstractParser
{
	/**
	 * Parse graph
	 * 
	 * @return graph if successful
	 */
	@Override
	protected Graph parseGraph(final Document thisDocument)
	{
		final MutableGraph thisGraph = new MutableGraph();

		// map
		final Hashtable<String, MutableGraphNode> theseGraphNodesById = new Hashtable<String, MutableGraphNode>();

		// C L A S S E S
		final Map<String, Style> thisStyleMap = GxmlParser.parseClasses(AbstractParser.getFirstElementByTagName(thisDocument.getDocumentElement(), "graph"));

		// N O D E S
		final NodeList theseNodeNodes = thisDocument.getElementsByTagName("node");
		for (int i = 0; i < theseNodeNodes.getLength(); i++)
		{
			final Node thisNodeNode = theseNodeNodes.item(i);
			final Element thisNodeElement = (Element) thisNodeNode;

			// get id
			final String thisId = thisNodeElement.getAttribute("name");
			if (thisId == null || thisId.isEmpty())
			{
				continue;
			}

			// create node
			final MutableGraphNode thisGraphNode = new MutableGraphNode(thisId);
			theseGraphNodesById.put(thisId, thisGraphNode);

			// add to graph
			thisGraph.add(thisGraphNode);

			// label
			final Element thisLabelElement = AbstractParser.getFirstLevel1ElementByTagName(thisNodeElement, "label");
			if (thisLabelElement != null)
			{
				String thisLabel = thisLabelElement.getTextContent();
				thisLabel = thisLabel.trim();
				thisGraphNode.setLabel(thisLabel);
			}

			// link
			final Element thisRefElement = AbstractParser.getFirstElementByTagName(thisNodeElement, "ref");
			if (thisRefElement != null)
			{
				String thisLink = thisRefElement.getAttribute("xlink:href");
				thisLink = thisLink.trim();
				thisGraphNode.setLink(thisLink);
			}

			// class style
			final String thisClass = thisNodeElement.getAttribute("class");
			if (thisClass != null)
			{
				GxmlParser.transferStyle(thisGraphNode, thisStyleMap.get(thisClass));
			}

			// style
			final Element thisStyleElement = AbstractParser.getFirstLevel1ElementByTagName(thisNodeElement, "style");
			if (thisStyleElement != null)
			{
				final Element thisLineElement = AbstractParser.getFirstLevel1ElementByTagName(thisStyleElement, "line");
				if (thisLineElement != null)
				{
					final Color thisColor = GxmlParser.toColor(thisLineElement.getAttribute("colour"));
					if (thisColor != null)
					{
						thisGraphNode.setEdgeColor(thisColor);
					}
				}

				final Element thisFillElement = AbstractParser.getFirstLevel1ElementByTagName(thisStyleElement, "fill");
				if (thisFillElement != null)
				{
					final Color thisColor = GxmlParser.toColor(thisFillElement.getAttribute("colour"));
					if (thisColor != null)
					{
						thisGraphNode.setBackColor(thisColor);
					}
					final String thisString = thisFillElement.getAttribute("xlink:href");
					if (thisString != null)
					{
						thisGraphNode.setImageFile(thisString);
					}
				}
			}
		}

		// E D G E S
		final NodeList theseEdgeNodes = thisDocument.getElementsByTagName("edge");
		for (int i = 0; i < theseEdgeNodes.getLength(); i++)
		{
			final Node thisEdgeNode = theseEdgeNodes.item(i);
			final Element thisEdgeElement = (Element) thisEdgeNode;

			// get end ids
			final String thisFromId = thisEdgeElement.getAttribute("source");
			if (thisFromId == null)
			{
				continue;
			}
			final String thisToId = thisEdgeElement.getAttribute("target");
			if (thisToId == null)
			{
				continue;
			}

			// get ends
			final MutableGraphNode thisFromNode = theseGraphNodesById.get(thisFromId);
			if (thisFromNode == null)
			{
				continue;
			}
			final MutableGraphNode thisToNode = theseGraphNodesById.get(thisToId);
			if (thisToNode == null)
			{
				continue;
			}

			// create edge
			final MutableEdge thisEdge = new MutableEdge(thisFromNode, thisToNode);

			// insert edge into graph
			final GraphEdge thisGraphEdge = new GraphEdge(thisFromNode, thisToNode, null);
			thisGraphEdge.setUserData(thisEdge);
			thisGraph.add(thisGraphEdge);

			// class style
			final String thisClass = thisEdgeElement.getAttribute("class");
			if (thisClass != null)
			{
				GxmlParser.transferStyle(thisEdge, thisStyleMap.get(thisClass));
			}

			// style
			GxmlParser.parseLine(thisEdgeElement, thisEdge);
		}
		return thisGraph;
	}

	/**
	 * Parse settings
	 * 
	 * @return graph if successful
	 */
	@Override
	protected Settings parseSettings(final Document thisDocument)
	{
		final Settings theseSettings = new Settings();
		theseSettings.theHasToolbarFlag = true;
		theseSettings.theHasStatusbarFlag = true;
		return theseSettings;
	}

	// S T Y L E

	/**
	 * Parse edge style
	 * 
	 * @param thisElement
	 *        DOM element to start from
	 * @param thisEdge
	 *        edge
	 */
	static private void parseLine(final Element thisElement, final MutableEdge thisEdge)
	{
		// style
		final Element thisStyleElement = AbstractParser.getFirstLevel1ElementByTagName(thisElement, "style");
		if (thisStyleElement != null)
		{
			// fore (line)
			final Element thisLineElement = AbstractParser.getFirstLevel1ElementByTagName(thisStyleElement, "line");
			if (thisLineElement != null)
			{
				final Color thisColor = GxmlParser.toColor(thisLineElement.getAttribute("colour"));
				if (thisColor != null)
				{
					thisEdge.setColor(thisColor);
				}

				final String thisLineStyle = thisLineElement.getAttribute("linestyle");
				if (thisLineStyle != null)
				{
					int thisStyle = 0;
					if (thisLineStyle != null)
					{
						if (thisLineStyle.equals("solid"))
						{
							thisStyle |= IEdge.SOLID;
						}
						else if (thisLineStyle.equals("dashed"))
						{
							thisStyle |= IEdge.DASH;
						}
						else if (thisLineStyle.equals("dotted"))
						{
							thisStyle |= IEdge.DOT;
						}
					}
					thisEdge.setStyle(thisStyle);
				}
			}
		}
	}

	/**
	 * Parse class styles
	 * 
	 * @param thisElement
	 *        DOM element to start from
	 * @return style map
	 */
	static private Map<String, Style> parseClasses(final Element thisElement)
	{
		final Map<String, Style> thisStyleMap = new HashMap<String, Style>();

		// style
		final Element thisStyleElement = AbstractParser.getFirstLevel1ElementByTagName(thisElement, "style");
		if (thisStyleElement != null)
		{
			// fore (line)
			final List<Element> theseLineElements = AbstractParser.getLevel1ChildElementsByTagName(thisStyleElement, "line");
			if (theseLineElements != null)
			{
				for (final Element thisLineElement : theseLineElements)
				{
					final String thisClass = thisLineElement.getAttribute("class");
					if (thisClass == null)
					{
						continue;
					}
					final String thisTag = thisLineElement.getAttribute("tag");
					if (thisTag == null)
					{
						continue;
					}

					final Color thisColor = GxmlParser.toColor(thisLineElement.getAttribute("colour"));
					if (thisColor != null)
					{
						final Style thisStyle = GxmlParser.getStyle(thisStyleMap, thisClass);
						if (thisTag.equals("node"))
						{
							thisStyle.theForeColor = thisColor;
						}
						else if (thisTag.equals("edge"))
						{
							thisStyle.theEdgeColor = thisColor;
						}
					}
				}
			}

			// back (fill)
			final List<Element> theseFillElements = AbstractParser.getLevel1ChildElementsByTagName(thisStyleElement, "fill");
			if (theseFillElements != null)
			{
				for (final Element thisFillElement : theseFillElements)
				{
					final String thisClass = thisFillElement.getAttribute("class");
					if (thisClass == null)
					{
						continue;
					}
					final String thisTag = thisFillElement.getAttribute("tag");
					if (thisTag == null)
					{
						continue;
					}

					final Color thisColor = GxmlParser.toColor(thisFillElement.getAttribute("colour"));
					if (thisColor != null)
					{
						final Style thisStyle = GxmlParser.getStyle(thisStyleMap, thisClass);
						if (thisTag.equals("node"))
						{
							thisStyle.theBackColor = thisColor;
						}
						else if (thisTag.equals("edge"))
						{
							thisStyle.theEdgeColor = thisColor;
						}
					}

					final String thisString = thisFillElement.getAttribute("xlink:href");
					if (thisString != null)
					{
						final Style thisStyle = GxmlParser.getStyle(thisStyleMap, thisClass);
						if (thisTag.equals("node"))
						{
							thisStyle.theNodeImage = thisString;
						}
						else if (thisTag.equals("edge"))
						{
							thisStyle.theEdgeImage = thisString;
						}
					}
				}
			}
		}
		return thisStyleMap;
	}

	/**
	 * Wrapper for style information
	 * 
	 * @author Bernard Bou
	 */
	static class Style
	{
		/**
		 * Node back color
		 */
		public Color theBackColor;

		/**
		 * Node fore color
		 */
		public Color theForeColor;

		/**
		 * Edge color
		 */
		public Color theEdgeColor;

		/**
		 * Nod image
		 */
		public String theNodeImage;

		/**
		 * Edge image
		 */
		public String theEdgeImage;

		/*
		 * (non-Javadoc)
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString()
		{
			return "nb=" + this.theBackColor + " nf=" + this.theForeColor + " ef=" + this.theEdgeColor + " ni=" + this.theNodeImage + " ei=" + this.theEdgeImage;
		}
	}

	/**
	 * Get style from map, allocate key and null value in map if needed
	 * 
	 * @param thisStyleMap
	 *        style map
	 * @param thisClass
	 *        class key
	 * @return style matching class in map (created if map did not previously contain key)
	 */
	static private Style getStyle(final Map<String, Style> thisStyleMap, final String thisClass)
	{
		Style thisStyle = thisStyleMap.get(thisClass);
		if (thisStyle == null)
		{
			thisStyle = new Style();
			thisStyleMap.put(thisClass, thisStyle);
		}
		return thisStyle;
	}

	/**
	 * Transfer class style to node
	 * 
	 * @param thisNode
	 *        node
	 * @param thisStyle
	 *        style
	 */
	static private void transferStyle(final MutableGraphNode thisNode, final Style thisStyle)
	{
		if (thisStyle == null)
			return;
		if (thisStyle.theBackColor != null)
		{
			thisNode.setBackColor(thisStyle.theBackColor);
		}
		if (thisStyle.theForeColor != null)
		{
			thisNode.setForeColor(thisStyle.theForeColor);
		}
		if (thisStyle.theEdgeColor != null)
		{
			thisNode.setEdgeColor(thisStyle.theEdgeColor);
		}
		if (thisStyle.theNodeImage != null)
		{
			thisNode.setImageFile(thisStyle.theNodeImage);
		}
		if (thisStyle.theEdgeImage != null)
		{
			thisNode.setEdgeImageFile(thisStyle.theEdgeImage);
		}
	}

	/**
	 * Transfer class style to edge
	 * 
	 * @param thisEdge
	 *        edge
	 * @param thisStyle
	 *        style
	 */
	static private void transferStyle(final MutableEdge thisEdge, final Style thisStyle)
	{
		if (thisStyle == null)
			return;
		if (thisStyle.theBackColor != null)
		{
			thisEdge.setColor(thisStyle.theEdgeColor);
		}
		if (thisStyle.theForeColor != null)
		{
			thisEdge.setColor(thisStyle.theEdgeColor);
		}
		if (thisStyle.theEdgeImage != null)
		{
			thisEdge.setImageFile(thisStyle.theEdgeImage);
		}
	}

	/**
	 * Color parsing
	 * 
	 * @param thisColorString
	 *        string to parse
	 * @return color
	 */
	static private Color toColor(final String thisColorString)
	{
		if (thisColorString == null || thisColorString.isEmpty())
			return null;

		if (thisColorString.equals("black"))
			return Color.BLACK;
		else if (thisColorString.equals("white"))
			return Color.WHITE;
		else if (thisColorString.equals("gray"))
			return Color.GRAY;
		else if (thisColorString.equals("darkGray"))
			return Color.DARK_GRAY;
		else if (thisColorString.equals("lightGray"))
			return Color.LIGHT_GRAY;
		else if (thisColorString.equals("red"))
			return Color.RED;
		else if (thisColorString.equals("blue"))
			return Color.BLUE;
		else if (thisColorString.equals("green"))
			return Color.GREEN;
		else if (thisColorString.equals("cyan"))
			return Color.CYAN;
		else if (thisColorString.equals("magenta"))
			return Color.MAGENTA;
		else if (thisColorString.equals("yellow"))
			return Color.YELLOW;
		else if (thisColorString.equals("orange"))
			return Color.ORANGE;
		else if (thisColorString.equals("pink"))
			return Color.PINK;

		return Utils.stringToColor(thisColorString.substring(1));
	}
}
