/**
 * Title :	Treebolic generator
 * Description: Treebolic generator
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.site;

import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.util.StringTokenizer;

import org.w3c.dom.Document;

import sun.net.ftp.FtpLoginException;
import treebolic.Messages;

/**
 * FTP extended client
 * 
 * @author Bernard Bou
 */
class FtpClientX extends FtpClient
{
	/**
	 * MKDIR a directory on a remote FTP server
	 * 
	 * @param remoteDirectory
	 *        remote directory to create
	 * @throws IOException
	 */
	public void mkdir(final String remoteDirectory) throws IOException
	{
		try
		{
			issueCommandCheck("MKD " + remoteDirectory); //$NON-NLS-1$
		}
		catch (final IOException e1)
		{
			// Well, "/" might not be the file delimitor for this particular ftp server,
			// so let's try a series of "cd" and "mkd" commands to get to the right place.

			String pathElement = null;

			// cd path
			final StringTokenizer t = new StringTokenizer(remoteDirectory, "/"); //$NON-NLS-1$
			while (t.hasMoreElements())
			{
				pathElement = t.nextToken();

				// cd this pathElement
				try
				{
					cd(pathElement);
				}
				catch (final IOException e2)
				{
					try
					{
						issueCommandCheck("MKDIR " + pathElement); //$NON-NLS-1$
						cd(pathElement);
					}
					catch (final IOException e3)
					{
						// giving up.
						throw e1;
					}
				}
			}
		}
	}

	/**
	 * MKDIR a directory on a remote FTP server
	 * 
	 * @throws IOException
	 */
	public void passive() throws IOException
	{
		try
		{
			issueCommandCheck("PASV "); //$NON-NLS-1$
		}
		catch (final IOException e1)
		{
			// do nothing
		}
	}
}

/**
 * FTP site maker
 * 
 * @author Bernard Bou
 */
public class FtpSiteMaker extends SiteMaker
{
	/**
	 * Server
	 */
	private final String theServer;

	/**
	 * Server directory
	 */
	private final String theDirectory;

	/**
	 * Server login
	 */
	private final String theLogin;

	/**
	 * Server login password
	 */
	private final String thePassword;

	/**
	 * FTP client
	 */
	private final FtpClientX theClient;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 * 
	 * @param thisDocument
	 *        DOM document (source)
	 * @param thisRepository
	 *        repository (source)
	 * @param thisImageRepository
	 *        image repository (source)
	 * @param thisHtmlFile
	 *        HTML file (target)
	 * @param thisXmlFile
	 *        XML file (target)
	 * @param thisTitle
	 *        HTML page title (target)
	 * @param thisServer
	 *        server (target)
	 * @param thisDirectory
	 *        server directory (target)
	 * @param thisLogin
	 *        server login (target)
	 * @param thisPassword
	 *        server login password (target)
	 * @throws URISyntaxException
	 * @throws IOException
	 */
	public FtpSiteMaker(final Document thisDocument, final String thisRepository, final String thisImageRepository, final String thisHtmlFile, final String thisXmlFile, final String thisTitle, final String thisServer, final String thisDirectory,
			final String thisLogin, final String thisPassword) throws URISyntaxException, IOException
	{
		super(thisDocument, thisRepository, thisImageRepository, thisHtmlFile, thisXmlFile, thisTitle);

		this.theServer = thisServer;
		this.theDirectory = thisDirectory;
		this.theLogin = thisLogin;
		this.thePassword = thisPassword;
		this.theClient = new FtpClientX();
	}

	// I N T E R F A C E

	/*
	 * (non-Javadoc)
	 * @see treebolic.site.SiteMaker#makeOutStream(java.lang.String)
	 */
	@Override
	public OutputStream makeOutStream(final String thisString) throws MalformedURLException, IOException
	{
		return this.theClient.put(thisString);
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.generator.sitemaker.SiteMaker#connect()
	 */
	@Override
	public boolean connect()
	{
		try
		{
			notifyOperation(Messages.getString("FtpSiteMaker.4"), SiteMaker.PROGRESSCONNECT); //$NON-NLS-1$
			this.theClient.openServer(this.theServer);

			try
			{
				notifyOperation(Messages.getString("FtpSiteMaker.1"), SiteMaker.PROGRESSCONNECT + 2); //$NON-NLS-1$
				this.theClient.login(this.theLogin, this.thePassword);
			}
			catch (final FtpLoginException thisException)
			{
				this.theClient.closeServer();
				throw new IOException("LOGGING IN"); //$NON-NLS-1$
			}

			notifyOperation(Messages.getString("FtpSiteMaker.2"), SiteMaker.PROGRESSCONNECT + 2); //$NON-NLS-1$
			this.theClient.passive();

			notifyOperation(Messages.getString("FtpSiteMaker.3"), SiteMaker.PROGRESSCONNECT + 3); //$NON-NLS-1$
			this.theClient.binary();

			try
			{
				notifyOperation("CD " + this.theDirectory, SiteMaker.PROGRESSCONNECT + 4); //$NON-NLS-1$
				this.theClient.cd(this.theDirectory);
			}
			catch (final IOException thisException)
			{
				this.theClient.mkdir(this.theDirectory);
				this.theClient.cd(this.theDirectory);
			}
		}
		catch (final IOException thisException)
		{
			notifyOperation(Messages.getString("FtpSiteMaker.10") + thisException.getMessage(), SiteMaker.PROGRESSCONNECTCOMPLETE); //$NON-NLS-1$
			return false;
		}

		notifyOperation(Messages.getString("FtpSiteMaker.5"), SiteMaker.PROGRESSCONNECTCOMPLETE); //$NON-NLS-1$
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.generator.sitemaker.SiteMaker#disconnect()
	 */
	@Override
	public boolean disconnect()
	{
		try
		{
			this.theClient.closeServer();
		}
		catch (final IOException thisException)
		{
			System.err.println(thisException.toString());
			thisException.printStackTrace();
			return false;
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see treebolic.generator.sitemaker.SiteMaker#changeFolder(java.lang.String)
	 */
	@Override
	public boolean changeFolder(final String thisRelativeDirectory)
	{
		try
		{
			this.theClient.cd(thisRelativeDirectory);
		}
		catch (final IOException thisException)
		{
			try
			{
				notifyOperation("MKDIR " + thisRelativeDirectory, SiteMaker.PROGRESSXFERIMAGES + 2); //$NON-NLS-1$
				this.theClient.mkdir(thisRelativeDirectory);
				notifyOperation("CD " + thisRelativeDirectory, SiteMaker.PROGRESSXFERIMAGES + 2); //$NON-NLS-1$
				this.theClient.cd(thisRelativeDirectory);
			}
			catch (final IOException thisSubException)
			{
				return false;
			}
		}
		return true;
	}
}
