/**
 * Title :	Treebolic generator
 * Description: Treebolic generator
 * Version:	provider
 * Copyright (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 */
package treebolic.site;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Date;
import java.util.Observable;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;

import treebolic.Messages;
import treebolic.dtd.Dtd;
import treebolic.generator.Pair;
import bsys.xml.transformer.DomTransformer;

/**
 * Make site
 * 
 * @author Bernard Bou
 */
public abstract class SiteMaker extends Observable
{
	// reference

	/**
	 * Jar file to transfer
	 */
	static private final String theJarFile = "TreebolicAppletDom.jar"; //$NON-NLS-1$

	/**
	 * Dtd file to transfer
	 */
	static private final String theDTDFile = "Treebolic.dtd"; //$NON-NLS-1$

	/**
	 * Applet class
	 */
	static private final String theAppclass = "treebolic.applet.Treebolic.class"; //$NON-NLS-1$

	/**
	 * HTML template file
	 */
	static public String theTemplateFile = "template.html"; //$NON-NLS-1$

	// progress steps

	/**
	 * Start
	 */
	static final int PROGRESSZERO = 0;

	/**
	 * Connect start
	 */
	static final int PROGRESSCONNECT = 0;

	/**
	 * Connect completed
	 */
	static final int PROGRESSCONNECTCOMPLETE = 10;

	/**
	 * Transfer XML
	 */
	static final int PROGRESSXFERXML = 10;

	/**
	 * Transfer HTML
	 */
	static final int PROGRESSXFERHTML = 40;

	/**
	 * Transfer applet Jar
	 */
	static final int PROGRESSXFERJAR = 50;

	/**
	 * Transfer DTD
	 */
	static final int PROGRESSXFERDTD = 75;

	/**
	 * Transfer images
	 */
	static final int PROGRESSXFERIMAGES = 80;

	/**
	 * Disconnect start
	 */
	static final int PROGRESSDISCONNECT = 95;

	/**
	 * Disconnect complete
	 */
	static final int PROGRESSCOMPLETE = 100;

	// D A T A

	/**
	 * Document (source)
	 */
	private final Document theDocument;

	/**
	 * Repository (source) for templetes
	 */
	private final String theRepository;

	/**
	 * Image repository (source)
	 */
	private final String theImageRepository;

	/**
	 * Title for HTML title (target)
	 */
	private final String theTitle;

	/**
	 * HTML file (target)
	 */
	protected String theHtmlFile;

	/**
	 * XML file (target)
	 */
	protected String theXmlFile;

	/**
	 * Operation string
	 */
	protected String theOperation;

	// C O N S T R U C T I O N

	/**
	 * Constructor
	 * 
	 * @param thisDocument
	 *        document (source)
	 * @param thisRepository
	 *        repository (source)
	 * @param thisImageRepository
	 *        image repository (source)
	 * @param thisHtmlFile
	 *        HTML file (target)
	 * @param thisXmlFile
	 *        XML file (target)
	 * @param thisTitle
	 *        HTML file title (target)
	 * @throws URISyntaxException
	 * @throws IOException
	 */
	public SiteMaker(final Document thisDocument, final String thisRepository, final String thisImageRepository, final String thisHtmlFile, final String thisXmlFile, final String thisTitle) throws URISyntaxException, IOException
	{
		this.theDocument = thisDocument;
		this.theTitle = thisTitle;
		this.theXmlFile = thisXmlFile;
		this.theHtmlFile = thisHtmlFile;
		this.theRepository = thisRepository;
		this.theImageRepository = thisImageRepository;
	}

	// I N T E R F A C E

	/**
	 * Connect to destination
	 * 
	 * @return true if successful
	 */
	abstract boolean connect();

	/**
	 * Disconnect from destination
	 * 
	 * @return true if successful
	 */
	abstract boolean disconnect();

	/**
	 * Change folder
	 * 
	 * @param thisRelativeFolder
	 *        relative folder
	 * @return true if successful
	 * @throws MalformedURLException
	 * @throws URISyntaxException
	 */
	abstract boolean changeFolder(String thisRelativeFolder) throws MalformedURLException, URISyntaxException;

	/**
	 * Make output stream
	 * 
	 * @param thisFilename
	 *        filename
	 * @return output stream
	 * @throws MalformedURLException
	 * @throws IOException
	 * @throws URISyntaxException
	 */
	abstract OutputStream makeOutStream(String thisFilename) throws MalformedURLException, IOException, URISyntaxException;

	// O B S E R V A B L E

	/**
	 * Notify operation progress
	 * 
	 * @param thisOperation
	 *        pending operation string
	 * @param thisProgressValue
	 *        progress value (max 100)
	 */
	void notifyOperation(final String thisOperation, final int thisProgressValue)
	{
		setChanged();
		this.theOperation = thisOperation;
		notifyObservers(new Pair<String, Integer>(thisOperation, thisProgressValue));
	}

	// O P E R A T I O N

	/**
	 * Make site operation
	 * 
	 * @return true if successful
	 */
	public boolean make()
	{
		// connect
		notifyOperation(Messages.getString("SiteMaker.10"), SiteMaker.PROGRESSCONNECT); //$NON-NLS-1$
		if (!connect())
		{
			notifyOperation(Messages.getString("SiteMaker.11"), SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$
			return false;
		}

		boolean success = false;
		try
		{
			// urls
			final URL thisRepositoryUrl = new URL(this.theRepository);
			final URL thisImageRepositoryUrl = new URL(this.theImageRepository);

			// xml file
			notifyOperation("STOR " + this.theXmlFile, SiteMaker.PROGRESSXFERXML); //$NON-NLS-1$
			OutputStream thisOutputStream = makeOutStream(this.theXmlFile);
			new DomTransformer().documentToStream(this.theDocument, thisOutputStream);
			thisOutputStream.close();

			// html file
			notifyOperation("STOR " + this.theHtmlFile, SiteMaker.PROGRESSXFERHTML); //$NON-NLS-1$
			final String[] theseMacros = { "%XMLFILE%", "%TITLE%", "%CLASS%", "%JAR%", "%WIDTH%", "%HEIGHT%", "%DATE%" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
			final String[] theseValues = { this.theXmlFile, this.theTitle, SiteMaker.theAppclass, SiteMaker.theJarFile, "100%", "100%", new Date().toString() }; //$NON-NLS-1$ //$NON-NLS-2$
			URL thisUrl = new URL(thisRepositoryUrl, SiteMaker.theTemplateFile);
			thisOutputStream = makeOutStream(this.theHtmlFile);
			copyStreamsReplace(thisUrl.openStream(), thisOutputStream, theseMacros, theseValues);
			thisOutputStream.close();

			// copy code jar
			notifyOperation("STOR " + SiteMaker.theJarFile, SiteMaker.PROGRESSXFERJAR); //$NON-NLS-1$
			thisUrl = new URL(thisRepositoryUrl, SiteMaker.theJarFile);
			thisOutputStream = makeOutStream(SiteMaker.theJarFile);
			SiteMaker.copyStreams(thisUrl.openStream(), thisOutputStream);
			thisOutputStream.close();

			// copy dtd
			notifyOperation("STOR " + SiteMaker.theDTDFile, SiteMaker.PROGRESSXFERDTD); //$NON-NLS-1$
			thisOutputStream = makeOutStream(SiteMaker.theDTDFile);
			Dtd.copyToUTF8Stream(thisOutputStream);
			thisOutputStream.close();

			// copy images
			notifyOperation("CD " + "images", SiteMaker.PROGRESSXFERIMAGES); //$NON-NLS-1$ //$NON-NLS-2$
			if (changeFolder("images")) //$NON-NLS-1$
			{
				for (final String thisImageFileName : DocumentSearch.makeImageList(this.theDocument))
				{
					notifyOperation("STOR " + "images" + "/" + thisImageFileName, SiteMaker.PROGRESSXFERIMAGES); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					thisUrl = new URL(thisImageRepositoryUrl, thisImageFileName);
					thisOutputStream = makeOutStream(thisImageFileName);
					SiteMaker.copyStreams(thisUrl.openStream(), thisOutputStream);
					thisOutputStream.close();
				}
			}

			// end
			notifyOperation(Messages.getString("SiteMaker.12"), SiteMaker.PROGRESSDISCONNECT); //$NON-NLS-1$
			success = true;
		}
		catch (final MalformedURLException thisException)
		{
			notifyOperation(Messages.getString("SiteMaker.20") + this.theOperation + " [" + thisException.toString() + "]", SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		catch (final IOException thisException)
		{
			notifyOperation(Messages.getString("SiteMaker.21") + this.theOperation + " [" + thisException.toString() + "]", SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		catch (final TransformerConfigurationException thisException)
		{
			notifyOperation(Messages.getString("SiteMaker.22") + this.theOperation + " [" + thisException.toString() + "]", SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		catch (final TransformerException thisException)
		{
			notifyOperation(Messages.getString("SiteMaker.23") + this.theOperation + " [" + thisException.toString() + "]", SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		catch (final URISyntaxException thisException)
		{
			notifyOperation(Messages.getString("SiteMaker.24") + this.theOperation + " [" + thisException.toString() + "]", SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		finally
		{
			disconnect();
		}
		if (success)
		{
			notifyOperation(Messages.getString("SiteMaker.13"), SiteMaker.PROGRESSCOMPLETE); //$NON-NLS-1$
		}
		return success;
	}

	/**
	 * Copy text file to output stream while replacing strings with values
	 * 
	 * @param thisInstream
	 *        input stream
	 * @param thisOutstream
	 *        output strean
	 * @param theseSources
	 *        what to replace
	 * @param theseDestinations
	 *        what to replace it with
	 * @return true if successful
	 */
	private boolean copyStreamsReplace(final InputStream thisInstream, final OutputStream thisOutstream, final String[] theseSources, final String[] theseDestinations)
	{
		try
		{
			final int thisBuffercount = 512;
			final BufferedReader theReader = new BufferedReader(new InputStreamReader(thisInstream), thisBuffercount);
			final PrintWriter theWriter = new PrintWriter(new BufferedOutputStream(thisOutstream));
			String thisLine;
			while ((thisLine = theReader.readLine()) != null)
			{
				for (int i = 0; i < theseSources.length; i++)
				{
					final int thisPos = thisLine.indexOf(theseSources[i]);
					if (thisPos != -1)
					{
						thisLine = thisLine.substring(0, thisPos) + theseDestinations[i] + thisLine.substring(thisPos + theseSources[i].length());
					}
				}
				theWriter.println(thisLine);
			}
			theReader.close();
			theWriter.close();
		}
		catch (final Exception unused)
		{
			return false;
		}
		return true;
	}

	/**
	 * Copy streams
	 * 
	 * @param thisInstream
	 *        input stream
	 * @param thisOutstream
	 *        output stream
	 * @return true if successful
	 */
	static private boolean copyStreams(final InputStream thisInstream, final OutputStream thisOutstream)
	{
		try
		{
			final int thisBufferCount = 512;
			final BufferedInputStream theReader = new BufferedInputStream(thisInstream, thisBufferCount);
			final BufferedOutputStream theWriter = new BufferedOutputStream(thisOutstream);
			final byte thisBuffer[] = new byte[thisBufferCount];
			int thisCount;
			while ((thisCount = theReader.read(thisBuffer, 0, thisBufferCount)) != -1)
			{
				theWriter.write(thisBuffer, 0, thisCount);
			}
			theReader.close();
			theWriter.close();
		}
		catch (final Exception unused)
		{
			return false;
		}
		return true;
	}
}
