/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.view;

import java.awt.Dimension;
import java.awt.Point;

import treebolic.core.Complex;
import treebolic.model.INode;
import treebolic.model.Location;

/**
 * Mapper of unit circle space to view
 * 
 * @author Bernard Bou
 */
public class Mapper
{
	/**
	 * Size of graphics context
	 */
	protected Dimension theSize;

	/**
	 * Scale x factor
	 */
	protected double theScaleX = 1.;

	/**
	 * Scale y factor
	 */
	protected double theScaleY = 1.;

	/**
	 * X shift
	 */
	protected double theXShift = 0.;

	/**
	 * Y shift
	 */
	protected double theYShift = 0.;

	/**
	 * Top coordinate
	 */
	protected int theTop;

	/**
	 * Left coordinate
	 */
	protected int theLeft;

	/**
	 * Zoom factor
	 */
	private double theZoomFactor = 1.;

	// S I Z E

	/**
	 * Get size
	 * 
	 * @return size
	 */
	public Dimension getSize()
	{
		return this.theSize;
	}

	// A C C E S S

	/**
	 * Set view zoom factor
	 * 
	 * @param thisFactor
	 *        zoom factor
	 */
	public void setZoomFactor(final Double thisFactor)
	{
		this.theZoomFactor *= thisFactor;
	}

	/**
	 * Set view x-shift
	 * 
	 * @param thisX
	 *        view x-shift (0,1)
	 */
	public void setXShift(final Double thisX)
	{
		this.theXShift = thisX;
	}

	/**
	 * Set view y-shift
	 * 
	 * @param thisY
	 *        view y-shift (0,1)
	 */
	public void setYShift(final Double thisY)
	{
		this.theYShift = thisY;
	}

	/**
	 * Get view x-shift
	 * 
	 * @return view x-shift
	 */
	public double getXShift()
	{
		return this.theXShift;
	}

	/**
	 * Get view y-shift
	 * 
	 * @return view y-shift
	 */
	public double getYShift()
	{
		return this.theYShift;
	}

	/**
	 * Compute scale from view size
	 */
	protected void computeScale()
	{
		if (this.theSize == null)
			return;
		this.theScaleX = (0.5 + (this.theXShift > 0. ? this.theXShift : -this.theXShift)) * this.theSize.width;
		this.theScaleY = (0.5 + (this.theYShift > 0. ? this.theYShift : -this.theYShift)) * this.theSize.height;
	}

	// M A P . U N I T C I R C L E . T O . V I E W

	/**
	 * Convert unit circle x-coordinate to view x-coordinate
	 * 
	 * @param x
	 *        unit circle x-coordinate
	 * @return view x-coordinate
	 */
	protected int xUnitCircleToView(final double x)
	{
		if (this.theSize == null)
			return 0;
		return (int) (this.theScaleX * this.theZoomFactor * x + this.theXShift * this.theSize.width);
	}

	/**
	 * Convert unit circle y-coordinate to view y-coordinate
	 * 
	 * @param y
	 *        unit circle y-coordinate
	 * @return view y-coordinate
	 */
	protected int yUnitCircleToView(final double y)
	{
		if (this.theSize == null)
			return 0;
		return (int) (this.theScaleY * this.theZoomFactor * y + this.theYShift * this.theSize.width);
	}

	/**
	 * Convert unit circle x-extent to view x-extent
	 * 
	 * @param cx
	 *        unit circle x-extent
	 * @return view x-extent
	 */
	protected int wUnitCircleToView(final double cx)
	{
		return (int) (this.theScaleX * this.theZoomFactor * cx);
	}

	/**
	 * Convert unit circle y-extent to view y-extent
	 * 
	 * @param cy
	 *        unit circle y-extent
	 * @return view y-extent
	 */
	protected int hUnitCircleToView(final double cy)
	{
		return (int) (this.theScaleY * this.theZoomFactor * cy);
	}

	// M A P . V I E W . T O . U N I T C I R C L E

	/**
	 * Convert view x-coordinate to unit circle x-coordinate
	 * 
	 * @param vx
	 *        view x-coordinate
	 * @return unit circle x-coordinate
	 */
	protected double xViewToUnitCircle(final double vx)
	{
		if (this.theSize == null)
			return 0.;
		return (vx - this.theXShift * this.theSize.width) / (this.theScaleX * this.theZoomFactor);
	}

	/**
	 * Convert view y-coordinate to unit circle y-coordinate
	 * 
	 * @param vy
	 *        view y-coordinate
	 * @return unit circle y-coordinate
	 */
	protected double yViewToUnitCircle(final double vy)
	{
		if (this.theSize == null)
			return 0.;
		return (vy - this.theYShift * this.theSize.height) / (this.theScaleY * this.theZoomFactor);
	}

	/**
	 * Convert view x-extent (width) to unit circle x-extent
	 * 
	 * @param cvx
	 *        view x-extent
	 * @return unit circle x-extent
	 */
	protected double wViewToUnitCircle(final double cvx)
	{
		return cvx / (this.theScaleX * this.theZoomFactor);
	}

	/**
	 * Convert view y-extent (height) to unit circle y-extent
	 * 
	 * @param cvy
	 *        view y-extent
	 * @return unit circle y-extent
	 */
	protected double hViewToUnitCircle(final double cvy)
	{
		return cvy / (this.theScaleY * this.theZoomFactor);
	}

	// S P A C E . C O N V E R S I O N

	/**
	 * Convert view coordinates to unit circle coordinates
	 * 
	 * @param vx
	 *        x coordinate in view space
	 * @param vy
	 *        y coordinate in view space
	 * @param thisSize
	 *        view size
	 * @return point in unit circle
	 */
	public Complex viewToUnitCircle(final int vx, final int vy, final Dimension thisSize)
	{
		final Complex p = new Complex(vx, vy);

		// this offsets for translation on cache graphics
		p.set(p.re - thisSize.width / 2, p.im - thisSize.height / 2);

		// convert
		p.set(xViewToUnitCircle(p.re), yViewToUnitCircle(p.im));

		// adjust
		if (p.abs2() > 1.)
		{
			p.normalize();
			p.multiply(.99);
		}
		return p;
	}

	/**
	 * Get view coordinates of node
	 * 
	 * @param thisNode
	 *        node
	 * @return view coordinate of node
	 */
	public Point getViewLocation(final INode thisNode)
	{
		final Location thisLocation = thisNode.getLocation();
		return new Point(xUnitCircleToView(thisLocation.euclidian.center.re) - this.theLeft, yUnitCircleToView(thisLocation.euclidian.center.im) - this.theTop);
	}
}
