/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.util.List;

import treebolic.core.Arc;
import treebolic.core.Complex;
import treebolic.model.IEdge;
import treebolic.model.INode;
import treebolic.model.Location;
import treebolic.model.MountPoint;

/**
 * Painter class
 * 
 * @author Bernard Bou
 */
public class Painter extends AbstractPainter
{
	// C O N S T A N T S

	/**
	 * Dot pattern
	 */
	static private final float[] theDotPattern = { 1, 3 };

	/**
	 * Dot stroke
	 */
	static private final BasicStroke theDotStroke = new BasicStroke(1.0F, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 10.0F, Painter.theDotPattern, 0);

	/**
	 * Dash pattern
	 */
	static private final float[] theDashPattern = { 10, 4 };

	/**
	 * Dash stroke
	 */
	static private final BasicStroke theDashStroke = new BasicStroke(1.0F, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 10.0F, Painter.theDashPattern, 0);

	/**
	 * Minimum span for terminators
	 */
	static private final int theMinSpanForTerminators = 50 * 50;

	/**
	 * Text padding
	 */
	static private final int theTextPadding = 25;

	/**
	 * Edge terminator length
	 */
	static private final double theTerminorLength = 8;

	/**
	 * Edge terminator width
	 */
	static private final double theTerminatorWidth = 5;

	/**
	 * Debug flag, bit1=draw circles, bit2=suppress
	 */
	static private int theDebugFlag = 0;

	// B A C K G R O U N D

	/*
	 * (non-Javadoc)
	 * @see treebolic.view.AbstractPainter#paintBackground()
	 */
	@Override
	public void paintBackground()
	{
		this.theGraphics.setColor(this.theBackColor);
		this.theGraphics.fillRect(this.theLeft, this.theTop, this.theSize.width, this.theSize.height);
		this.theGraphics.setPaintMode();

		if (this.theBackgroundImage != null)
		{
			final int xinc = this.theBackgroundImage.getWidth(null);
			final int yinc = this.theBackgroundImage.getHeight(null);
			if (xinc <= 0 || yinc <= 0)
				return;

			final int xmax = this.theSize.width;
			final int ymax = this.theSize.height;
			for (int y = this.theTop; y < ymax; y += yinc)
			{
				for (int x = this.theLeft; x < xmax; x += xinc)
				{
					this.theGraphics.drawImage(this.theBackgroundImage, x, y, null);
				}
			}
		}

		// boundary circle
		if (Painter.theDebugFlag != 0)
		{
			drawCircle(0., 0., 1., Color.lightGray);
		}
	}

	// D R A W

	/*
	 * (non-Javadoc)
	 * @see treebolic.view.AbstractPainter#paint(treebolic.model.INode, treebolic.model.IEdgeList)
	 */
	@Override
	public void paint(final INode thisRoot, final List<IEdge> thisEdgeList)
	{
		// anti aliasing
		final Graphics2D thisGraphics2 = (Graphics2D) this.theGraphics;
		thisGraphics2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

		// compute node data
		computeTree(thisRoot);

		// edges
		if (thisEdgeList != null)
		{
			for (final IEdge thisEdge : thisEdgeList)
			{
				draw(thisEdge);
			}
		}

		// tree
		drawTree(thisRoot);
	}

	/**
	 * Compute tree recursively
	 * 
	 * @param thisNode
	 *        starting node
	 */
	private void computeTree(final INode thisNode)
	{
		if (thisNode == null)
			return;

		// hyper circle
		final Location thisLocation = thisNode.getLocation();
		if (thisLocation.hyper.isDirty)
		{
			thisLocation.mapToUnitCircle();
		}

		// node data and attach to node
		final NodeData thisNodeData = computeNodeData(thisNode);
		thisNode.getLocation().theViewData = thisNodeData;

		// recurse to compute the children
		final List<INode> theseChildren = thisNode.getChildren();
		if (theseChildren != null)
		{
			for (final INode thisChild : theseChildren)
			{
				computeTree(thisChild);
			}
		}
	}

	/**
	 * Draw tree recursively
	 * 
	 * @param thisNode
	 *        starting node
	 */
	private void drawTree(final INode thisNode)
	{
		if (thisNode == null)
			return;

		// edge to parent
		final INode thisParent = thisNode.getParent();
		if (thisParent != null)
		{
			// color
			Color thisColor = thisNode.getEdgeColor();
			if (thisColor == null)
			{
				thisColor = this.theTreeEdgeColor;
			}
			this.theGraphics.setColor(thisColor);

			// draw
			drawTreeEdge(thisParent, thisNode);
		}

		// recurse to draw the children
		final List<INode> theseChildren = thisNode.getChildren();
		if (theseChildren != null)
		{
			for (final INode thisChild : theseChildren)
			{
				drawTree(thisChild);
			}
		}

		// debug
		if ((Painter.theDebugFlag & 1) != 0)
		{
			drawSpace(thisNode);
		}

		if ((Painter.theDebugFlag & 2) != 0)
			return;

		// draw node
		final Location thisLocation = thisNode.getLocation();
		if (!thisLocation.hyper.isBorder)
		{
			drawNode((NodeData) thisNode.getLocation().theViewData);
		}
	}

	// D R A W . N O D E

	/**
	 * Class collecting drawing data
	 * 
	 * @author Bernard Bou
	 */
	private class NodeData
	{
		/**
		 * Node box
		 */
		public Rectangle2D theBox;

		/**
		 * Node label
		 */
		public String theLabel;

		/**
		 * Node label x coordinate
		 */
		public int theLabelX;

		/**
		 * Node label y coordinate
		 */
		public int theLabelY;

		/**
		 * Node image
		 */
		public Image theImage;

		/**
		 * Node image x-coordinate
		 */
		public int theImageX;

		/**
		 * Node image y-coordinate
		 */
		public int theImageY;

		/**
		 * Node backcolor
		 */
		public Color theBackColor;

		/**
		 * Node forecolor
		 */
		public Color theForeColor;

		/**
		 * Node font
		 */
		public Font theFont;

		/**
		 * Is mountable
		 */
		public Boolean isMountable;
	}

	/**
	 * Draw node
	 * 
	 * @param thisNodeData
	 *        computed node data
	 */

	private void drawNode(final NodeData thisNodeData)
	{
		if (thisNodeData == null)
			return;

		// box
		if (thisNodeData.theBox != null)
		{
			// graphics context
			final Graphics2D thisGraphics2 = (Graphics2D) this.theGraphics;

			// rectangle
			final RoundRectangle2D thisRoundRectangle = new RoundRectangle2D.Double();
			thisRoundRectangle.setRoundRect(thisNodeData.theBox.getX(), thisNodeData.theBox.getY(), thisNodeData.theBox.getWidth(), thisNodeData.theBox.getHeight(), 10, 10);

			// fill
			thisGraphics2.setColor(thisNodeData.theBackColor);
			thisGraphics2.fill(thisRoundRectangle);

			// outline
			thisGraphics2.setColor(thisNodeData.theForeColor);
			thisGraphics2.draw(thisRoundRectangle);

			// mount clue
			if (thisNodeData.isMountable != null)
			{
				final int w = 3;
				final int h = 2 * w;
				final int hm = 2;
				final int x = (int) thisRoundRectangle.getX() + (int) thisRoundRectangle.getWidth() + 5;
				final int x1 = x - w;
				final int x2 = x + w;
				final int y = (int) thisRoundRectangle.getY() + hm;
				final int xs[] = { x1, x, x2 };
				if (thisNodeData.isMountable)
				{
					final int hr = (int) thisRoundRectangle.getHeight() - 2 * hm;
					final int y2 = y + hr;
					final int y1 = y2 - h;
					final int ys[] = { y1, y2, y1 };
					thisGraphics2.fillPolygon(xs, ys, 3);
				}
				else
				{
					final int y1 = y + h;
					final int ys[] = { y1, y, y1 };
					thisGraphics2.fillPolygon(xs, ys, 3);
				}
			}
		}

		// image
		if (thisNodeData.theImage != null)
		{
			this.theGraphics.drawImage(thisNodeData.theImage, thisNodeData.theImageX, thisNodeData.theImageY, null);
		}

		// label
		if (thisNodeData.theLabel != null)
		{
			this.theGraphics.setFont(thisNodeData.theFont);
			this.theGraphics.drawString(thisNodeData.theLabel, thisNodeData.theLabelX, thisNodeData.theLabelY);
		}
	}

	/**
	 * Compute node data
	 * 
	 * @param thisNode
	 *        node
	 * @return node data
	 */
	@SuppressWarnings("synthetic-access")
	private NodeData computeNodeData(final INode thisNode)
	{
		final NodeData thisNodeData = new NodeData();

		// hyper circle
		final Location thisLocation = thisNode.getLocation();
		if (thisLocation.hyper.isBorder)
			return null;

		// font size
		int thisFontIdx = (int) Math.round(thisLocation.hyper.dist * AbstractPainter.NFONTS);
		thisFontIdx = Math.min(this.theFonts.length - 1, thisFontIdx);
		thisNodeData.theFont = this.theFonts[thisFontIdx];
		this.theGraphics.setFont(this.theFonts[thisFontIdx]);

		// color
		thisNodeData.theBackColor = thisNode.getBackColor();
		if (thisNodeData.theBackColor == null)
		{
			thisNodeData.theBackColor = this.theNodeBackColor;
		}
		thisNodeData.theForeColor = thisNode.getForeColor();
		if (thisNodeData.theForeColor == null)
		{
			thisNodeData.theForeColor = this.theNodeForeColor;
		}

		// center
		final int x = xUnitCircleToView(thisLocation.euclidian.center.re);
		final int y = yUnitCircleToView(thisLocation.euclidian.center.im);

		// node space actual diameter
		final int rad = wUnitCircleToView(thisLocation.euclidian.radius);
		final int dia = 2 * rad;
		final int span = dia;

		// image
		thisNodeData.theImage = thisNode.getImage();
		if (thisNodeData.theImage == null)
		{
			thisNodeData.theImage = this.theDefaultNodeImage;
		}

		// string
		thisNodeData.theLabel = thisNode.toString();
		if (thisNodeData.theLabel == null || thisNodeData.theLabel.isEmpty())
		{
			if (thisNodeData.theImage != null)
			{
				final int hi = thisNodeData.theImage.getHeight(null);
				final int xi = x - thisNodeData.theImage.getWidth(null) / 2;
				final int yi = y - hi / 2;

				thisNodeData.theImageX = xi;
				thisNodeData.theImageY = yi;
			}
			return thisNodeData;
		}

		// label font size
		final FontMetrics fm = this.theGraphics.getFontMetrics();
		final int h = fm.getAscent() /* +fm.getDescent() */;
		int w = fm.stringWidth(thisNodeData.theLabel);

		// truncate label
		if (w > span)
		{
			// compute average character width
			final int wunit = fm.stringWidth("x");

			// compute trailing dots width
			final int wdots = fm.stringWidth("...");

			// compute number of characters that fit before dots
			int thisNChars = (span - wdots) / wunit;

			// ensure at least one
			if (thisNChars < 1)
			{
				thisNChars = 1;
			}

			// perform truncation if we actually truncate
			final int thisLen = thisNodeData.theLabel.length();
			if (thisLen > thisNChars)
			{
				thisNodeData.theLabel = thisNodeData.theLabel.substring(0, thisNChars) + "...";
				w = fm.stringWidth(thisNodeData.theLabel);
			}
		}

		// box computation
		final int xr = x - w / 2 - AbstractPainter.theXInset;
		int yr = y - h / 2 - AbstractPainter.theYInset;
		final int wr = w + 2 * AbstractPainter.theXInset;
		final int hr = h + 2 * AbstractPainter.theYInset;

		// image computation
		int hi2 = 0; // height image
		int rs = 0; // rectangle shift
		if (thisNodeData.theImage != null)
		{
			hi2 = thisNodeData.theImage.getHeight(null) / 2;
			rs = hi2 > rad ? rad : hi2; // limit shift to radius
			yr += rs;
		}

		// box
		thisNodeData.theBox = new Rectangle2D.Double(xr, yr, wr, hr);

		// image
		if (thisNodeData.theImage != null)
		{
			final int xi = x - thisNodeData.theImage.getWidth(null) / 2;
			final int yi = y - hi2;
			thisNodeData.theImageX = xi;
			thisNodeData.theImageY = yi;
		}

		// label
		thisNodeData.theLabelX = xr + AbstractPainter.theXInset;
		thisNodeData.theLabelY = yr + AbstractPainter.theYInset + h;

		// is mountable
		final MountPoint thisMountPoint = thisNode.getMountPoint();
		thisNodeData.isMountable = thisMountPoint == null ? null : thisMountPoint instanceof MountPoint.Mounting ? true : false;

		return thisNodeData;
	}

	// D R A W . E D G E

	/**
	 * Draw tree edge, from parent to child
	 * 
	 * @param thisParent
	 *        from-node
	 * @param thisNode
	 *        to-node
	 */
	private void drawTreeEdge(final INode thisParent, final INode thisNode)
	{
		// style
		final Integer thatStyle = thisNode.getEdgeStyle();
		final int thisStyle = thatStyle != null ? thatStyle : this.theTreeEdgeStyle;
		if ((thisStyle & IEdge.HIDDEN) != 0)
			return;

		// hyper circles
		final Location thisFrom = thisParent.getLocation();
		final Location thisTo = thisNode.getLocation();

		// data
		final NodeData thisFromData = (NodeData) thisParent.getLocation().theViewData;
		final NodeData thisToData = (NodeData) thisNode.getLocation().theViewData;
		final Rectangle2D thisFromBox = thisFromData == null ? null : thisFromData.theBox;
		final Rectangle2D thisToBox = thisToData == null ? null : thisToData.theBox;

		// do not draw edge if boxes intersect
		if (boxesIntersect(thisFromBox, thisToBox))
			return;

		// image
		Image thisImage = thisNode.getEdgeImage();
		if (thisImage == null)
		{
			thisImage = this.theDefaultTreeEdgeImage;
		}

		// draw
		final String thisLabel = thisNode.getEdgeLabel();
		final boolean isBorder = thisFrom.hyper.isBorder;
		drawArc(thisFrom.euclidian.center, thisTo.euclidian.center, thisLabel, thisImage, thisStyle, thisFromBox, thisToBox, isBorder);
	}

	/**
	 * Draw non-tree edge
	 * 
	 * @param thisEdge
	 *        edge
	 */
	private void draw(final IEdge thisEdge)
	{
		// style
		final Integer thatStyle = thisEdge.getStyle();
		final int thisStyle = thatStyle != null ? thatStyle : this.theEdgeStyle;
		if ((thisStyle & IEdge.HIDDEN) != 0)
			return;

		// hyper circles
		final INode thisFromNode = thisEdge.getFrom();
		final INode thisToNode = thisEdge.getTo();
		if (thisFromNode == null || thisToNode == null)
			return;

		final Location thisFromLocation = thisFromNode.getLocation();
		final Location thisToLocation = thisToNode.getLocation();
		if (thisFromLocation.hyper.isBorder || thisToLocation.hyper.isBorder)
			return;

		if (thisFromLocation.hyper.isDirty)
		{
			thisFromLocation.mapToUnitCircle();
		}
		if (thisToLocation.hyper.isDirty)
		{
			thisToLocation.mapToUnitCircle();
		}

		// data
		final NodeData thisFromData = (NodeData) thisFromNode.getLocation().theViewData;
		final NodeData thisToData = (NodeData) thisToNode.getLocation().theViewData;
		if (thisFromData == null || thisToData == null)
			return;

		// do not draw edge if boxes intersect
		if (boxesIntersect(thisFromData.theBox, thisToData.theBox))
			return;

		// label
		final String thisLabel = thisEdge.getLabel();
		if (thisLabel != null)
		{
			// font size
			int thisFontIdx = (int) Math.round(thisToLocation.hyper.dist * AbstractPainter.NFONTS);
			thisFontIdx = Math.min(this.theFonts.length - 1, thisFontIdx);
			this.theGraphics.setFont(this.theFonts[thisFontIdx]);
		}

		// draw
		Color thisColor = thisEdge.getColor();
		if (thisColor == null)
		{
			thisColor = this.theEdgeColor;
		}
		this.theGraphics.setColor(thisColor);

		// image
		Image thisImage = thisEdge.getImage();
		if (thisImage == null)
		{
			thisImage = this.theDefaultEdgeImage;
		}

		// draw arc
		drawArc(thisFromLocation.euclidian.center, thisToLocation.euclidian.center, thisLabel, thisImage, thisStyle, thisFromData.theBox, thisToData.theBox, false);
	}

	// D R A W . A R C

	/**
	 * Draw geodesic arc from z1 to z2 which models line from z1 to z2
	 * 
	 * @param z1
	 *        from-end
	 * @param z2
	 *        to-end
	 * @param thatLabel
	 *        arc label
	 * @param thisImage
	 *        arc image
	 * @param thisStyle
	 *        code for edge style
	 * @param thisFromBox
	 *        from-node box
	 * @param thisToBox
	 *        to-node box
	 * @param isBorder
	 *        true if arc neighbours border
	 */
	private void drawArc(final Complex z1, final Complex z2, final String thatLabel, final Image thisImage, final int thisStyle, final Rectangle2D thisFromBox, final Rectangle2D thisToBox, final boolean isBorder)
	{
		String thisLabel = thatLabel;

		if (!this.hyperbolicEdges || this.isDragging)
		{
			final Point2D thisFrom = new Point2D.Double(xUnitCircleToView(z1.re), yUnitCircleToView(z1.im));
			final Point2D thisTo = new Point2D.Double(xUnitCircleToView(z2.re), yUnitCircleToView(z2.im));

			drawLine(thisFrom, thisTo, thisStyle);
			if (!isBorder && thisLabel != null && !thisLabel.isEmpty())
			{
				// fit
				thisLabel = mangleString(thisLabel, thisFrom, thisTo);
				if (thisLabel == null)
					return;

				// draw
				drawText(thisLabel, thisFrom, thisTo);
			}
		}
		else
		{
			final Arc thisArc = new Arc(z1, z2);
			draw(thisArc, thisLabel, thisImage, thisStyle, thisFromBox, thisToBox, isBorder);
		}
	}

	/**
	 * Draw arc
	 * 
	 * @param thisArc
	 *        arc
	 * @param thatLabel
	 *        arc label
	 * @param thisImage
	 *        arc image
	 * @param isBorder
	 *        true if arc neighbours border
	 */
	private void draw(final Arc thisArc, final String thatLabel, final Image thisImage, final int thisStyle, final Rectangle2D thisFromBox, final Rectangle2D thisToBox, final boolean isBorder)
	{
		String thisLabel = thatLabel;

		// if(r == 0.) it is segment of line
		if (thisArc.r == 0.)
		{
			Point2D thisFrom = new Point2D.Double(xUnitCircleToView(thisArc.from.re), yUnitCircleToView(thisArc.from.im));
			Point2D thisTo = new Point2D.Double(xUnitCircleToView(thisArc.to.re), yUnitCircleToView(thisArc.to.im));

			// adjust to anchors
			Point2D thisFromAnchor = null;
			if (!isBorder)
			{
				thisFromAnchor = getIntersection(thisFromBox, thisTo, thisFrom);
			}
			Point2D thisToAnchor = null;
			thisToAnchor = getIntersection(thisToBox, thisFrom, thisTo);

			// adjust line ends
			if (thisFromAnchor != null)
			{
				thisFrom = thisFromAnchor;
			}
			if (thisToAnchor != null)
			{
				thisTo = thisToAnchor;
			}

			// line
			drawLine(thisFrom, thisTo, thisStyle);

			// image
			if (thisImage != null && !isBorder)
			{
				drawImage(thisImage, getMiddle(thisFrom, thisTo));
			}

			// ends
			drawEdgeEnds(thisFrom, thisTo, null, thisStyle);

			// label
			if (!isBorder && thisLabel != null && !thisLabel.isEmpty())
			{
				// fit
				thisLabel = mangleString(thisLabel, thisFrom, thisTo);
				if (thisLabel == null)
					return;

				drawText(thisLabel, thisFrom, thisTo);
			}
		}
		else
		{
			final Arc2D thisArc2D = toArc2D(thisArc);

			// anchors
			Point2D thisFromAnchor = null;
			if (!isBorder)
			{
				thisFromAnchor = getIntersection(thisFromBox, thisArc2D);
			}
			if (thisFromAnchor == null)
			{
				thisFromAnchor = thisArc2D.getStartPoint();
			}

			Point2D thisToAnchor = null;
			thisToAnchor = getIntersection(thisToBox, thisArc2D);
			if (thisToAnchor == null)
			{
				thisToAnchor = thisArc2D.getEndPoint();
			}

			// adjust arc ends to anchors
			if (thisArc2D.getAngleExtent() >= 0.)
			{
				thisArc2D.setAngles(thisFromAnchor, thisToAnchor);
			}
			else
			{
				thisArc2D.setAngles(thisToAnchor, thisFromAnchor);
			}

			// draw
			drawArc(thisArc2D, thisFromAnchor, thisToAnchor, thisStyle);

			// image
			if (thisImage != null)
			{
				drawImage(thisImage, getMidArc(thisArc2D));
			}

			// ends
			drawEdgeEnds(thisFromAnchor, thisToAnchor, thisArc2D, thisStyle);

			// draw edge label
			if (!isBorder && thisLabel != null && !thisLabel.isEmpty())
			{
				// fit
				thisLabel = mangleString(thisLabel, thisFromAnchor, thisToAnchor);
				if (thisLabel == null)
					return;

				// mid arc
				final Point2D thisMidArc = getMidArc(thisArc2D);

				// tangent
				final double thisTangent = getTangent(thisArc2D, thisMidArc);

				// draw text
				drawText(thisLabel, thisMidArc, thisTangent);
			}
		}
	}

	/**
	 * Draw arc2D
	 * 
	 * @param thisArc2D
	 *        arc
	 */
	private void drawArc(final Arc2D thisArc2D, final Point2D thisFromAnchor, final Point2D thisToAnchor, final int thisStyle)
	{
		final Graphics2D thisGraphics = (Graphics2D) this.theGraphics;
		if ((thisStyle & IEdge.STROKEMASK) != 0)
		{
			final Stroke thatStroke = thisGraphics.getStroke();
			Stroke thisStroke;
			if ((thisStyle & IEdge.DASH) != 0)
			{
				thisStroke = Painter.theDashStroke;
			}
			else if ((thisStyle & IEdge.DOT) != 0)
			{
				thisStroke = Painter.theDotStroke;
			}
			else
			{
				thisStroke = null;
			}

			if (thisStroke != null)
			{
				thisGraphics.setStroke(thisStroke);
				thisGraphics.draw(thisArc2D);
				thisGraphics.setStroke(thatStroke);
				return;
			}
		}
		thisGraphics.draw(thisArc2D);
	}

	/**
	 * Draw edge ends
	 * 
	 * @param thisFrom
	 *        from-end
	 * @param thisTo
	 *        to-end
	 * @param thisArc2D
	 *        arc
	 * @param thisStyle
	 *        style code
	 */
	private void drawEdgeEnds(final Point2D thisFrom, final Point2D thisTo, final Arc2D thisArc2D, final int thisStyle)
	{
		if (thisStyle == 0)
			return;

		final double x1 = thisFrom.getX();
		final double y1 = thisFrom.getY();
		final double x2 = thisTo.getX();
		final double y2 = thisTo.getY();
		final double cx = x2 - x1;
		final double cy = y2 - y1;

		// do nothing if too near
		final double span2 = cx * cx + cy * cy;
		if (span2 < Painter.theMinSpanForTerminators)
			return;

		double sx1;
		double sy1;
		double dx1;
		double dy1;
		double sx2;
		double sy2;
		double dx2;
		double dy2;
		if (thisArc2D == null)
		{
			final double thisOrientation = Math.atan2(y2 - y1, x2 - x1);
			sx1 = Painter.theTerminorLength * Math.cos(thisOrientation);
			sy1 = Painter.theTerminorLength * Math.sin(thisOrientation);
			dx1 = Painter.theTerminatorWidth * Math.cos(thisOrientation + Math.PI / 2);
			dy1 = Painter.theTerminatorWidth * Math.sin(thisOrientation + Math.PI / 2);
			sx2 = sx1;
			sy2 = sy1;
			dx2 = dx1;
			dy2 = dy1;
		}
		else
		{
			final double thisOrientation1 = getTangent(thisArc2D, thisFrom);
			sx1 = Painter.theTerminorLength * Math.cos(thisOrientation1);
			sy1 = Painter.theTerminorLength * Math.sin(thisOrientation1);
			dx1 = Painter.theTerminatorWidth * Math.cos(thisOrientation1 + Math.PI / 2);
			dy1 = Painter.theTerminatorWidth * Math.sin(thisOrientation1 + Math.PI / 2);

			final double thisOrientation2 = getTangent(thisArc2D, thisTo);
			sx2 = Painter.theTerminorLength * Math.cos(thisOrientation2);
			sy2 = Painter.theTerminorLength * Math.sin(thisOrientation2);
			dx2 = Painter.theTerminatorWidth * Math.cos(thisOrientation2 + Math.PI / 2);
			dy2 = Painter.theTerminatorWidth * Math.sin(thisOrientation2 + Math.PI / 2);
		}

		// from
		switch (thisStyle >> IEdge.FROMSHIFT & IEdge.SHAPEMASK)
		{
		case IEdge.TRIANGLE:
		{
			// t
			final int x[] = { (int) x1, (int) (x1 + sx1 + dx1), (int) (x1 + sx1 - dx1) };
			final int y[] = { (int) y1, (int) (y1 + sy1 + dy1), (int) (y1 + sy1 - dy1) };
			if ((thisStyle & IEdge.FROMSOLID) != 0)
			{
				this.theGraphics.fillPolygon(x, y, x.length);
			}
			else
			{
				final Color thatColor = this.theGraphics.getColor();
				this.theGraphics.setColor(this.theBackColor);
				this.theGraphics.fillPolygon(x, y, x.length);
				this.theGraphics.setColor(thatColor);
				this.theGraphics.drawPolygon(x, y, x.length);
			}
			break;
		}

		case IEdge.CIRCLE:
		{
			// c
			final double x = x1 + sx1 / 2 - Painter.theTerminorLength / 2;
			final double y = y1 + sy1 / 2 - Painter.theTerminorLength / 2;
			if ((thisStyle & IEdge.FROMSOLID) != 0)
			{
				this.theGraphics.fillOval((int) x, (int) y, (int) Painter.theTerminorLength, (int) Painter.theTerminorLength);
			}
			else
			{
				final Color thatColor = this.theGraphics.getColor();
				this.theGraphics.setColor(this.theBackColor);
				this.theGraphics.fillOval((int) x, (int) y, (int) Painter.theTerminorLength, (int) Painter.theTerminorLength);
				this.theGraphics.setColor(thatColor);
				this.theGraphics.drawOval((int) x, (int) y, (int) Painter.theTerminorLength, (int) Painter.theTerminorLength);
			}
			break;
		}

		case IEdge.DIAMOND:
		{
			// d
			final int x[] = { (int) x1, (int) (x1 + sx1 + dx1), (int) (x1 + sx1 + sx1), (int) (x1 + sx1 - dx1) };
			final int y[] = { (int) y1, (int) (y1 + sy1 + dy1), (int) (y1 + sy1 + sy1), (int) (y1 + sy1 - dy1) };
			if ((thisStyle & IEdge.FROMSOLID) != 0)
			{
				this.theGraphics.fillPolygon(x, y, x.length);
			}
			else
			{
				final Color thatColor = this.theGraphics.getColor();
				this.theGraphics.setColor(this.theBackColor);
				this.theGraphics.fillPolygon(x, y, x.length);
				this.theGraphics.setColor(thatColor);
				this.theGraphics.drawPolygon(x, y, x.length);
			}
			break;
		}

		case IEdge.ARROW:
		{
			// a
			final int x[] = { (int) (x1 + sx1 + dx1), (int) x1, (int) (x1 + sx1 - dx1) };
			final int y[] = { (int) (y1 + sy1 + dy1), (int) y1, (int) (y1 + sy1 - dy1) };
			this.theGraphics.drawPolyline(x, y, x.length);
			break;
		}

		case IEdge.HOOK:
		{
			// h
			final int x[] = { (int) (x1 + sx1 + dx1), (int) x1 };
			final int y[] = { (int) (y1 + sy1 + dy1), (int) y1 };
			this.theGraphics.drawPolyline(x, y, x.length);
			break;
		}
		}

		// to
		switch (thisStyle >> IEdge.TOSHIFT & IEdge.SHAPEMASK)
		{
		case IEdge.TRIANGLE:
		{
			// t
			final int x[] = { (int) x2, (int) (x2 - sx2 + dx2), (int) (x2 - sx2 - dx2) };
			final int y[] = { (int) y2, (int) (y2 - sy2 + dy2), (int) (y2 - sy2 - dy2) };
			if ((thisStyle & IEdge.TOSOLID) != 0)
			{
				this.theGraphics.fillPolygon(x, y, x.length);
			}
			else
			{
				final Color thatColor = this.theGraphics.getColor();
				this.theGraphics.setColor(this.theBackColor);
				this.theGraphics.fillPolygon(x, y, x.length);
				this.theGraphics.setColor(thatColor);
				this.theGraphics.drawPolygon(x, y, x.length);
			}
			break;
		}

		case IEdge.CIRCLE:
		{
			// c
			final double x = x2 - sx2 / 2 - Painter.theTerminorLength / 2;
			final double y = y2 - sy2 / 2 - Painter.theTerminorLength / 2;
			if ((thisStyle & IEdge.TOSOLID) != 0)
			{
				this.theGraphics.fillOval((int) x, (int) y, (int) Painter.theTerminorLength, (int) Painter.theTerminorLength);
			}
			else
			{
				final Color thatColor = this.theGraphics.getColor();
				this.theGraphics.setColor(this.theBackColor);
				this.theGraphics.fillOval((int) x, (int) y, (int) Painter.theTerminorLength, (int) Painter.theTerminorLength);
				this.theGraphics.setColor(thatColor);
				this.theGraphics.drawOval((int) x, (int) y, (int) Painter.theTerminorLength, (int) Painter.theTerminorLength);
			}
			break;
		}

		case IEdge.DIAMOND:
		{
			// d
			final int x[] = { (int) x2, (int) (x2 - sx2 + dx2), (int) (x2 - sx2 - sx2), (int) (x2 - sx2 - dx2) };
			final int y[] = { (int) y2, (int) (y2 - sy2 + dy2), (int) (y2 - sy2 - sy2), (int) (y2 - sy2 - dy2) };
			if ((thisStyle & IEdge.TOSOLID) != 0)
			{
				this.theGraphics.fillPolygon(x, y, x.length);
			}
			else
			{
				final Color thatColor = this.theGraphics.getColor();
				this.theGraphics.setColor(this.theBackColor);
				this.theGraphics.fillPolygon(x, y, x.length);
				this.theGraphics.setColor(thatColor);
				this.theGraphics.drawPolygon(x, y, x.length);
			}
			break;
		}

		case IEdge.ARROW:
		{
			// a
			final int x[] = { (int) (x2 - sx2 + dx2), (int) x2, (int) (x2 - sx2 - dx2) };
			final int y[] = { (int) (y2 - sy2 + dy2), (int) y2, (int) (y2 - sy2 - dy2) };
			this.theGraphics.drawPolyline(x, y, x.length);
			break;
		}

		case IEdge.HOOK:
		{
			// h
			final int x[] = { (int) (x2 - sx2 + dx2), (int) x2 };
			final int y[] = { (int) (y2 - sy2 + dy2), (int) y2 };
			this.theGraphics.drawPolyline(x, y, x.length);
			break;
		}
		}
	}

	/**
	 * Convert Arc to Arc2D
	 * 
	 * @param thisArc
	 *        arc
	 * @return arc
	 */
	private Arc2D toArc2D(final Arc thisArc)
	{
		final double x1 = xUnitCircleToView(thisArc.x);
		final double y1 = yUnitCircleToView(thisArc.y);
		final double x2 = xUnitCircleToView(thisArc.x - thisArc.r);
		final double y2 = yUnitCircleToView(thisArc.y - thisArc.r);
		double thisStart = Math.toDegrees(-thisArc.start);
		if (thisStart < 0.)
		{
			thisStart += 360.;
		}
		final double thisExtent = Math.toDegrees(-thisArc.angle);
		final Arc2D.Double thisArc2D = new Arc2D.Double();
		thisArc2D.setFrameFromCenter(x1, y1, x2, y2);
		thisArc2D.setAngleStart(thisStart);
		thisArc2D.setAngleExtent(thisExtent);
		thisArc2D.setArcType(Arc2D.OPEN);
		return thisArc2D;
	}

	// D R A W . A R C . T E X T

	/**
	 * Get arc eccentricity
	 * 
	 * @param thisArc2D
	 *        arc
	 * @return eccentricity of arc
	 */
	static private double getEccentricity(final Arc2D thisArc2D)
	{
		final double a = thisArc2D.getWidth() / 2.;
		final double b = thisArc2D.getHeight() / 2.;
		final double d = a > b ? b * b / (a * a) : a * a / (b * b);
		return Math.sqrt(1 - d);
	}

	/**
	 * Get mid-arc point
	 * 
	 * @param thisArc2D
	 *        arc
	 * @return point at mid-arc
	 */
	private Point2D getMidArc(final Arc2D thisArc2D)
	{
		final double a = thisArc2D.getWidth() / 2.;
		final double b = thisArc2D.getHeight() / 2.;
		final double theta = Math.toRadians(-thisArc2D.getAngleStart()) + Math.toRadians(-thisArc2D.getAngleExtent() / 2.);
		final double x = a * Math.cos(theta);
		final double y = b * Math.sin(theta);
		return new Point2D.Double(x + thisArc2D.getCenterX(), y + thisArc2D.getCenterY());
	}

	/**
	 * Get tangent to this Arc2D at given point
	 * 
	 * @param thisArc2D
	 *        arc
	 * @param thisWhere
	 *        point on arc
	 * @return tangent of this arc at given point
	 */
	private double getTangent(final Arc2D thisArc2D, final Point2D thisWhere)
	{
		// center, eccentricity, major axis, minor axis
		final double x = thisWhere.getX() - thisArc2D.getCenterX();
		final double y = thisWhere.getY() - thisArc2D.getCenterY();
		final double e = Painter.getEccentricity(thisArc2D);
		final double a = thisArc2D.getWidth() / 2.;
		final double b = thisArc2D.getHeight() / 2.;

		// angles between foci and point
		double thisAngle1;
		double thisAngle2;
		if (a >= b)
		{
			final double ea = e * a;
			thisAngle1 = Math.atan2(y, x - ea);
			thisAngle2 = Math.atan2(y, x + ea);
		}
		else
		{
			final double eb = e * b;
			thisAngle1 = Math.atan2(y - eb, x);
			thisAngle2 = Math.atan2(y + eb, x);
		}
		// tangent
		final double thisBissector = (thisAngle1 + thisAngle2) / 2.;
		return thisBissector + Math.PI / 2.;
	}

	/**
	 * Draw text
	 * 
	 * @param thisString
	 *        string to be drawn
	 * @param thisWhere
	 *        where to put text (centered on this point)
	 * @param thatOrientation
	 *        text orientation
	 */
	private void drawText(final String thisString, final Point2D thisWhere, final double thatOrientation)
	{
		double thisOrientation = thatOrientation;
		final Graphics2D thisGraphics = (Graphics2D) this.theGraphics;

		// save current transform
		final AffineTransform thatTransform = thisGraphics.getTransform();

		// font metrics
		final FontMetrics thisFontMetrics = thisGraphics.getFontMetrics();
		final int thisWidth = thisFontMetrics.stringWidth(thisString);

		// text orientation
		final boolean reverse = thisOrientation > Math.PI / 2.;
		int thisYShift = 0;
		if (reverse)
		{
			thisOrientation += Math.PI;
			thisYShift = -thisFontMetrics.getAscent();
		}
		else
		{
			thisYShift = thisFontMetrics.getDescent();
		}

		// translate to center of text
		final AffineTransform thisTransform = new AffineTransform(thatTransform);
		thisTransform.translate(thisWhere.getX(), thisWhere.getY());
		thisTransform.rotate(thisOrientation);
		thisGraphics.setTransform(thisTransform);
		thisGraphics.drawString(thisString, -thisWidth / 2, -thisYShift);

		// restore transform
		thisGraphics.setTransform(thatTransform);
	}

	/**
	 * Draw text from (x1,y1) to (x2,y2)
	 * 
	 * @param thisString
	 *        text
	 * @param thisFrom
	 *        from-point
	 * @param thisTo
	 *        to-point
	 */
	private void drawText(final String thisString, final Point2D thisFrom, final Point2D thisTo)
	{
		final double x1 = thisFrom.getX();
		final double y1 = thisFrom.getY();
		final double x2 = thisTo.getX();
		final double y2 = thisTo.getY();

		// orientation
		double thisOrientation = Math.atan2(y1 - y2, x1 - x2);
		if (thisOrientation < 0)
		{
			thisOrientation += Math.PI;
		}

		// where
		final Point2D thisWhere = new Point2D.Double((x2 + x1) / 2., (y2 + y1) / 2.);

		// draw
		drawText(thisString, thisWhere, thisOrientation);
	}

	/**
	 * Mangle string to fit in
	 * 
	 * @param thatString
	 *        string
	 * @param thisFrom
	 *        from-point
	 * @param thisTo
	 *        to-point
	 * @return mangled string or null
	 */
	private String mangleString(final String thatString, final Point2D thisFrom, final Point2D thisTo)
	{
		String thisString = thatString;
		final double cx = thisTo.getX() - thisFrom.getX();
		final double cy = thisTo.getY() - thisFrom.getY();
		final int span = (int) Math.sqrt(cx * cx + cy * cy) - Painter.theTextPadding;

		// label font size
		final FontMetrics fm = this.theGraphics.getFontMetrics();
		final int w = fm.stringWidth(thisString);

		// truncate label
		if (w > span)
		{
			// compute average character width
			final int wunit = fm.stringWidth("x");

			// compute trailing dots width
			final int wdots = fm.stringWidth("...");

			// compute number of characters that fit before dots
			int thisNChars = (span - wdots) / wunit;

			// ensure at least one
			if (thisNChars < 1)
			{
				thisNChars = 1;
			}

			// perform truncation if we actually truncate
			final int thisLen = thisString.length();
			if (thisLen > thisNChars)
			{
				thisString = thisString.substring(0, thisNChars) + "...";
			}
		}
		return thisString;
	}

	// D R A W . L I N E

	/**
	 * Draw line from p1 to p2
	 * 
	 * @param thisFromPoint
	 *        from-point
	 * @param thisToPoint
	 *        to-point
	 */
	private void drawLine(final Point2D thisFromPoint, final Point2D thisToPoint, final int thisStyle)
	{
		if ((thisStyle & IEdge.STROKEMASK) != 0)
		{
			final Graphics2D thisGraphics = (Graphics2D) this.theGraphics;
			final Stroke thatStroke = thisGraphics.getStroke();
			Stroke thisStroke;
			if ((thisStyle & IEdge.DASH) != 0)
			{
				thisStroke = Painter.theDashStroke;
			}
			else if ((thisStyle & IEdge.DOT) != 0)
			{
				thisStroke = Painter.theDotStroke;
			}
			else
			{
				thisStroke = null;
			}

			if (thisStroke != null)
			{
				thisGraphics.setStroke(thisStroke);
				this.theGraphics.drawLine((int) thisFromPoint.getX(), (int) thisFromPoint.getY(), (int) thisToPoint.getX(), (int) thisToPoint.getY());
				thisGraphics.setStroke(thatStroke);
				return;
			}
		}
		this.theGraphics.drawLine((int) thisFromPoint.getX(), (int) thisFromPoint.getY(), (int) thisToPoint.getX(), (int) thisToPoint.getY());
	}

	// D R A W . I M A G E

	/**
	 * Draw image at point
	 * 
	 * @param thisImage
	 *        image
	 * @param where
	 *        location
	 */
	private void drawImage(final Image thisImage, final Point2D where)
	{
		if (thisImage != null)
		{
			final int hi2 = thisImage.getHeight(null) / 2;
			final int wi2 = thisImage.getWidth(null) / 2;
			final int x = (int) where.getX() - wi2;
			final int y = (int) where.getY() - hi2;
			this.theGraphics.drawImage(thisImage, x, y, null);
		}
	}

	/**
	 * Get middle
	 * 
	 * @param x1
	 *        from point
	 * @param x2
	 *        to mpoint
	 * @return middle
	 */
	private Point2D getMiddle(final Point2D x1, final Point2D x2)
	{
		final double x = (x1.getX() + x2.getX()) / 2.0;
		final double y = (x1.getY() + x2.getY()) / 2.0;
		return new Point2D.Double(x, y);
	}

	// D R A W . N O D E S P A C E

	/**
	 * Draw node's space
	 * 
	 * @param thisNode
	 *        node
	 */
	private void drawSpace(final INode thisNode)
	{
		// color
		Color thisBackColor = thisNode.getBackColor();
		if (thisBackColor == null)
		{
			thisBackColor = this.theNodeBackColor;
		}
		// draw
		final Location thisLocation = thisNode.getLocation();
		drawCircle(thisLocation.euclidian.center.re, thisLocation.euclidian.center.im, thisLocation.euclidian.radius, thisBackColor);
	}

	/**
	 * Draw circle
	 * 
	 * @param x
	 *        center x-coordinate
	 * @param y
	 *        center y-coordinate
	 * @param r
	 *        radius
	 * @param thisColor
	 *        color
	 */
	private void drawCircle(final double x, final double y, final double r, final Color thisColor)
	{
		this.theGraphics.setColor(thisColor);
		this.theGraphics.drawOval(xUnitCircleToView(x - r), yUnitCircleToView(y - r), wUnitCircleToView(2 * r), hUnitCircleToView(2 * r));
	}

	// I N T E R S E C T I O N

	/**
	 * Get intersection between line and rectangle
	 * 
	 * @param thisRect
	 *        rectangle
	 * @param thisFrom
	 *        from-point on line
	 * @param thisTo
	 *        to-point on line
	 * @return intersection point
	 */
	private Point2D getIntersection(final Rectangle2D thisRect, final Point2D thisFrom, final Point2D thisTo)
	{
		if (thisRect == null)
			return null;

		final int thisCode = thisRect.outcode(thisFrom);

		// rectangle
		final double x0 = thisRect.getCenterX();
		final double y0 = thisRect.getCenterY();
		final double x = thisRect.getX();
		final double y = thisRect.getY();
		final double w = thisRect.getWidth();
		final double h = thisRect.getHeight();

		// points on top, bottom, left, right
		final double ty = y;
		final double by = y + h;
		final double lx = x;
		final double rx = x + w;

		// line
		final double tanAlpha = (thisTo.getY() - thisFrom.getY()) / (thisTo.getX() - thisFrom.getX());

		// drawPoint((ty - y0) / tanAlpha + x0, ty, Color.MAGENTA);
		// drawPoint((by - y0) / tanAlpha + x0, by, Color.ORANGE);
		// drawPoint(lx, (lx - x0) * tanAlpha + y0, Color.BLUE);
		// drawPoint(rx, (rx - x0) * tanAlpha + y0, Color.CYAN);
		// drawLine(thisFrom, thisTo);

		if ((thisCode & Rectangle2D.OUT_TOP) != 0)
		{
			// compute top intersection
			final double tx = (ty - y0) / tanAlpha + x0;
			if (tx >= lx && tx <= rx)
				return new Point2D.Double(tx, ty);
		}
		if ((thisCode & Rectangle2D.OUT_BOTTOM) != 0)
		{
			// compute bottom intersection
			final double bx = (by - y0) / tanAlpha + x0;
			if (bx >= lx && bx <= rx)
				return new Point2D.Double(bx, by);
		}
		if ((thisCode & Rectangle2D.OUT_LEFT) != 0)
		{
			// compute left intersection
			final double ly = (lx - x0) * tanAlpha + y0;
			if (ly >= ty && ly <= by)
				return new Point2D.Double(lx, ly);
		}
		if ((thisCode & Rectangle2D.OUT_RIGHT) != 0)
		{
			// compute right intersection
			final double ry = (rx - x0) * tanAlpha + y0;
			if (ry >= ty && ry <= by)
				return new Point2D.Double(rx, ry);
		}
		return null;
	}

	/**
	 * Get whether boxes intersect (boxes have to be infalted by one pixel)
	 * 
	 * @param thisBox1
	 *        box1
	 * @param thisBox2
	 *        box2
	 * @return true if nodes' boxes intersect
	 */
	private boolean boxesIntersect(final Rectangle2D thisBox1, final Rectangle2D thisBox2)
	{
		if (thisBox1 != null && thisBox2 != null)
		{
			// inflate
			final Rectangle2D thisFromBox2 = new Rectangle2D.Double();
			thisFromBox2.setFrame(new Point2D.Double(thisBox1.getMinX() - 1, thisBox1.getMinY() - 1), new Dimension((int) thisBox1.getWidth() + 2, (int) thisBox1.getHeight() + 2));
			final Rectangle2D thisToBox2 = new Rectangle2D.Double();
			thisToBox2.setFrame(new Point2D.Double(thisBox2.getMinX() - 1, thisBox2.getMinY() - 1), new Dimension((int) thisBox2.getWidth() + 2, (int) thisBox2.getHeight() + 2));

			// intersection test
			if (thisFromBox2.intersects(thisToBox2))
				return true;
		}
		return false;
	}

	/**
	 * Get intersection between arc and rectangle
	 * 
	 * @param thisRect
	 *        rectangle
	 * @param thisArc
	 *        arc
	 * @return intersection point
	 */
	private Point2D getIntersection(final Rectangle2D thisRect, final Arc2D thisArc)
	{
		if (thisRect == null)
			return null;

		// rectangle
		final double x = thisRect.getCenterX();
		final double y = thisRect.getCenterY();
		final double w = thisRect.getWidth() / 2.;
		final double h = thisRect.getHeight() / 2.;

		// arc
		final double ax = thisArc.getCenterX();
		final double ay = thisArc.getCenterY();
		final double a = thisArc.getWidth() / 2.;
		final double b = thisArc.getHeight() / 2.;
		final double a2 = a * a;
		final double b2 = b * b;

		// rectangle center relative to arc center
		final double drx = x - ax;
		final double dry = y - ay;
		final boolean isSouth = dry < 0.;
		final boolean isEast = drx < 0.;

		// rectangle top, bottom, left, right relative to arc center
		final double drl = drx - w;
		final double drr = drx + w;
		final double drt = dry - h;
		final double drb = dry + h;

		// points on top, bottom, left, right
		final double ty = dry - h;
		final double by = dry + h;
		final double lx = drx - w;
		final double rx = drx + w;
		double k;

		// compute top intersection
		k = ty * ty / b2;
		if (k <= 1.)
		{
			double tx = Math.sqrt((1. - k) * a2);
			if (isEast)
			{
				tx = -tx;
			}
			// if other coodinate is on rectangle
			if (tx >= drl && tx <= drr)
			{
				if (isOnArc(tx, ty, thisArc))
					return new Point2D.Double(tx + ax, ty + ay);
			}
		}

		// compute bottom intersection
		k = by * by / b2;
		if (k <= 1.)
		{
			double bx = Math.sqrt((1. - k) * a2);
			if (isEast)
			{
				bx = -bx;
			}
			// if other coodinate is on rectangle
			if (bx >= drl && bx <= drr)
			{
				if (isOnArc(bx, by, thisArc))
					return new Point2D.Double(bx + ax, by + ay);
			}
		}

		// compute left intersection
		k = lx * lx / a2;
		if (k <= 1.)
		{
			double ly = Math.sqrt((1. - k) * b2);
			if (isSouth)
			{
				ly = -ly;
			}
			// if other coodinate is on rectangle
			if (ly >= drt && ly <= drb)
			{
				if (isOnArc(lx, ly, thisArc))
					return new Point2D.Double(lx + ax, ly + ay);
			}
		}

		// compute right intersection
		k = rx * rx / a2;
		if (k <= 1.)
		{
			double ry = Math.sqrt((1. - k) * b2);
			if (isSouth)
			{
				ry = -ry;
			}
			// if other coodinate is on rectangle
			if (ry >= drt && ry <= drb)
			{
				if (isOnArc(rx, ry, thisArc))
					return new Point2D.Double(rx + ax, ry + ay);
			}
		}
		return null;
	}

	// O N . A R C

	/**
	 * Test whether point is on arc
	 * 
	 * @param dx
	 *        x-offset from arc center
	 * @param dy
	 *        y-offset from arc center
	 * @param thisArc
	 *        arc
	 * @return whether the point is on arc
	 */
	private boolean isOnArc(final double dx, final double dy, final Arc2D thisArc)
	{
		final double a = pointToAngle(dx, dy, thisArc);
		return thisArc.containsAngle(Math.toDegrees(a));
	}

	/**
	 * Compute angle for point
	 * 
	 * @param dx
	 *        x-offset from arc center
	 * @param dy
	 *        y-offset from arc center
	 * @param thisArc
	 *        arc
	 * @return computed angle
	 */
	private double pointToAngle(final double dx, final double dy, final Arc2D thisArc)
	{
		final double thisResult = Math.atan2(-dy * thisArc.getWidth(), dx * thisArc.getHeight());
		return thisResult;
	}
}
