/**
 * Title: Hyperbolic engine
 * Description: Treebolic Engine
 * Version: provider
 * Copyright: (c) 2001-2008
 * Terms of use:see license agreement at http://treebolic.sourceforge.net/en/license.htm
 * Author: Bernard Bou
 * Company: bsys
 * Update: Mon Mar 10 00:00:00 CEST 2008
 */
package treebolic.view;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.JPanel;
import javax.swing.Timer;

import treebolic.control.Controller;
import treebolic.control.MouseAdapter;
import treebolic.core.AbstractLayerOut;
import treebolic.core.Complex;
import treebolic.core.HyperRotation;
import treebolic.core.HyperTransform;
import treebolic.core.Transformer;
import treebolic.model.INode;
import treebolic.model.Location;
import treebolic.model.Model;
import treebolic.model.Settings;

/**
 * View
 * 
 * @author Bernard Bou
 */
public class View extends JPanel
{
	static private final long serialVersionUID = 1L;

	// D A T A

	/**
	 * Data model
	 */
	private Model theModel;

	// A G E N T S

	/**
	 * Painter
	 */
	private final AbstractPainter thePainter;

	/**
	 * Transformer
	 */
	private final Transformer theTransformer;

	/**
	 * Animator
	 */
	private final Animator theAnimator;

	/**
	 * LayerOut
	 */
	private AbstractLayerOut theLayerOut;

	/**
	 * Mouse adapter
	 */
	private MouseAdapter theMouseAdapter;

	/**
	 * Timer
	 */
	private Timer theTimer;

	// G R A P H I C S

	/**
	 * Full redraw flag
	 */
	private boolean invalidatePainterGraphics;

	/**
	 * View size
	 */
	private Dimension theSize;

	/**
	 * Drawing cache
	 */
	private Image theCache;

	// R E F E R E N C E . N O D E S

	/**
	 * Hot node
	 */
	private INode theHotNode;

	/**
	 * Focus node
	 */
	private INode theFocusNode;

	// B E H A V I O U R

	/**
	 * Whether hovering on nodes triggers gaining focus
	 */
	private boolean focusOnHover;

	/**
	 * Whether it has tooltip
	 */
	private boolean hasTooltip;

	/**
	 * Whether tooltips display contents
	 */
	public boolean tooltipDisplaysContent;

	/**
	 * Tooltip break
	 */
	static private int theTooltipLineSpan = 80;

	// A N I M A T I O N

	/**
	 * Animation time slice
	 */
	static private final int ANIMATIONTIMESLICE = 100;

	/**
	 * Animation start delay
	 */
	static private final int ANIMATIONSTARTDELAY = 1000;

	/**
	 * Hot node time slice
	 */
	static private final int HOTNODETIMESLICE = 1000;

	// C O N S T R U C T O R

	/**
	 * Constructor
	 */
	public View()
	{
		// graphics
		setLayout(null);

		// painter
		this.thePainter = new Painter();
		this.thePainter.resetColors();

		// transformer
		this.theTransformer = new Transformer();

		// animator
		this.theAnimator = new Animator(this, View.ANIMATIONTIMESLICE, View.ANIMATIONSTARTDELAY);

		// graphics
		this.theSize = new Dimension(0, 0);
		this.theCache = null;
		this.invalidatePainterGraphics = true;

		// reference nodes
		this.theHotNode = null;
		this.theFocusNode = null;

		// behaviour flags
		this.focusOnHover = true;
		this.hasTooltip = true;
		this.tooltipDisplaysContent = true;
	}

	// C O N N E C T

	/**
	 * Connect controller
	 * 
	 * @param thisController
	 *        controller
	 */
	public void connect(final Controller thisController)
	{
		// mouse
		this.theMouseAdapter = new MouseAdapter(thisController);
		addMouseListener(this.theMouseAdapter);
		addMouseMotionListener(this.theMouseAdapter);
	}

	/**
	 * Set model
	 * 
	 * @param thisModel
	 *        data model
	 */
	public void connect(final Model thisModel)
	{
		this.theModel = thisModel;
		this.theAnimator.start();
	}

	/**
	 * Connect layout agent
	 * 
	 * @param thisLayerOut
	 *        layout agent
	 */
	public void connect(final AbstractLayerOut thisLayerOut)
	{
		this.theLayerOut = thisLayerOut;
	}

	// P A I N T I N G

	/*
	 * (non-Javadoc)
	 * @see treebolic.appletcore.IView#repaint()
	 */
	@Override
	public void repaint()
	{
		if (this.theMouseAdapter != null)
		{
			this.theMouseAdapter.resetHotNode();
		}
		super.repaint();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#paint(java.awt.Graphics)
	 */
	@Override
	public void paint(final Graphics thisScreenGraphics)
	{
		// drag
		if (this.theMouseAdapter.drag())
		{
			this.invalidatePainterGraphics = true;
		}

		// create a cached image the first time or when size changes
		final Dimension thisSize = getSize();
		if (this.theCache == null || thisSize.width != this.theSize.width || thisSize.height != this.theSize.height)
		{
			// invalidate old cache
			this.invalidatePainterGraphics = true;

			// size
			this.theSize = thisSize;

			// cache
			this.theCache = createImage(this.theSize.width, this.theSize.height);
		}

		// setup painter cached graphics and size
		if (this.invalidatePainterGraphics)
		{
			// cached graphics context
			final Graphics thisCacheGraphics = this.theCache.getGraphics();

			// init
			thisCacheGraphics.setFont(getFont());

			// tell painter
			this.thePainter.setup(thisCacheGraphics, this.theSize);
		}

		// paint background
		this.thePainter.paintBackground();

		// paint tree
		this.thePainter.paint(this.theModel.theTree.getRoot(), this.theModel.theTree.getEdges());

		// transfer cache on screen
		thisScreenGraphics.drawImage(this.theCache, 0, 0, null);
	}

	/**
	 * Invalidate graphics and repaint
	 */
	private void invalidateAndRepaint()
	{
		this.invalidatePainterGraphics = true;
		repaint();
	}

	/**
	 * Set default cursor
	 * 
	 * @param linkCursor
	 *        whether to use 'link' cursor
	 */
	public void setHoverCursor(final boolean linkCursor)
	{
		if (linkCursor)
		{
			setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
		}
		else
		{
			setCursor(Cursor.getDefaultCursor());
		}
	}

	// D R A G . N O T I F I C A T I O N

	/**
	 * Drag notification hook
	 */
	public void enterDrag()
	{
		this.thePainter.enterDrag();
		invalidateAndRepaint();
	}

	/**
	 * Leave drag notification hook
	 */
	public void leaveDrag()
	{
		this.thePainter.leaveDrag();
		invalidateAndRepaint();
	}

	// M O U N T . N O T I F I C A T I O N

	/**
	 * Mount notification hook
	 * 
	 * @param thisMountedRoot
	 *        mounted node
	 */
	public void mount(final INode thisMountedRoot)
	{
		this.theTransformer.transform(thisMountedRoot);
		repaint();
	}

	/**
	 * Unmount notification hook
	 */
	public void umount(final INode thisMountingRoot)
	{
		this.theTransformer.transform(thisMountingRoot);
		repaint();
	}

	// F O C U S

	/**
	 * Control whether hovering on node triggers gaining focus
	 * 
	 * @param thisFlag
	 *        whether hovering on node triggers gaining focus (null toggles value)
	 */
	@SuppressWarnings("serial")
	public void setFocusOnHover(final Boolean thisFlag)
	{
		this.focusOnHover = thisFlag != null ? thisFlag : !this.focusOnHover;

		// implement
		if (this.focusOnHover)
		{
			// timer creation
			if (this.theTimer == null)
			{
				this.theTimer = new Timer(View.HOTNODETIMESLICE, new AbstractAction()
				{
					@SuppressWarnings("synthetic-access")
					@Override
					public void actionPerformed(final ActionEvent e)
					{
						final INode thisHotNode = View.this.theMouseAdapter.getHotNode();
						final boolean again = View.this.theHotNode == thisHotNode;
						View.this.theHotNode = thisHotNode;
						if (!again || View.this.theAnimator.isRunning() || View.this.theHotNode == null || View.this.theHotNode.getLocation().hyper.center.equals(Complex.ZERO))
							return;
						animateToCenter(View.this.theHotNode, true);
					}
				});
				this.theTimer.setRepeats(true);
			}

			// start
			this.theTimer.start();
		}
		else
		{
			if (this.theTimer != null)
			{
				this.theTimer.stop();
			}
		}
	}

	// O R I E N T A T I O N

	/**
	 * Preserve orientation
	 * 
	 * @param thisFlag
	 *        whether transformations preserve orientation (null toggles value)
	 */
	public void setPreserveOrientation(final Boolean thisFlag)
	{
		this.theTransformer.setPreserveOrientation(thisFlag != null ? thisFlag : !this.theTransformer.getPreserveOrientation());
	}

	// O F F S E T

	/**
	 * Set shift
	 * 
	 * @param cx0
	 *        x shift
	 * @param cy0
	 *        y shift
	 * @param xinc
	 *        whether to add to current x shift
	 * @param yinc
	 *        whether to add to current y shift
	 */
	public void setShift(final double cx0, final double cy0, final boolean xinc, final boolean yinc)
	{
		double cx = cx0;
		double cy = cy0;
		if (xinc)
		{
			cx += this.thePainter.getXShift();
		}
		if (yinc)
		{
			cy += this.thePainter.getYShift();
		}
		this.thePainter.setXShift(cx);
		this.thePainter.setYShift(cy);
	}

	/**
	 * Set zoom factor
	 * 
	 * @param f
	 *        zoom factor
	 */
	public void setZoomFactor(final double f)
	{
		this.thePainter.setZoomFactor(f);
	}

	// R E N D E R I N G

	/**
	 * Set linear/hyperbolic rendering of edges
	 * 
	 * @param thisFlag
	 *        whether edges are rendered in hyperbolic way (null toggles value)
	 */
	public void setHyperbolicEdges(final Boolean thisFlag)
	{
		this.thePainter.setHyperbolicEdges(thisFlag != null ? thisFlag : !this.thePainter.getHyperbolicEdges());
	}

	// T O O L T I P

	/**
	 * Enable/disable tooltips
	 * 
	 * @param thisFlag
	 *        whether to display tooltip (null toggles value)
	 */
	public void setHasTooltip(final Boolean thisFlag)
	{
		this.hasTooltip = thisFlag != null ? thisFlag : !this.hasTooltip;
		setToolTipText(null);
	}

	/**
	 * Enable/disable displaying content in tooltips
	 * 
	 * @param thisFlag
	 *        whether tooltip displays content (null toggles value)
	 */
	public void setTooltipDisplaysContent(final Boolean thisFlag)
	{
		this.tooltipDisplaysContent = thisFlag != null ? thisFlag : !this.tooltipDisplaysContent;
	}

	/**
	 * Set tooltip to contain information about this node
	 * 
	 * @param thisNode
	 *        node
	 */
	public void setToolTip(final INode thisNode)
	{
		if (!this.hasTooltip)
			return;

		final String thisLabel = thisNode.getLabel();
		final String thisContent = thisNode.getContent();
		if (thisLabel == null && (this.tooltipDisplaysContent ? thisContent == null : true))
			return;

		final StringBuffer thisBuffer = new StringBuffer();
		thisBuffer.append("<html>");

		if (thisLabel != null && !thisLabel.isEmpty())
		{
			thisBuffer.append("<strong>");
			thisBuffer.append(thisLabel);
			thisBuffer.append("</strong><br>");
		}

		if (this.tooltipDisplaysContent)
		{
			if (thisContent != null && !thisContent.isEmpty())
			{
				final String[] theseLines = thisContent.split("\n");
				for (final String thisLine : theseLines)
				{
					final StringBuffer thisLineBuffer = new StringBuffer(thisLine);

					// force break after x characters
					for (int offset = View.theTooltipLineSpan; offset < thisLineBuffer.length(); offset += View.theTooltipLineSpan)
					{
						thisLineBuffer.insert(offset, "<br>");
					}

					// append processed line with break
					thisBuffer.append(thisLineBuffer);
					thisBuffer.append("<br>");
				}
			}
		}
		thisBuffer.append("</html>");
		setToolTipText(thisBuffer.toString());
	}

	// P O P U P

	/**
	 * Control whether view has popup menu
	 * 
	 * @param thisFlag
	 *        whether view has popup menu (null toggles value)
	 */
	public void setPopUpMenu(final Boolean thisFlag)
	{
		this.theMouseAdapter.hasPopUp = thisFlag != null ? thisFlag : !this.theMouseAdapter.hasPopUp;
	}

	// A P P L Y . S E T T I N G S

	/**
	 * Set view behaviour
	 * 
	 * @param theseSettings
	 *        settings
	 */
	public void apply(final Settings theseSettings)
	{
		// view settings
		if (theseSettings.theHasPopUpMenuFlag != null)
		{
			setPopUpMenu(theseSettings.theHasPopUpMenuFlag);
		}
		if (theseSettings.theFocusOnHoverFlag != null)
		{
			setFocusOnHover(theseSettings.theFocusOnHoverFlag);
		}
		if (theseSettings.theHasToolTipFlag != null)
		{
			setHasTooltip(theseSettings.theHasToolTipFlag);
		}
		if (theseSettings.theToolTipDisplaysContentFlag != null)
		{
			setTooltipDisplaysContent(theseSettings.theToolTipDisplaysContentFlag);
		}
		if (theseSettings.thePreserveOrientationFlag != null)
		{
			setPreserveOrientation(theseSettings.thePreserveOrientationFlag);
		}
		if (theseSettings.theXShift != null || theseSettings.theYShift != null)
		{
			setShift(theseSettings.theXShift == null ? 0. : theseSettings.theXShift, theseSettings.theYShift == null ? 0. : theseSettings.theYShift, false, false);
		}

		// painter
		this.thePainter.setColors(theseSettings.theBackColor, theseSettings.theForeColor, theseSettings.theNodeBackColor, theseSettings.theNodeForeColor, theseSettings.theTreeEdgeColor, theseSettings.theEdgeColor);
		this.thePainter.setFonts(theseSettings.theFontFace, theseSettings.theFontSize, theseSettings.theFontSizeStep);
		this.thePainter.setHyperbolicEdges(theseSettings.theHyperbolicEdgesFlag);
		this.thePainter.setEdgeStyles(theseSettings.theTreeEdgeStyle, theseSettings.theEdgeStyle);
	}

	/**
	 * Set images
	 * 
	 * @param thisBackgroundImage
	 *        background image
	 * @param thisDefaultNodeImage
	 *        default node image
	 * @param thisDefaultTreeEdgeImage
	 *        default tree edge image
	 * @param thisDefaultEdgeImage
	 *        default edge image
	 */
	public void setImages(final Image thisBackgroundImage, final Image thisDefaultNodeImage, final Image thisDefaultTreeEdgeImage, final Image thisDefaultEdgeImage)
	{
		this.thePainter.setImages(thisBackgroundImage, thisDefaultNodeImage, thisDefaultTreeEdgeImage, thisDefaultEdgeImage);
	}

	// S E T . T R A N S F O R M

	public void resetTransform()
	{
		this.theTransformer.setTransform(HyperTransform.NULLTRANSFORM);
	}

	// A P P L Y . T R A N S F O R M

	/**
	 * Compose this transform with current and apply
	 * 
	 * @param thisTransform
	 *        transform to compose with current transform
	 */
	public void applyComposedTransform(final HyperTransform thisTransform)
	{
		this.theTransformer.composeTransform(thisTransform);
		this.theTransformer.transform(this.theModel.theTree.getRoot());
	}

	/**
	 * Apply transform
	 * 
	 * @param thisTransform
	 *        transform
	 */
	public void applyTransform(final HyperTransform thisTransform)
	{
		this.theTransformer.setTransform(thisTransform);
		this.theTransformer.transform(this.theModel.theTree.getRoot());
	}

	/**
	 * Apply null transform
	 */
	public void applyNullTransform()
	{
		this.theTransformer.reset(this.theModel.theTree.getRoot());
	}

	/**
	 * Apply initial transform
	 */
	public void applyInitialTransform()
	{
		final Complex thisStart = new Complex(-1, -1).normalize().multiply(0.9);
		final HyperTransform thisTransform = this.theTransformer.makeTransform(thisStart, Complex.ZERO, this.theLayerOut.getOrientation());
		applyTransform(thisTransform);
	}

	// D E L T A . M O V E

	/**
	 * Compose delta translation
	 * 
	 * @param thisStart
	 *        unit circle delta vector start
	 * @param thisEnd
	 *        unit circle delta vector start
	 */
	public void composeTranslate(final Complex thisStart, final Complex thisEnd)
	{
		final HyperTransform thisTransform = this.theTransformer.makeTransform(thisStart, thisEnd, this.theLayerOut.getOrientation());
		applyComposedTransform(thisTransform);
		repaint();
	}

	/**
	 * Compose delta rotation
	 * 
	 * @param thisStart
	 *        unit circle delta vector start
	 * @param thisEnd
	 *        unit circle delta vector end
	 */
	public void composeRotate(final Complex thisStart, final Complex thisEnd)
	{
		final HyperRotation thisRotation = new HyperRotation(Complex.ONE);
		thisRotation.div(thisEnd, thisStart);
		thisRotation.normalize();
		applyComposedTransform(new HyperTransform(thisRotation));
		repaint();
	}

	// M O V E

	/**
	 * Translate tree according to vector
	 * 
	 * @param thisStart
	 *        unit circle vector start
	 * @param thisEnd
	 *        unit circle vector end
	 */
	private void translate(final Complex thisStart, final Complex thisEnd)
	{
		final HyperTransform thisTransform = this.theTransformer.makeTransform(thisStart, thisEnd, this.theLayerOut.getOrientation());
		applyTransform(thisTransform);
		repaint();
	}

	/**
	 * Translate node to point
	 * 
	 * @param thisNode
	 *        node
	 * @param thisDestination
	 *        unit circle destination location
	 */
	public void moveTo(final INode thisNode, final Complex thisDestination)
	{
		final Location thisLocation = thisNode.getLocation();
		translate(thisLocation.hyper.center0, thisDestination);
		this.theFocusNode = thisNode;
	}

	/**
	 * Translate node to unit circle center
	 * 
	 * @param thisNode
	 *        node
	 */
	public void moveToCenter(final INode thisNode)
	{
		final Location thisLocation = thisNode.getLocation();
		translate(thisLocation.hyper.center0, Complex.ZERO);
		this.theFocusNode = thisNode;
	}

	// A N I M A T E

	/**
	 * Animate to unit circle center
	 * 
	 * @param thisNode
	 *        node
	 * @param now
	 *        whether to start now
	 */
	public void animateToCenter(final INode thisNode, final boolean now)
	{
		final Location thisLocation = thisNode.getLocation();
		final Complex thisFrom = thisLocation.hyper.center;
		animate(thisFrom, Complex.ZERO, now);
		this.theFocusNode = thisNode;
	}

	/**
	 * Animate node to destination location
	 * 
	 * @param thisNode
	 *        node
	 * @param thisDestination
	 *        unit circle destination location
	 * @param now
	 *        whether to start now
	 */
	public void animateTo(final INode thisNode, final Complex thisDestination, final boolean now)
	{
		final Location thisLocation = thisNode.getLocation();
		final Complex thisFrom = thisLocation.hyper.center;
		animate(thisFrom, thisDestination, now);
		this.theFocusNode = thisNode;
	}

	/**
	 * Animate tree. Translate as per vector(start,end)
	 * 
	 * @param thisFrom
	 *        unit circle vector start
	 * @param thisTo
	 *        unit circle vector end
	 * @param now
	 *        whether to start now
	 */
	private synchronized void animate(final Complex thisFrom, final Complex thisTo, final boolean now)
	{
		final Animation thisAnimation = Animation.make(thisFrom, thisTo, this.theTransformer, this.theLayerOut.getOrientation(), 0);
		this.theAnimator.run(thisAnimation, now);
	}

	// S E A R C H I N G

	/**
	 * Find node given view location
	 * 
	 * @param vx
	 *        view x location
	 * @param vy
	 *        view y location
	 * @return node
	 */
	public INode findNode(final int vx, final int vy)
	{
		final Complex thisEuclidianLocation = this.thePainter.viewToUnitCircle(vx, vy, this.theSize);
		return this.theModel.theTree.findNodeAt(this.theModel.theTree.getRoot(), thisEuclidianLocation);
	}

	/**
	 * Get focused node
	 * 
	 * @return focus noe (at center of unit circle)
	 */
	public INode getFocusNode()
	{
		return this.theFocusNode;
	}

	// S P A C E . C O N V E R S I O N

	/**
	 * Convert from view coordinates to unit circle euclidian coordinates
	 * 
	 * @param vx
	 *        view x-coordinate
	 * @param vy
	 *        view x-coordinate
	 * @return unit circle euclidian coordinate
	 */
	public Complex viewToUnitCircle(final int vx, final int vy)
	{
		return this.thePainter.viewToUnitCircle(vx, vy, getSize());
	}
}
